/*
 * Copyright (c) 2009,
 * @author Mitsuharu Morisawa
 *
 * AIST
 *
 * All rights reserved.
 *
 * This program is made available under the terms of the Eclipse Public License
 * v1.0 which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 */
#include "mwalk.h"


mwalk::mwalk(istringstream& strm)
{
  wait_request = false;
  state_logger_mode = false;
  setupMethodTable();
  seq = NULL;
  pg = new MwalkGenerator(strm);
  fp = new MwalkFootPlanner(pg->getFootStepManagerPtr());
}

mwalk::~mwalk()
{
  delete fp;
  delete pg;
}

bool mwalk::setup( robot_state *rs, motor_command *mc )
{
  clearExecTime(mExecTime);
  pg->Reset();
  
  if( !seq ){
    try{
      seq = SequencePlayer::_narrow(manager->find("seq", ""));
    }catch(...){
      cerr << "Sequence player can't be found." << endl;

      return false;
    }
  }
  
  return true;
}

void mwalk::control( robot_state *rs, motor_command *mc )
{
  initExecTime(mExecTime);
  pushExecTime(mExecTime);
  
  switch(pg->mmode){
    
  case INITCONF:
    // stand by walk
    pg->initConfiguration(mc->basePos.get_buffer(), mc->baseAtt.get_buffer(),
			  mc->angle.get_buffer(), mc->zmp.get_buffer());
    pg->setMotionCommand(mc->basePos.get_buffer(), mc->baseAtt.get_buffer(),
			 mc->angle.get_buffer(), mc->zmp.get_buffer(),
			 mc->accel[0].get_buffer());
    break;
    
  case ZMPFIX:
    // apply ZMP compensation to offline pattern
    pg->setStateForZmpfix(mc->basePos.get_buffer(), mc->baseAtt.get_buffer(),
			  mc->angle.get_buffer(), mc->zmp.get_buffer());
    pg->setMotionCommand(mc->basePos.get_buffer(), mc->baseAtt.get_buffer(),
			 mc->angle.get_buffer(), mc->zmp.get_buffer(),
			 mc->accel[0].get_buffer());
    
    if( pg->mmode == INITCONF ){
      if( seq != NULL )
	seq->setReferenceState(*mc, DT);

      if( wait_request )
	sem_post(&wait_sem);
    }
    
    break;
    
  case WALKING:
    // main routine for walking pattern generation
    pg->calcZMPTrackingControl(rs->zmp.get_buffer(), rs->accel[0].get_buffer());
    pg->updateWalkingState();
    pushExecTime(mExecTime);
    
    pg->makeSwingLegMotion();
    pg->makeUpperBodyMotion();
    
    //#define WHOLE_BODY_MOTION
#ifndef WHOLE_BODY_MOTION
    pg->makeCOGMotion();
#else
    pushExecTime(mExecTime);
    pg->makeWholeBodyMotion();
    //pg->updateWholeBodyMotion();
#endif
    pushExecTime(mExecTime);
    
    pg->setMotionCommand(mc->basePos.get_buffer(), mc->baseAtt.get_buffer(),
			 mc->angle.get_buffer(), mc->zmp.get_buffer(),
			 mc->accel[0].get_buffer());
    
    if( pg->mmode == INITCONF ){
      if( seq != NULL )
	seq->setReferenceState(*mc, DT);
      
      if( wait_request )
	sem_post(&wait_sem);
    }
  }
  
  if( state_logger_mode )
    pushStateLogger();
  
  pushExecTime(mExecTime);
  calcExecTime(mExecTime);
}

#define FORCE_THRESHOLD 100 //[N]
bool mwalk::cleanup( robot_state *rs, motor_command *mc )
{
  if( pg->mmode == INITCONF )
    return true;
  
  return false;
}

plugin *create_plugin(istringstream &strm)
{
  return new mwalk(strm);
}

void mwalk::setupMethodTable(void)
{
  register_method(":start-walk", (method)&mwalk::m_start_walk);
  register_method(":start-walk-nowait", (method)&mwalk::m_start_walk_nowait);
  register_method(":start-zmpfix", (method)&mwalk::m_start_zmpfix);
  register_method(":stop-zmpfix", (method)&mwalk::m_stop_zmpfix);
  register_method(":set-foot-pos",(method)&mwalk::m_set_foot_pos);
  register_method(":set-next-foot-pos",(method)&mwalk::m_set_next_foot_pos);
  register_method(":set-target-pos", (method)&mwalk::m_set_target_pos);
  register_method(":set-target-pos-nowait", (method)&mwalk::m_set_target_pos_nowait);
  register_method(":set-arc", (method)&mwalk::m_set_arc);
  register_method(":set-arc-nowait", (method)&mwalk::m_set_arc_nowait);
  register_method(":set-target-pos-passgate", (method)&mwalk::m_set_target_pos_passgate);
  register_method(":push-walk-command", (method)&mwalk::m_push_walk_command);
  register_method(":load-footpos", (method)&mwalk::m_load_footpos);
  register_method(":first-step", (method)&mwalk::m_first_step);
  register_method(":set-profile-horizontal", (method)&mwalk::m_set_profile_horizontal);
  register_method(":set-profile-vertical", (method)&mwalk::m_set_profile_vertical);
  register_method(":set-pivot-translation", (method)&mwalk::m_set_pivot_translation);
  register_method(":set-ankle-rotation", (method)&mwalk::m_set_ankle_rotation);
  register_method(":set-waist-middle-ratio", (method)&mwalk::m_set_waist_middle_ratio);
  register_method(":set-swing-height", (method)&mwalk::m_set_swing_height);
  register_method(":set-landoff-angle", (method)&mwalk::m_set_landoff_angle);
  register_method(":set-touchdown-angle", (method)&mwalk::m_set_touchdown_angle);
  register_method(":set-backstep-pitch-angle", (method)&mwalk::m_set_backstep_pitch_angle);
  register_method(":set-step-limit-x", (method)&mwalk::m_set_step_limit_x);
  register_method(":set-step-limit-y", (method)&mwalk::m_set_step_limit_y);
  register_method(":set-step-limit-z", (method)&mwalk::m_set_step_limit_z);
  register_method(":set-step-limit-yaw", (method)&mwalk::m_set_step_limit_yaw); 
  register_method(":set-step-yaw-ratio", (method)&mwalk::m_set_step_yaw_ratio);
  register_method(":set-waist-yaw-ratio", (method)&mwalk::m_set_waist_yaw_ratio);
  register_method(":set-arm-swing-ratio", (method)&mwalk::m_set_arm_swing_ratio);
  register_method(":set-zmp-ankle",(method)&mwalk::m_set_zmp_ankle);
  register_method(":set-zmp-path",(method)&mwalk::m_set_zmp_path);
  register_method(":set-ankle-ds-rotation-ratio", (method)&mwalk::m_set_ankle_ds_rotation_ratio);
  register_method(":trunk-roll-mode", (method)&mwalk::m_trunk_roll_mode);
  register_method(":time-adj", (method)&mwalk::m_time_adj);
  register_method(":zmpshaping", (method)&mwalk::m_zmpshaping);
  register_method(":dfilter-preview", (method)&mwalk::m_dfilter_preview);
  register_method(":init-cog-offset-gain", (method)&mwalk::m_init_cog_offset_gain);
  register_method(":update-cog-offset-gain", (method)&mwalk::m_update_cog_offset_gain);
  register_method(":load-zmpshaping-preview", (method)&mwalk::m_load_zmpshaping_pgain);
  register_method(":load-dfilter-preview", (method)&mwalk::m_load_dfilter_pgain);
  register_method(":reactive-motion", (method)&mwalk::m_reactive_motion);
  register_method(":state-logger-mode", (method)&mwalk::m_state_logger_mode);
  register_method(":save-state-logger", (method)&mwalk::m_save_state_logger);
  register_method(":stat", (method)&mwalk::m_stat);
  register_method(":check-joints-status", (method)&mwalk::m_check_joints_status);
  register_method(":joints-status", (method)&mwalk::m_joints_status);
  register_method(":debug-level", (method)&mwalk::m_debug_level);
  register_method(":debug-info", (method)&mwalk::m_debug_info);
  
#ifdef WITH_TOE_JOINT
  register_method(":set-zmp-toe", (method)&mwalk::m_set_zmp_toe);
  register_method(":set-toe-limit-pitch", (method)&mwalk::m_set_toe_limit_pitch);
#endif
}

void mwalk::pushStateLogger(void)
{
  vector<double> pos, vel;
  if( !pg->flg_status[DFILTER_PREVIEW] ){
    for( int i = 0 ; i < DOF ; i++ ){
      pos.push_back(pg->q_cmd[i]);
      vel.push_back(pg->qdot_cmd[i]);
    }
  }
  else{
    for( int i = 0 ; i < DOF ; i++ ){
      pos.push_back(pg->q_cmd2[i]);
      vel.push_back(pg->qdot_cmd2[i]);
    }
  }
  vector3 hip(rpyFromRot(pg->waistLink->R));
  
  state_logger.pushWaist(pg->waistLink->p, pg->waistLink->R, pg->waistLink->v, pg->waistLink->w);
  state_logger.pushPos(pos);
  state_logger.pushVel(vel);
  state_logger.pushZMP(pg->Zmp_ref);
  state_logger.pushHip(hip);
  state_logger.pushGsens(pg->Gsensor_offset);
  state_logger.pushZMPRef(pg->Zmp_ideal);
  state_logger.pushPivot(pg->Pivot_cmd);
  state_logger.pushAnkle(pg->Ankle_cmd[RLEG].P, pg->Ankle_cmd[RLEG].R, pg->Ankle_cmd[RLEG].V, pg->Ankle_cmd[RLEG].W,
			 pg->Ankle_cmd[LLEG].P, pg->Ankle_cmd[LLEG].R, pg->Ankle_cmd[LLEG].V, pg->Ankle_cmd[LLEG].W
			 );
  
  vector3 Foot_rpy[2];
  Foot_rpy[RLEG] = rpyFromRot(pg->Foot_cmd[RLEG].R);
  Foot_rpy[LLEG] = rpyFromRot(pg->Foot_cmd[LLEG].R);
  state_logger.pushFoot(pg->Foot_cmd[RLEG].P, pg->Foot_cmd[RLEG].V, Foot_rpy[RLEG],
			pg->Foot_cmd[LLEG].P, pg->Foot_cmd[LLEG].V, Foot_rpy[LLEG]);
}

int mwalk::getFloor(string& str_floor)
{
  if( str_floor ==  "flat" || str_floor == "FLAT" )
    return FLAT;
  else if( str_floor == "upstairs" || str_floor == "up" || str_floor == "UPSTAIRS" || str_floor == "UP" )
    return UPSTAIRS;
  else if( str_floor == "downstairs" || str_floor == "down" || str_floor == "DOWNSTAIRS" || str_floor == "DOWN" )
    return DOWNSTAIRS;
  
  return FLAT;
}

void mwalk::m_set_profile_horizontal(istringstream& strm)
{
  double Tprofile[4];
  string str_floor;
  strm >> str_floor >> Tprofile[0] >> Tprofile[1] >> Tprofile[2] >> Tprofile[3];
  
  int floor = getFloor(str_floor);
  for( int i = 0 ; i < 4 ; i++ )
    pg->Tprofile_horizontal[floor][i] = Tprofile[i];
  
  cout << str_floor << " : Tprofile_horizontal = "
       << pg->Tprofile_horizontal[floor][0] << "," << pg->Tprofile_horizontal[floor][1] << ","
       << pg->Tprofile_horizontal[floor][2] << "," << pg->Tprofile_horizontal[floor][3] << endl;
}

void mwalk::m_set_profile_vertical(istringstream &strm)
{
  double Tprofile[8];
  string str_floor;
  strm >> str_floor >> Tprofile[0] >> Tprofile[1] >> Tprofile[2] >> Tprofile[3] >> Tprofile[4] >> Tprofile[5] >> Tprofile[6] >> Tprofile[7];  
  
  int floor = getFloor(str_floor);
  for( int i = 0 ; i < 8 ; i++ )
    pg->Tprofile_vertical[floor][i] = Tprofile[i];
  
  cout << str_floor << " : Tprofile_vertial = " 
       << pg->Tprofile_vertical[floor][0] << "," << pg->Tprofile_vertical[floor][1] << ","
       << pg->Tprofile_vertical[floor][2] << "," << pg->Tprofile_vertical[floor][3] << ","
       << pg->Tprofile_vertical[floor][4] << "," << pg->Tprofile_vertical[floor][5] << ","
       << pg->Tprofile_vertical[floor][6] << "," << pg->Tprofile_vertical[floor][7] << endl;  
}

void mwalk::m_set_pivot_translation(istringstream& strm)
{
  double Tp[2];
  string str_floor;
  strm >> str_floor >> Tp[0] >> Tp[1];
  
  int floor = getFloor(str_floor);
  for( int i = 0 ; i < 2 ; i++ )  
    pg->TP_pivot[floor][i] = Tp[i];
  
  cout << str_floor << " : TP_pivot = " 
       << pg->TP_pivot[floor][0] << "," << pg->TP_pivot[floor][1] << endl;
}

void mwalk::m_set_ankle_rotation(istringstream& strm)
{
  double Tr[4];
  string str_floor;
  strm >> str_floor >> Tr[0] >> Tr[1] >> Tr[2] >> Tr[3];
  
  int floor = getFloor(str_floor);
  for( int i = 0 ; i < 4 ; i++ )  
    pg->TR_ankle[floor][i] = Tr[i];
  
  cout << str_floor << " : TR_ankle = "
       << pg->TR_ankle[floor][0] << "," << pg->TR_ankle[floor][1] << ","
       << pg->TR_ankle[floor][2] << "," << pg->TR_ankle[floor][3] << endl;
}

void mwalk::m_set_swing_height(istringstream& strm)
{
  double XPmiddle, SWmiddle;
  string str_floor;
  strm >> str_floor >> XPmiddle >> SWmiddle;
  
  int floor = getFloor(str_floor);
  pg->XPmiddle_ratio[floor] = XPmiddle;
  pg->SWmiddle_point[floor] = SWmiddle;
  
  cout << str_floor << " : XPmiddle_ratio = " << pg->XPmiddle_ratio[floor] << ", SWmiddle_point = " << pg->SWmiddle_point[floor] << endl;
}

void mwalk::m_set_waist_middle_ratio(istringstream& strm)
{
  strm >> pg->THmiddle_ratio >> pg->XHmiddle_ratio;
  
  cout << "THmiddle_ratio = " << pg->THmiddle_ratio << ", XHmiddle_ratio = " << pg->XHmiddle_ratio << endl;
}

void mwalk::m_set_landoff_angle(istringstream& strm)
{
  vector3 rlandoff(0.0);
  string str_floor;
  strm >> str_floor >> rlandoff(X) >> rlandoff(Y) >> rlandoff(Z);
  
  int floor = getFloor(str_floor);
  pg->R_landoff[floor] = rlandoff * DegToRad; 
  
  cout << str_floor << " : Landoff angle : " << rlandoff << endl;
}

void mwalk::m_set_touchdown_angle(istringstream& strm)
{
  vector3 rtouchdown(0.0);
  string str_floor;
  strm >> str_floor >> rtouchdown(X) >> rtouchdown(Y) >> rtouchdown(Z);
  
  int floor = getFloor(str_floor);
  pg->R_touchdown[floor] = rtouchdown * DegToRad;
  
  cout << str_floor << " : Touchdown angle : " << rtouchdown << endl;
}

void mwalk::m_set_backstep_pitch_angle(istringstream& strm)
{
  double blandoff = 0.0, btouchdown = 0.0;
  string str_floor;
  strm >> str_floor >> blandoff >> btouchdown;
  
  int floor = getFloor(str_floor);
  pg->R_blandoff[floor]   = blandoff   * DegToRad;
  pg->R_btouchdown[floor] = btouchdown * DegToRad;
  
  cout << str_floor << " : Backstep pich angle(landoff/touchdown) : " << blandoff << "," << btouchdown << endl;
}

void mwalk::m_set_step_limit_x(istringstream& strm)
{
  strm >> pg->STEP_X_FORWARD_MAX >> pg->STEP_X_BACK_MAX;
  if( pg->STEP_X_BACK_MAX > 0.0 )
    pg->STEP_X_BACK_MAX = -pg->STEP_X_BACK_MAX;
  
  cout << "step-limit-x(fwd,bck) = (" << pg->STEP_X_FORWARD_MAX << "," << pg->STEP_X_BACK_MAX << ")" << endl;
}


void mwalk::m_set_step_limit_y(istringstream& strm)
{
  strm >> pg->STEP_Y_IN_MAX >> pg->STEP_Y_OUT_MAX;
  cout << "step-limit-y(inside,outside) = (" << pg->STEP_Y_IN_MAX << "," << pg->STEP_Y_OUT_MAX << ")" << endl;
}

void mwalk::m_set_step_limit_z(istringstream& strm)
{
  strm >> pg->STEP_Z_UP_MAX >> pg->STEP_Z_DOWN_MAX;
  cout << "step-limit-z(up,down) = (" << pg->STEP_Z_UP_MAX << "," << pg->STEP_Z_DOWN_MAX << ")" << endl;
}

void mwalk::m_set_step_limit_yaw(istringstream& strm)
{
  strm >> pg->STEP_YAW_IN_MAX >> pg->STEP_YAW_OUT_MAX;
  pg->STEP_YAW_IN_MAX = ToRad(fabs(pg->STEP_YAW_IN_MAX));
  pg->STEP_YAW_OUT_MAX = ToRad(fabs(pg->STEP_YAW_OUT_MAX));
  
  cout << "step-limit-yaw(inside,outside) = " << ToDeg(pg->STEP_YAW_IN_MAX) << "," << ToDeg(pg->STEP_YAW_OUT_MAX) << endl;
}

void mwalk::m_set_step_yaw_ratio(istringstream& strm)
{
  strm >> pg->STEP_YAW_RATIO;
  cout << "step-yaw-ratio = " << pg->STEP_YAW_RATIO << endl;
}

void mwalk::m_set_waist_yaw_ratio(istringstream& strm)
{
  strm >> pg->WAIST_YAW_RATIO;
  cout << "waist-yaw-ratio = " << pg->WAIST_YAW_RATIO << endl;
}

void mwalk::m_set_arm_swing_ratio(istringstream& strm)
{
  double arm_swing_ratio_ = 0.0;
  strm >> arm_swing_ratio_;
  pg->ARM_SWING_RATIO = arm_swing_ratio_*ToRad(20.0)/pg->STEP_X_FORWARD_MAX;
  cout << "arm-swing-ratio = " << arm_swing_ratio_ << endl;
}

void mwalk::m_start_walk(istringstream& strm)
{
  pg->FootQueueNext.clear();
  pg->flg_status[WALK_REQUEST] = true;
  wait_request = true;
  cout << "start walk" << endl;

  sem_wait(&wait_sem);
}

void mwalk::m_start_walk_nowait(istringstream& strm)
{
  pg->FootQueueNext.clear();
  pg->flg_status[WALK_REQUEST] = true;
  wait_request = false;
  cout << "start walk" << endl;
}

void mwalk::m_start_zmpfix(istringstream& strm)
{
  pg->FootQueueNext.clear();
  pg->flg_status[ZMPFIX_REQUEST] = true;
  wait_request = true;
  cout << "start zmpfix" << endl;
}

void mwalk::m_stop_zmpfix(istringstream& strm)
{
  pg->FootQueueNext.clear();
  pg->flg_status[STOP_ZMPFIX] = true;
  wait_request = false;
  cout << "stop zmpfix" << endl;
}

void mwalk::m_first_step(istringstream& strm)
{
  string swg_leg = "RLEG";
  strm >> swg_leg;
  
  if( swg_leg == "RLEG" )
    pg->first_step = RLEG;
  else if( swg_leg == "LLEG" )
    pg->first_step = LLEG;
  else
    pg->first_step = -1;
  
  cout << "first step is " << ((pg->first_step == RLEG) ? "RLEG" : ((pg->first_step == LLEG) ? "LLEG" : "FREE")) << endl;
}

void mwalk::m_set_target_pos(istringstream& strm)
{
  double tss = (double)Nss*DT;
  double tds = (double)Nds*DT;
  double dx_cmd = 0.0, dy_cmd = 0.0, dyaw_cmd = 0.0;
  strm >> dx_cmd >> dy_cmd >> dyaw_cmd >> tss >> tds;
  
  int nss = (int)round(tss/DT);
  int nds = (int)round(tds/DT);
  fp->setTargetPos(pg->FootQueue, dx_cmd, dy_cmd, ToRad(dyaw_cmd), nss, nds);
  
  // last step
  FootPlace FP;
  FP.p = 0.0;
  FP.rpy = 0.0;
  FP.nss = nss;
  FP.nds = nds;
  FP.support_leg = pg->FootQueue.back().support_leg == RLEG ? LLEG : RLEG;
  pg->FootQueue.push_back(FP);  
  
  if( pg->verbose ){
    for( int i = 0 ; i < (int)pg->FootQueue.size() ; i++ )
      cout << "pg->FootQueue(x,y,th) = " << pg->FootQueue[i].p(X) << "," << pg->FootQueue[i].p(Y) << "," << ToDeg(pg->FootQueue[i].rpy(Z))
	   << ", sup = " << ((pg->FootQueue[i].support_leg==RLEG) ? "RLEG" : ((pg->FootQueue[i].support_leg==LLEG) ? "LLEG" : "Non"))
	   << ", nss = " << nss << ", nds = " << nds << endl;
  }
  
  pg->flg_status[WALK_REQUEST] = true;
  wait_request = true;
  sem_wait(&wait_sem);
}

void mwalk::m_set_target_pos_nowait(istringstream& strm)
{
  double tss = (double)Nss*DT;
  double tds = (double)Nds*DT;
  double dx_cmd = 0.0, dy_cmd = 0.0, dyaw_cmd = 0.0;
  strm >> dx_cmd >> dy_cmd >> dyaw_cmd >> tss >> tds;
  
  int nss = (int)round(tss/DT);
  int nds = (int)round(tds/DT);
  fp->setTargetPos(pg->FootQueue, dx_cmd, dy_cmd, ToRad(dyaw_cmd), nss, nds);
  
  // last step
  FootPlace FP;
  FP.p = 0.0;
  FP.rpy = 0.0;
  FP.nss = nss;
  FP.nds = nds;
  FP.support_leg = pg->FootQueue.back().support_leg == RLEG ? LLEG : RLEG;
  pg->FootQueue.push_back(FP);  
  
  pg->flg_status[WALK_REQUEST] = true;
  wait_request = false;
  
  if( pg->verbose ){
    for( int i = 0 ; i < (int)pg->FootQueue.size() ; i++ )
      cout << "pg->FootQueue(x,y,th) = " << pg->FootQueue[i].p(X) << "," << pg->FootQueue[i].p(Y) << "," << ToDeg(pg->FootQueue[i].rpy(Z))
	   << ", sup = " << ((pg->FootQueue[i].support_leg==RLEG) ? "RLEG" : ((pg->FootQueue[i].support_leg==LLEG) ? "LLEG" : "Non"))
	   << ", nss = " << nss << ", nds = " << nds << endl;
  }
}

void mwalk::m_set_target_pos_passgate(istringstream& strm)
{
  double tss = (double)Nss*DT;
  double tds = (double)Nds*DT;
  vector3 gate_pos1(0.0), gate_pos2(0.0), target_pos(0.0);
  strm >> gate_pos1(X) >> gate_pos1(Y) >> gate_pos1(Z)
       >> gate_pos2(X) >> gate_pos2(Y) >> gate_pos2(Z)
       >> target_pos(X) >> target_pos(Y) >> target_pos(Z)
       >> tss >> tds;
  
  target_pos(Z)= ToRad(target_pos(Z));  
  int nss = (int)round(tss/DT);
  int nds = (int)round(tds/DT);
  
  fp->setTargetPosPassGate(pg->FootQueue, gate_pos1, gate_pos2, target_pos, nss, nds);
  
  // last step
  FootPlace FP;
  FP.p = 0.0;
  FP.rpy = 0.0;
  FP.nss = nss;
  FP.nds = nds;
  FP.support_leg = pg->FootQueue.back().support_leg == RLEG ? LLEG : RLEG;
  pg->FootQueue.push_back(FP);  
  
  if( pg->verbose ){
    for( int i = 0 ; i < (int)pg->FootQueue.size() ; i++ ) 
      cout << "pg->FootQueue(x,y,th) = " << pg->FootQueue[i].p(X) << "," << pg->FootQueue[i].p(Y) << "," << ToDeg(pg->FootQueue[i].rpy(Z))
	   << ", sup = " << ((pg->FootQueue[i].support_leg==RLEG) ? "RLEG" : ((pg->FootQueue[i].support_leg==LLEG) ? "LLEG" : "Non"))
	   << ", nss = " << nss << ", nds = " << nds << endl;
  }
  
  pg->flg_status[WALK_REQUEST] = true;
  wait_request = true;
  sem_wait(&wait_sem);
}


// usage :set-arc x y th
// rotate center(x[m],y[m]) rotate angle th[deg]
void mwalk::m_set_arc(istringstream& strm)
{
  double tss = (double)Nss*DT;
  double tds = (double)Nds*DT;
  double cx_cmd = 0.0, cy_cmd = 0.0,  dyaw_cmd = 0.0;
  strm >> cx_cmd >> cy_cmd >> dyaw_cmd >> tss >> tds;

  int nss = (int)round(tss/DT);
  int nds = (int)round(tds/DT);
  fp->setArcWalk(pg->FootQueue, cx_cmd, cy_cmd, ToRad(dyaw_cmd), nss, nds);
  
  // last step
  FootPlace FP;
  FP.p = 0.0;
  FP.rpy = 0.0;
  FP.nss = nss;
  FP.nds = nds;
  FP.support_leg = pg->FootQueue.back().support_leg == RLEG ? LLEG : RLEG;
  pg->FootQueue.push_back(FP);  
  
  if( pg->verbose ){
    for( int i = 0 ; i < (int)pg->FootQueue.size() ; i++ )
      cout << "pg->FootQueue(x,y,th) = " << pg->FootQueue[i].p(X) << "," << pg->FootQueue[i].p(Y) << "," << ToDeg(pg->FootQueue[i].rpy(Z))
	   << ", sup = " << ((pg->FootQueue[i].support_leg==RLEG) ? "RLEG" : ((pg->FootQueue[i].support_leg==LLEG) ? "LLEG" : "Non"))
	   << ", nss = " << nss << ", nds = " << nds << endl;
  }
  
  pg->flg_status[WALK_REQUEST] = true;
  wait_request = true;
  sem_wait(&wait_sem);
}

void mwalk::m_set_arc_nowait(istringstream& strm)
{
  double tss = (double)Nss*DT;
  double tds = (double)Nds*DT;
  double cx_cmd = 0.0, cy_cmd = 0.0,  dyaw_cmd = 0.0;
  strm >> cx_cmd >> cy_cmd >> dyaw_cmd >> tss >> tds;
  
  int nss = (int)round(tss/DT);
  int nds = (int)round(tds/DT);
  fp->setArcWalk(pg->FootQueue, cx_cmd, cy_cmd, ToRad(dyaw_cmd), nss, nds);
  
  // last step
  FootPlace FP;
  FP.p = 0.0;
  FP.rpy = 0.0;
  FP.nss = nss;
  FP.nds = nds;
  FP.support_leg = pg->FootQueue.back().support_leg == RLEG ? LLEG : RLEG;
  pg->FootQueue.push_back(FP);  
  
  pg->flg_status[WALK_REQUEST] = true;
  wait_request = false;
  
  if( pg->verbose ){
    for( int i = 0 ; i < (int)pg->FootQueue.size() ; i++ )
      cout << "pg->FootQueue(x,y,th) = " << pg->FootQueue[i].p(X) << "," << pg->FootQueue[i].p(Y) << "," << ToDeg(pg->FootQueue[i].rpy(Z))
	   << ", sup = " << ((pg->FootQueue[i].support_leg==RLEG) ? "RLEG" : ((pg->FootQueue[i].support_leg==LLEG) ? "LLEG" : "Non"))
	   << ", nss = " << nss << ", nds = " << nds << endl;
  }
}

// :push-walk-command ODV 1 0 0   (x[m],y[m],theta[deg]) //
// :push-walk-command ARC 0 1 30  (x[m],y[m],theta[deg]) //
void mwalk::m_push_walk_command(istringstream& strm)
{
  string walk_command = "no-walk";
  double tss = (double)Nss*DT;
  double tds = (double)Nds*DT;
  double x = 0.0, y = 0.0, th = 0.0;
  strm >> walk_command >> x >> y >> th >> tss >> tds;

  int nss = (int)round(tss/DT);
  int nds = (int)round(tds/DT);  
  if( walk_command == "ODV" )
    fp->setTargetPos(pg->FootQueue, x, y, ToRad(th), nss, nds);
  else if( walk_command == "ARC" )
    fp->setArcWalk(pg->FootQueue, x, y, ToRad(th), nss, nds);
  else if( walk_command == "RFOOT" )
    pg->pushFootPlace(pg->FootQueue, RLEG, x, y, 0.0, ToRad(th), 0.0, 0.0, 0.0, 0.0, (double)nss*DT, (double)nds*DT);  
  else if( walk_command == "LFOOT" )
    pg->pushFootPlace(pg->FootQueue, LLEG, x, y, 0.0, ToRad(th), 0.0, 0.0, 0.0, 0.0, (double)nss*DT, (double)nds*DT);  
  else if( walk_command == "EXEC_NOWAIT" || walk_command == "EXEC" ){
    if( !pg->FootQueue.empty() )
      pg->pushFootPlace(pg->FootQueue, -1, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, (double)pg->FootQueue.back().nss*DT, (double)pg->FootQueue.back().nds*DT);  
    else
      pg->pushFootPlace(pg->FootQueue, -1, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, (double)nss*DT, (double)nds*DT);
    
    pg->FootQueueNext.clear();
    pg->flg_status[WALK_REQUEST] = true;
    
    if( walk_command == "EXEC_NOWAIT" )
      wait_request = false;
    else{
      wait_request = true;
      sem_wait(&wait_sem);
    }
  }
}

void mwalk::m_set_foot_pos(istringstream& strm)
{
  string swingleg_name = "NONE";
  double x = 0.0, y = 0.0, z = 0.0, th = 0.0, hip_z = 0.0, roll = 0.0, pitch = 0.0, yaw = 0.0, tss = 0.0, tds = 0.0;
  strm >> x >> y >> z >> th >> hip_z >> roll >> pitch >> yaw >> tss >> tds >> swingleg_name;
  
  int support_leg = -1;
  if( swingleg_name == "RLEG" )
    support_leg = LLEG;
  else if( swingleg_name == "LLEG" )
    support_leg = RLEG;
  
  bool isModify = pg->pushFootPlace(pg->FootQueue,support_leg, x, y, z, ToRad(th), hip_z, ToRad(roll), ToRad(pitch), ToRad(yaw), tss, tds);
  
  if( pg->verbose ){
    if( isModify)
      cout << "!!!can not allocate foot place" << endl;
    
    cout << "pg->FootQueue(x,y,z,th) = "
	 << pg->FootQueue.back().p(X) << ","
	 << pg->FootQueue.back().p(Y) << ","
	 << pg->FootQueue.back().p(Z) << ","
	 << ToDeg(pg->FootQueue.back().rpy(2)) << endl;
    
    if( fabs(hip_z) > 0.0 || fabs(pitch) > 0.0 || fabs(yaw) > 0.0 || fabs(hip_z) > 0.0 )
      cout << "hip_z = " << hip_z << ", Ankle(roll,pitch,yaw) = " << roll << "," << pitch << "," << yaw << endl;
    
    if( tss > 0.0 && tds > 0.0)
      cout << "Tss = " << tss << "[s]" << ",Tds = " << tds << "[s]" << endl;
    else if( tss > 0.0)
      cout << "Tss = " << tss << "[s]" << endl;
    else if( tds > 0.0)
      cout << "Tds = " << tds << "[s]" << endl;
  }
}

void mwalk::m_set_next_foot_pos(istringstream& strm)
{
  double x = 0.0, y = 0.0, z = 0.0, th = 0.0, hip_z = 0.0, tss = 0.0, tds = 0.0, roll = 0.0, pitch = 0.0, yaw = 0.0;
  strm >> x >> y >> z >> th >> hip_z >> roll >> pitch >> yaw >> tss >> tds;
  
  pg->FootQueueNext.front();
  pg->pushFootPlace(pg->FootQueueNext,-1,x,y,z,ToRad(th),hip_z,ToRad(roll),ToRad(pitch),ToRad(yaw),tss,tds);
  
  if( pg->verbose ){
    cout << "pg->FootQueueNext(x,y,z,th) = " << x << "," << y << "," << z << "," << th;
    if( fabs(hip_z) > 0.0 || fabs(roll) > 0.0 || fabs(pitch) > 0.0 || fabs(yaw) > 0.0 )
      cout << "hip_z = " << hip_z << ", Ankle(roll,pitch,yaw) = " << roll << "," << pitch << "," << yaw << endl;
    
    cout << "Tss = " << tss << "[s]" << ", Tds = " << tds << "[s]" << endl;
  }
}

void mwalk::m_load_footpos(istringstream& strm)
{
  string filename;
  strm >> filename;
  ifstream istrm(filename.c_str());
  if (!istrm.is_open()) {
    cerr << "file not found (" << filename << ")" << endl;
    return;
  }
  else{
    cout << "load-footpos (" << filename << ")" << endl;
    while( !istrm.eof() ){
      int n = 0;
      double data[10]={0.0};
      while( read_text_parameter::GetData(istrm, data[n], 1) && n < 10) n++;
      
      if( n > 3 ){
        pg->pushFootPlace(pg->FootQueue,pg->first_step,data[0],data[1],data[2],ToRad(data[3]),data[4],
			  ToRad(data[5]),ToRad(data[6]),ToRad(data[7]),data[8],data[9]);
        
        if( pg->verbose ){
          cout << "pg->FootQueue(x,y,z,th) = " << data[0] << "," << data[1] << "," << data[2] << "," << data[3] << endl;

          if( fabs(data[4]) > 0.0 || fabs(data[5]) > 0.0 || fabs(data[6]) > 0.0 || fabs(data[7]) > 0.0 )
            cout << "hip_z = " << data[4] << ", Ankle(roll,pitch,yaw) = " << data[5] << "," << data[6] << "," << data[7] << endl;

          if(data[8] > 0.0 && data[9] > 0.0)
            cout << "Tss = " << data[8] << "[s]" << ", Tds = " << data[9] << "[s]" << endl;
          else if(data[8] > 0.0)
            cout << "Tss = " << data[8] << "[s]" << endl;
          else if(data[9] > 0.0)
            cout << "Tds = " << data[9] << "[s]" << endl;
        }
      }
      else{
	cout << "Too a few parameters : [x,y,z,th,hip_z,Tss,Tds,ankle_roll,ankle_pitch,ankle_yaw]" << endl;
      }
    }
    istrm.close();
  }
}

void mwalk::m_set_zmp_ankle(istringstream& strm)
{
  vector3 zmp_ankle_(zmp_ankle);
  strm >> zmp_ankle_(X) >> zmp_ankle_(Y) >> zmp_ankle_(Z);
  pg->ANcog.setZMPankle(zmp_ankle_);
  
  cout << "ZMP pos. at ankle = (" << zmp_ankle_(X) << "," << zmp_ankle_(Y) << "," << zmp_ankle_(Z) << ")" << endl;
}

void mwalk::m_set_zmp_path(istringstream& strm)
{
  vector3 zmp_ankle_ini_(zmp_ankle);
  vector3 zmp_ankle_end_(zmp_ankle);
  strm >> zmp_ankle_ini_(X) >> zmp_ankle_ini_(Y) >> zmp_ankle_ini_(Z) >> zmp_ankle_end_(X) >> zmp_ankle_end_(Y) >> zmp_ankle_end_(Z);
  
  pg->ANcog.setZMPankle(zmp_ankle_ini_, zmp_ankle_end_);
  
  cout << "ZMP moves from = ("
       << zmp_ankle_ini_(X) << "," << zmp_ankle_ini_(Y) << "," << zmp_ankle_ini_(Z) << ") to ("
       << zmp_ankle_end_(X) << "," << zmp_ankle_end_(Y) << "," << zmp_ankle_end_(Z) << ")" << endl;
}

#ifdef WITH_TOE_JOINT
void mwalk::m_set_zmp_toe(istringstream& strm)
{
  vector3 zmp_toe_(zmp_toe);
  strm >> zmp_toe_(X) >> zmp_toe_(Y) >> zmp_toe_(Z);
  
  pg->ANcog.setZMPtoe(zmp_toe_);

  cout << "ZMP pos. at toe = (" << zmp_toe_(X) << "," << zmp_toe_(Y) << "," << zmp_toe_(Z) << ")" << endl;
}
#endif

void mwalk::m_set_ankle_ds_rotation_ratio(istringstream& strm)
{
  strm >> pg->TR_ankle_ds_ratio[0] >> pg->TR_ankle_ds_ratio[1];
  
  if( pg->TR_ankle_ds_ratio[0] == 0.0 && pg->TR_ankle_ds_ratio[1] == 0.0 ){
    pg->flg_status[ANKLE_DS_ROTATION] = false;
    cout << "Prohibit ankle rotation in half of double support phase" << endl;
  }
  else{
    pg->flg_status[ANKLE_DS_ROTATION] = true;
    cout << "Permit ankle rotation in half of double support phase" << endl;
    
    cout << "TR_ankle_ds_ratio = "
	 << pg->TR_ankle_ds_ratio[0] << "," << pg->TR_ankle_ds_ratio[1] << endl;
  }
}

void mwalk::m_trunk_roll_mode(istringstream& strm)
{
  double fixed_height = 1.0;
  string mode;
  strm >> mode >> fixed_height;
  
  if( mode == "on" || mode == "true" || mode == "1" ){
    pg->flg_status[TRUNK_ROLL] = true;
    pg->Trunk_roll_fixed_height = fixed_height;
    cout << "Trunk roll motion on " << fixed_height << endl;
  }
  else{
    pg->flg_status[TRUNK_ROLL] = false;
    cout << "Trunk roll motion off" << endl;
  }
}

void mwalk::m_time_adj(istringstream& strm)
{
  bool flg = false;
  strm >> flg;
  
  pg->flg_status[TIME_ADJUSTMENT] = flg;
  if( pg->flg_status[TIME_ADJUSTMENT] )
    cout << "time adjustment mode" << endl;
}

void mwalk::m_zmpshaping(istringstream& strm)
{
  double tprev = (double)(Nss+Nds)*DT;
  strm >> tprev;
  
  pg->flg_status[DFILTER_PREVIEW] = false;
  
  const double cog_height = pg->COG_cmd(Z);
  
  if( tprev > DT && pg->mmode == INITCONF && pg->Pcog[X].calcPreviewGains(cog_height, (int)round(tprev/DT)) ){
    pg->Pcog[Y].calcPreviewGains(cog_height, (int)round(tprev/DT));
    pg->flg_status[DFILTER_PREVIEW] = true;
    
    cout << "Preview gains for zmp shaping were calculated." << endl;
  }
  else
    cout << "can't calculate Preview gains for zmp shaping" << endl;
}

void mwalk::m_dfilter_preview(istringstream& strm)
{
  double tprev = (double)(Nss+Nds)*DT*2.0;
  string mode;
  strm >> mode >> tprev;
  
  pg->flg_status[DFILTER_PREVIEW] = false;
  if( mode == "on" ){
    const double cog_height = pg->COG_cmd(Z) - pg->Zmp_cmd(Z);
    
    if( tprev > DT && pg->mmode == INITCONF && pg->Pcog_dzmp[X].calcPreviewGains(cog_height, (int)round(tprev/DT)) ){
      pg->Pcog_dzmp[Y].calcPreviewGains(cog_height, (int)round(tprev/DT));
      pg->flg_status[DFILTER_PREVIEW] = true;
      
      cout << "Preview gains for dynamics filter were calculated." << endl;
    }
    else
      cout << "can't calculate Preview gains for dynamics filter" << endl;
  }
  else{
    cout << "dfilter-preview off" << endl;
  }
}

void mwalk::m_init_cog_offset_gain(istringstream& strm)
{
  strm >> pg->INIT_COG_OFFSET_GAIN;
  
  if( pg->INIT_COG_OFFSET_GAIN < 0.0 )
    pg->INIT_COG_OFFSET_GAIN = 0.0;
  else if( pg->INIT_COG_OFFSET_GAIN > 1.0/(2.0*DT) )
    pg->INIT_COG_OFFSET_GAIN = 1.0/(2.0*DT);
  
  cout << "set init cog offset gain (0 - " << 1.0/(2.0*DT) << ") = " << pg->INIT_COG_OFFSET_GAIN << endl;
}

void mwalk::m_update_cog_offset_gain(istringstream& strm)
{
  strm >> pg->UPDATE_COG_OFFSET_GAIN;
  
  if( pg->UPDATE_COG_OFFSET_GAIN < 0.0 )
    pg->UPDATE_COG_OFFSET_GAIN = 0.0;
  else if( pg->UPDATE_COG_OFFSET_GAIN > 1.0/(2.0*DT) )
    pg->UPDATE_COG_OFFSET_GAIN = 1.0/(2.0*DT);
  
  cout << "set update cog offset gain (0 - " << 1.0/(2.0*DT) << ") = " << pg->UPDATE_COG_OFFSET_GAIN << endl;
}

void mwalk::m_load_zmpshaping_pgain(istringstream& strm)
{
  string filename;
  strm >> filename;
  
  if( pg->Pcog[X].loadGains(filename.c_str(), pg->verbose) ){
    
    pg->flg_status[ZMP_SHAPING_PREVIEW] = true;
    
    pg->Pcog[Y].loadGains(filename.c_str(), pg->verbose);
    
    pg->zmp_err[X].resize(pg->Pcog[X].Nprev*2);
    pg->zmp_err[Y].resize(pg->Pcog[Y].Nprev*2);
    
    noalias(pg->zmp_err[X]) = dzerovector(pg->zmp_err[X].size());
    noalias(pg->zmp_err[Y]) = dzerovector(pg->zmp_err[Y].size());
    
    cout << "Use ZMP reference shaping by preview control" << endl;
  }
  else{
    pg->flg_status[ZMP_SHAPING_PREVIEW] = false;
    cout << "Can not use ZMP shaping" << endl;
  }
}

// Dynamic filter by preview control
void mwalk::m_load_dfilter_pgain(istringstream& strm)
{
  string filename;
  strm >> filename;
  
  if( pg->mmode == INITCONF && pg->Pcog_dzmp[X].loadGains(filename.c_str(), pg->verbose) ){
    
    pg->flg_status[DFILTER_PREVIEW] = true;
    pg->flg_status[ZMP_SHAPING_PREVIEW] = false;
    pg->flg_status[TIME_ADJUSTMENT] = false;
    
    pg->Pcog_dzmp[Y].loadGains(filename.c_str(), pg->verbose);
    
    cout << "Use dynamic filter by preview control" << endl;
  }
  else{
    pg->flg_status[DFILTER_PREVIEW] = false;
    cout << "Can not use preview filter" << endl;
  }
}

void mwalk::m_reactive_motion(istringstream& strm)
{
  string mode;
  strm >> mode;
  
  if( mode == "on" || mode == "true" || mode == "1" ){
    pg->flg_status[REACTIVE_MOTION] = true;
    cout << "reactive motion on" << endl;
  }
  else{
    pg->flg_status[REACTIVE_MOTION] = false;
    cout << "reactive motion off" << endl;
  } 
}

void mwalk::m_state_logger_mode(istringstream& strm)
{
  string mode;
  strm >> mode;
  
  state_logger.clearLogData();
  if( mode == "on" || mode == "true" || mode == "1" ){
    state_logger_mode = true;
    cout << "State logger on" << endl;
  }
  else{
    state_logger_mode = false;
    cout << "State logger off" << endl;
  }
}

void mwalk::m_save_state_logger(istringstream& strm)
{
  state_logger_mode = false;
  
  string filename;
  strm >> filename;
  
  state_logger.saveLogData(filename);
  cout << "pattern files were created as " << filename << "(.pos, .hip, .zmp, .gsens, .vel, .waist, .zmpref, .pivot, .ankle, .foot), size=" << endl;
}

void mwalk::m_stat(istringstream& strm)
{
#ifdef CHECK_EXEC_TIME
  cout << "execT_max(" << mExecTime.execT_max.size()-1 << ") = ";
  for( unsigned int i = 0 ; i < mExecTime.execT_max.size() - 1 ; i++ )
    cout << mExecTime.execT_max[i] << " ";
  cout << endl;
  cout << "Maximum calculation time = " << mExecTime.execT_max.back() << endl;
#endif
  
  cout << "\nmawlk stat" << endl;
  cout << "\ntime " << (double)pg->n_clock*DT << "[s], walking state = " << getWalkingStateString(pg->ws) << endl;
  cout << "\nlatest adjusting time = " << (double)pg->n_comp * DT << endl;
  cout << "COG pos= " << pg->total_com << endl;
}

void mwalk::m_check_joints_status(istringstream& strm)
{
  bool flg = false;
  strm >> flg;
  
  pg->flg_status[CHECK_JOINT_STATUS] = flg;
  if( pg->flg_status[CHECK_JOINT_STATUS] )
    cout << "check joints status" << endl;
}

void mwalk::m_joints_status(istringstream& strm)
{
  pg->showJointLimits();
}

void mwalk::m_debug_level(istringstream& strm)
{
  string debug_level;
  strm >> debug_level;
  
  if( debug_level == "on" || debug_level == "true" || debug_level == "1" ){
    pg->verbose = true;
    cout << "debug mode on" << endl;
  }
}

void mwalk::m_debug_info(istringstream& strm)
{
  
  for( int i = 0 ; i < Ncog ; i++ )
    cout << "cog_height[" << i << "] = " << pg->cog_height[i] << endl;
  
  cout << "Waist_cmd pos = " << pg->Waist_cmd.P << endl;
  
  cout << "Xg_offset_ini = " << pg->Xg_offset_ini << endl;
  
  for( int i = 0 ; i <= Ncog ; i++ )
    cout << "zmp_des[" << i   << "] = " << pg->zmp_des[i] << endl;
  
  int FQsize = pg->FootQueue.size();
  for( int i = 0 ; i < FQsize ; i++ )
    cout << "pg->FootQueue[" << i << "] = " << pg->FootQueue[i].hip_z << endl;
}

#ifdef WITH_TOE_JOINT

void mwalk::m_set_toe_limit_pitch(istringstream& strm)
{
  double toe_pitch_max_   = pg->TOE_PITCH_MAX;
  double toe_pitch_min_   = pg->TOE_PITCH_MIN;
  double toe_pitch_ratio_ = pg->TOE_PITCH_RATIO;
  
  strm >> toe_pitch_max_ >> toe_pitch_min_ >> toe_pitch_ratio_;
  
  pg->TOE_PITCH_MAX   = ToRad(toe_pitch_max_);
  pg->TOE_PITCH_MIN   = ToRad(toe_pitch_min_);
  pg->TOE_PITCH_RATIO = toe_pitch_ratio_;
  
  cout << "step-toe-limit-pitch(Max,Min,ratio) = (" << toe_pitch_max_ << "," << toe_pitch_min_ << "," << toe_pitch_ratio_ << ")" << endl;
}

#endif

mwalk_state::mwalk_state(mwalk *mw)
{
  n_clock = mw->pg->n_clock;
  ws      = mw->pg->ws;
  ws_pre  = mw->pg->ws_pre;

  Zmp_cmd    = mw->pg->Zmp_cmd; // + mg->Waist_cmd.P;
  Zmp_shp    = mw->pg->Zmp_shp;
  Zmp_ref    = mw->pg->Zmp_ref;
  Zmp_model  = mw->pg->Zmp_model;
  Zmp_ideal  = mw->pg->Zmp_ideal;
  Zmp_model2 = mw->pg->Zmp_model2;
  VZmp_cmd   = mw->pg->VZmp_cmd; // + mg->Waist_cmd.V;;
  Gsensor_offset = mw->pg->Gsensor_offset;
  
  Xg_cmd     = mw->pg->COG_cmd - mw->pg->COG_shp;
  Xg_shp     = mw->pg->COG_shp;
  Xg_dfilter = mw->pg->COG_dfilter;
  Vg_cmd     = mw->pg->VCOG_cmd - mw->pg->VCOG_shp;
  Vgdot_cmd  = mw->pg->VCOGdot_cmd;
  
  for( int i = RLEG ; i <= LLEG ; i++ ){
    Xf_cmd[i] = mw->pg->Foot_cmd[i].P;
    Vf_cmd[i] = mw->pg->Foot_cmd[i].V;
    Xa_cmd[i] = mw->pg->Ankle_cmd[i].P;
    Va_cmd[i] = mw->pg->Ankle_cmd[i].V;
    Ra_cmd[i] = mw->pg->Ankle_cmd[i].R_rpy;
    Wa_cmd[i] = mw->pg->Ankle_cmd[i].W;
    Pivot_cmd[i] = mw->pg->Pivot_cmd[i];
  }
  
  Xwaist_cmd = mw->pg->Waist_cmd.P;
  Vwaist_cmd = mw->pg->Waist_cmd.V;
  Rwaist_cmd = mw->pg->Waist_cmd.R_rpy;
  Wwaist_cmd = mw->pg->Waist_cmd.W;
  
  Zmp_res = mw->pg->Zmp_res;
  COG_cmp = mw->pg->COG_cmp;
  VCOG_cmp = mw->pg->VCOG_cmp;
  
  VCOGdot_res = mw->pg->VCOGdot_res;
  
  P_cmd(0) = mw->pg->PL_cmd(0);
  P_cmd(1) = mw->pg->PL_cmd(1);
  P_cmd(2) = mw->pg->PL_cmd(2);
  L_cmd(0) = mw->pg->PL_cmd(3);
  L_cmd(1) = mw->pg->PL_cmd(4);
  L_cmd(2) = mw->pg->PL_cmd(5);
  P_model = mw->pg->lin_mom_pre;
  L_model = mw->pg->ang_mom_pre;
  
  if( !mw->pg->flg_status[DFILTER_PREVIEW] ){
    q_cmd    = mw->pg->q_cmd;
    qdot_cmd = mw->pg->qdot_cmd;
  }
  else{
    q_cmd    = mw->pg->q_cmd2;
    qdot_cmd = mw->pg->qdot_cmd2;
  }
  
#if 0  
  const double  foot_size[2][2] = FOOT_SIZE;
  vector3 Pvertex_local[4][2];
  Pvertex_local[0][RLEG] = -foot_size[FB][HEEL],  foot_size[RL][INSIDE],  -ANKLE_HEIGHT;
  Pvertex_local[1][RLEG] = -foot_size[FB][HEEL], -foot_size[RL][OUTSIDE], -ANKLE_HEIGHT;
  Pvertex_local[2][RLEG] =  foot_size[FB][TOE],  -foot_size[RL][OUTSIDE], -ANKLE_HEIGHT;
  Pvertex_local[3][RLEG] =  foot_size[FB][TOE],   foot_size[RL][INSIDE],  -ANKLE_HEIGHT;
  
  Pvertex_local[0][LLEG] = -foot_size[FB][HEEL], -foot_size[RL][INSIDE],  -ANKLE_HEIGHT;
  Pvertex_local[1][LLEG] = -foot_size[FB][HEEL],  foot_size[RL][OUTSIDE], -ANKLE_HEIGHT;
  Pvertex_local[2][LLEG] =  foot_size[FB][TOE],   foot_size[RL][OUTSIDE], -ANKLE_HEIGHT;
  Pvertex_local[3][LLEG] =  foot_size[FB][TOE],  -foot_size[RL][INSIDE],  -ANKLE_HEIGHT;
  
  for( int leg = RLEG ; leg <= LLEG ; leg++ ){
    for( int i = 0 ; i < 4 ; i++ ){
      Pvertex[i][leg] = mw->pg->Ankle_cmd[leg].P + mw->pg->Ankle_cmd[leg].R * Pvertex_local[i][leg];
      Vvertex[i][leg] = mw->pg->Ankle_cmd[leg].V + mw->pg->Ankle_cmd[leg].W * Pvertex_local[i][leg];
    }
  }
#endif

}

mwalk_state::~mwalk_state()
{
}

const char *mwalk_state::header()
{
  return "%time ws ws_pre Zmp_cmd(X) Zmp_cmd(Y) Zmp_cmd(Z) VZmp_cmd(X) VZmp_cmd(Y) VZmp_cmd(Z) Xg_cmd(X) Xg_cmd(Y) Xg_cmd(Z) Vg_cmd(X) Vg_cmd(Y) Vg_cmd(Z) Vgdot_cmd(X) Vgdot_cmd(Y) Vgdot_cmd(Z)  Xf_cmd[RLEG](X) Xf_cmd[RLEG](Y) Xf_cmd[RLEG](Z) Vf_cmd[RLEG](X) Vf_cmd[RLEG](Y) Vf_cmd[RLEG](Z) Xf_cmd[LLEG](X) Xf_cmd[LLEG](Y) Xf_cmd[LLEG](Z) Vf_cmd[LLEG](X) Vf_cmd[LLEG](Y) Vf_cmd[LLEG](Z) Xa_cmd[RLEG](X) Xa_cmd[RLEG](Y) Xa_cmd[RLEG](Z) Xa_cmd[LLEG](X) Xa_cmd[LLEG](Y) Xa_cmd[LLEG](Z) R_cmd[RLEG](X) R_cmd[RLEG](Y) R_cmd[RLEG](Z) R_cmd[LLEG](X) R_cmd[LLEG](Y) R_cmd[LLEG](Z) Xwaist_cmd(X) Xwaist_cmd(Y) Xwaist_cmd(Z) Vwaist_cmd(X) Vwaist_cmd(Y) Vwaist_cmd(Z) Rwaist_cmd(X) Rwaist_cmd(Y) Rwaist_cmd(Z) Wwaist_cmd(X) Wwaist_cmd(Y) Wwaist_cmd(Z) Zmp_model(X) Zmp_model(Y) Zmp_model(Z) Zmp_ideal(X) Zmp_ideal(Y) Zmp_ideal(Z) Zmp_model2(X) Zmp_model2(Y) Zmp_model2(Z) Zmp_shp(X) Zmp_shp(Y) Xg_shp(X) Xg_shp(Y) Xg_dfilter(X) Xg_dfilter(Y) Zmp_ref(X) Zmp_ref(Y) Zmp_ref(Z) Gsensor_offset(X) Gsensor_offset(Y) Gsensor_offset(Z) q_cmd qdot_cmd";
}

void mwalk_state::pprint(ostream &s) const
{
  s << (double)n_clock*DT     << " " << ws                     << " " << ws_pre                 << " ";  //  1 -  3
  s << Zmp_cmd(X)             << " " << Zmp_cmd(Y)             << " " << Zmp_cmd(Z)             << " ";  //  4 -  6
  s << VZmp_cmd(X)            << " " << VZmp_cmd(Y)            << " " << VZmp_cmd(Z)            << " ";  //  7 -  9
  s << Xg_cmd(X)              << " " << Xg_cmd(Y)              << " " << Xg_cmd(Z)              << " ";  // 10 - 12
  s << Vg_cmd(X)              << " " << Vg_cmd(Y)              << " " << Vg_cmd(Z)              << " ";  // 13 - 15
  s << Vgdot_cmd(X)           << " " << Vgdot_cmd(Y)           << " " << Vgdot_cmd(Z)           << " ";  // 16 - 18
  s << Xf_cmd[RLEG](X)        << " " << Xf_cmd[RLEG](Y)        << " " << Xf_cmd[RLEG](Z)        << " ";  // 19 - 21
  s << Vf_cmd[RLEG](X)        << " " << Vf_cmd[RLEG](Y)        << " " << Vf_cmd[RLEG](Z)        << " ";  // 22 - 24
  s << Xf_cmd[LLEG](X)        << " " << Xf_cmd[LLEG](Y)        << " " << Xf_cmd[LLEG](Z)        << " ";  // 25 - 27
  s << Vf_cmd[LLEG](X)        << " " << Vf_cmd[LLEG](Y)        << " " << Vf_cmd[LLEG](Z)        << " ";  // 28 - 30
  s << Xa_cmd[RLEG](X)        << " " << Xa_cmd[RLEG](Y)        << " " << Xa_cmd[RLEG](Z)        << " ";  // 31 - 33
  s << Xa_cmd[LLEG](X)        << " " << Xa_cmd[LLEG](Y)        << " " << Xa_cmd[LLEG](Z)        << " ";  // 34 - 36
  s << Va_cmd[RLEG](X)        << " " << Va_cmd[RLEG](Y)        << " " << Va_cmd[RLEG](Z)        << " ";  // 37 - 39
  s << Va_cmd[LLEG](X)        << " " << Va_cmd[LLEG](Y)        << " " << Va_cmd[LLEG](Z)        << " ";  // 40 - 42
  s << Wa_cmd[RLEG](X)        << " " << Wa_cmd[RLEG](Y)        << " " << Wa_cmd[RLEG](Z)        << " ";  // 43 - 45
  s << Wa_cmd[LLEG](X)        << " " << Wa_cmd[LLEG](Y)        << " " << Wa_cmd[LLEG](Z)        << " ";  // 46 - 48
  s << ToDeg(Ra_cmd[RLEG](X)) << " " << ToDeg(Ra_cmd[RLEG](Y)) << " " << ToDeg(Ra_cmd[RLEG](Z)) << " ";  // 49 - 51
  s << ToDeg(Ra_cmd[LLEG](X)) << " " << ToDeg(Ra_cmd[LLEG](Y)) << " " << ToDeg(Ra_cmd[LLEG](Z)) << " ";  // 52 - 54
  s << Xwaist_cmd(X)          << " " << Xwaist_cmd(Y)          << " " << Xwaist_cmd(Z)          << " ";  // 55 - 57
  s << Vwaist_cmd(X)          << " " << Vwaist_cmd(Y)          << " " << Vwaist_cmd(Z)          << " ";  // 58 - 60
  s << ToDeg(Rwaist_cmd(X))   << " " << ToDeg(Rwaist_cmd(Y))   << " " << ToDeg(Rwaist_cmd(Z))   << " ";  // 61 - 63
  s << Wwaist_cmd(X)          << " " << Wwaist_cmd(Y)          << " " << Wwaist_cmd(Z)          << " ";  // 64 - 66
  s << Pivot_cmd[RLEG](X)     << " " << Pivot_cmd[RLEG](Y)     << " ";                                   // 67 - 68
  s << Pivot_cmd[LLEG](X)     << " " << Pivot_cmd[LLEG](Y)     << " ";                                   // 69 - 70
  s << Zmp_model(X)           << " " << Zmp_model(Y)           << " " << Zmp_model(Z)           << " ";  // 71 - 73
  s << Zmp_ideal(X)           << " " << Zmp_ideal(Y)           << " " << Zmp_ideal(Z)           << " ";  // 74 - 76
  s << Zmp_model2(X)          << " " << Zmp_model2(Y)          << " " << Zmp_model2(Z)          << " ";  // 77 - 79
  s << Zmp_shp(X)             << " " << Zmp_shp(Y)             << " ";                                   // 80 , 81
  s << Xg_shp(X)              << " " << Xg_shp(Y)              << " ";                                   // 82 , 83
  s << Xg_dfilter(X)          << " " << Xg_dfilter(Y)          << " ";                                   // 84 , 85
  s << Zmp_ref(X)             << " " << Zmp_ref(Y)             << " " << Zmp_ref(Z)             << " ";  // 86 - 88
  s << Gsensor_offset(X)      << " " << Gsensor_offset(Y)      << " " << Gsensor_offset(Z)      << " ";  // 89 - 91
  
#if 0
  s << P_cmd(X) << " " << P_cmd(Y) << " " << P_cmd(Z) << " ";
  s << L_cmd(X) << " " << L_cmd(Y) << " " << L_cmd(Z) << " ";
  s << P_model(X) << " " << P_model(Y) << " " << P_model(Z) << " ";
  s << L_model(X) << " " << L_model(Y) << " " << L_model(Z) << " ";
#endif
  
#if 0
  s << Pvertex[0][RLEG](X) << " " << Pvertex[0][RLEG](Y) << " " << Pvertex[0][RLEG](Z) << " ";  //  92 - 94
  s << Pvertex[1][RLEG](X) << " " << Pvertex[1][RLEG](Y) << " " << Pvertex[1][RLEG](Z) << " ";  //  95 - 97
  s << Pvertex[2][RLEG](X) << " " << Pvertex[2][RLEG](Y) << " " << Pvertex[2][RLEG](Z) << " ";  //  98 - 100
  s << Pvertex[3][RLEG](X) << " " << Pvertex[3][RLEG](Y) << " " << Pvertex[3][RLEG](Z) << " ";  // 101 - 103
  
  s << Pvertex[0][LLEG](X) << " " << Pvertex[0][LLEG](Y) << " " << Pvertex[0][LLEG](Z) << " ";  // 104 - 106
  s << Pvertex[1][LLEG](X) << " " << Pvertex[1][LLEG](Y) << " " << Pvertex[1][LLEG](Z) << " ";  // 107 - 109
  s << Pvertex[2][LLEG](X) << " " << Pvertex[2][LLEG](Y) << " " << Pvertex[2][LLEG](Z) << " ";  // 110 - 112
  s << Pvertex[3][LLEG](X) << " " << Pvertex[3][LLEG](Y) << " " << Pvertex[3][LLEG](Z) << " ";  // 113 - 115

  s << Vvertex[0][RLEG](X) << " " << Vvertex[0][RLEG](Y) << " " << Vvertex[0][RLEG](Z) << " ";  // 116 - 118
  s << Vvertex[1][RLEG](X) << " " << Vvertex[1][RLEG](Y) << " " << Vvertex[1][RLEG](Z) << " ";  // 119 - 121
  s << Vvertex[2][RLEG](X) << " " << Vvertex[2][RLEG](Y) << " " << Vvertex[2][RLEG](Z) << " ";  // 122 - 124
  s << Vvertex[3][RLEG](X) << " " << Vvertex[3][RLEG](Y) << " " << Vvertex[3][RLEG](Z) << " ";  // 125 - 127

  s << Vvertex[0][LLEG](X) << " " << Vvertex[0][LLEG](Y) << " " << Vvertex[0][LLEG](Z) << " ";  // 128 - 130
  s << Vvertex[1][LLEG](X) << " " << Vvertex[1][LLEG](Y) << " " << Vvertex[1][LLEG](Z) << " ";  // 131 - 133
  s << Vvertex[2][LLEG](X) << " " << Vvertex[2][LLEG](Y) << " " << Vvertex[2][LLEG](Z) << " ";  // 134 - 136
  s << Vvertex[3][LLEG](X) << " " << Vvertex[3][LLEG](Y) << " " << Vvertex[3][LLEG](Z) << " ";  // 137 - 139
#endif
  
  s << Zmp_res(X)      << " " << Zmp_res(Y)      << " " << Zmp_res(Z)      << " ";                    //  92 -  94 
  s << COG_cmp(X)      << " " << COG_cmp(Y)      << " " << COG_cmp(Z)      << " ";                    //  95 -  97 
  s << VCOG_cmp(X)     << " " << VCOG_cmp(Y)     << " " << VCOG_cmp(Z)     << " ";                    //  98 - 100
  
  s << VCOGdot_res(X)  << " " << VCOGdot_res(Y)  << " " << VCOGdot_res(Z)  << " ";                    // 101 - 103
  
  for( int i = 0 ; i < DOF ; i++ ) // 104 - 104+DOF-1
    s << q_cmd[i] << " ";
  
  for( int i = 0 ; i < DOF ; i++ ) // 104+DOF - 104+2*DOF - 1
    s << qdot_cmd[i] << " ";
  
  s << endl;
}

plugin_state *mwalk::get_state()
{
  return new mwalk_state(this);
}
