/*
 * Copyright (c) 2009,
 * @author Mitsuharu Morisawa
 *
 * AIST
 *
 * All rights reserved.
 *
 * This program is made available under the terms of the Eclipse Public License
 * v1.0 which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 */
#ifndef __MSWING_GENERATOR_H__
#define __MSWING_GENERATOR_H__

#include "mwalk_common.h"
#include "mwalkstate.h"
#include "State.h"
#include "MotionInterpolator.h"

using namespace motion_interpolator;


namespace mwalk_pattern_generator{

  typedef MotionInterpolator MotionInterpolator2[2];
  typedef MotionInterpolator MotionInterpolator3[3];
  
  enum TRAJECTORY_TYPE{
    FLAT = 0,
    UPSTAIRS,
    DOWNSTAIRS,
    SWING_TYPE_NUM
  };
  
  class FootMotionGenerator{
  private:
    MotionInterpolator3 SwingLegPos;
    MotionInterpolator3 AnkleRpy[2];
    MotionInterpolator2 SwingRpy[2];
    MotionInterpolator2 Pivot[2];
#ifdef WITH_TOE_JOINT
    MotionInterpolator  ToePitch[2];
#endif
    
  protected:
    int  selectFloor(double step);
    void setPivotParam(MotionInterpolator2& SupLegPivot_, MotionInterpolator2& SwgLegPivot_, vector3& step, WalkingState& ws, int n_start, int n_period);
    void setSwingRpyParam(MotionInterpolator2& SwingRpyDat, State& Leg_ini, State& Leg_end, unsigned int FQsize, int n_start, int n_period);
    void setSwingRpyParam(MotionInterpolator2& SwingRpyDat, vector3& step, double step_yaw, unsigned int FQsize, int n_start, int n_period);
    void setAnkleParam(MotionInterpolator3& AnkleRpyDat, State& Leg_ini, State& Leg_end, int n_start, int n_period);

#ifndef WITH_TOE_JOINT
    /*! @brief Generate sole(foot) and ankle position and rotation (for swing leg)
     * @param[in]  PivotDat    pivot position interpolator
     * @param[in]  SwingRpyDat pitch and roll angles interpolator at touchdown and liftoff
     * @param[in]  foot_       sole states(position, velocity and acceleration of translation and rotation)
     * @param[out] ankle_      ankle states(position, velocity and acceleration of translation and rotation)
     * @param[out] pivot_      pivot position
     * @param[in]  n_clock     current time
     */
    void getAnkleMotion(MotionInterpolator2& PivotDat, MotionInterpolator2& SwingRpyDat, State& foot_, State& ankle_, vector3& pivot_, int n_clock);
#else
    void setToeParam(MotionInterpolator& SupToe, MotionInterpolator& SwgToe, unsigned int FQsize, int n_start, int n_period);
    void getAnkleMotion(MotionInterpolator2& PivotDat, MotionInterpolator2& SwingRpyDat, MotionInterpolator& ToePitch, State& Dat_pivot, State& Dat, double& toe_pitch_cmd,  vector3& Pivot_cmd, int n_clock);
#endif
    
  public:
    double swg_height_threshold;
    double swg_rpy_threshold;
    
#ifdef WITH_TOE_JOINT
    double TOE_PITCH_MAX;
    double TOE_PITCH_MIN;
    double TOE_PITCH_RATIO;
#endif
    
    double Tprofile_horizontal[SWING_TYPE_NUM][4];       // (initial, acceleration, deceleration, end)
    double Tprofile_vertical[SWING_TYPE_NUM][8];         // (initial, acceleration, deceleration, end)x(landoff,touchdown)
    double TP_pivot[SWING_TYPE_NUM][2];                  // (initial  end)
    double TR_ankle[SWING_TYPE_NUM][4];                  // (initial, landoff, touchdown, end)
    double XPmiddle_ratio[SWING_TYPE_NUM];
    double SWmiddle_point[SWING_TYPE_NUM];
    double THmiddle_ratio[SWING_TYPE_NUM];
    double XHmiddle_ratio[SWING_TYPE_NUM];
    
    double  R_blandoff[SWING_TYPE_NUM];
    double  R_btouchdown[SWING_TYPE_NUM];
    vector3 R_landoff[SWING_TYPE_NUM];
    vector3 R_touchdown[SWING_TYPE_NUM];
    
    FootMotionGenerator(void);
    ~FootMotionGenerator();
    
    void Reset(int n_clock, int n_period);
    void setSwingPosParam(State& Leg_ini, State& Leg_end, int n_start, int n_period, int n_total);
    void setSwingVerticalPosParam(State& Leg_ini, State& Leg_end, int n_start, int n_period, int n_total);
    void setSwingHorizontalPosParam(State& Leg_ini, State& Leg_end, int n_start, int n_total);
    inline void setPivotParam(vector3& step, int sup_leg, int swg_leg, WalkingState ws, int n_start, int n_period){
      setPivotParam(Pivot[sup_leg], Pivot[swg_leg], step, ws, n_start, n_period); 
    }
    inline void setSwingRpyParam(vector3& step, double& step_yaw, const int leg, unsigned int FQsize, int n_start, int n_period){
      setSwingRpyParam(SwingRpy[leg], step, step_yaw, FQsize, n_start, n_period);
    }
    inline void setAnkleParam(State *Foot_cmd, State *Foot_next_cmd, int n_start, int n_period){
      setAnkleParam(AnkleRpy[RLEG], Foot_cmd[RLEG], Foot_next_cmd[RLEG], n_start, n_period);
      setAnkleParam(AnkleRpy[LLEG], Foot_cmd[LLEG], Foot_next_cmd[LLEG], n_start, n_period);
    }
#ifndef WITH_TOE_JOINT
    /*! @brief Generate sole(foot) and ankle states and pivot position
     * @param[out] Foot_cmd  sole states(position, velocity and acceleration of translation and rotation)
     * @param[out] Ankle_cmd ankle states(position, velocity and acceleration of translation and rotation)
     * @param[out] Pivot_cmd pivot position
     * @param[in]  ws        walking state
     * @param[in]  n_clock   current time
     */
    void makeSwingLegMotion(State Foot_cmd[2], State Ankle_cmd[2], vector3 Pivot_cmd[2], WalkingState ws, int n_clock);
#else
    void makeSwingLegMotion(State *Foot_cmd, State *Ankle_cmd, double& toe_angle_right, double& toe_angle_left, vector3 *Pivot_cmd, WalkingState ws, int n_clock);
    inline void setToeParam(int sup_leg, int swg_leg, unsigned int FQsize, int n_start, int n_period){
      setToeParam(ToePitch[sup_leg], ToePitch[swg_leg], FQsize, n_start, n_period);
    }
#endif
  };
  
}  // end of namespace mwalk_pattern_generator


#endif // __MWALK_GENERATOR_H__
