/*
 * Copyright (c) 2009,
 * @author Mitsuharu Morisawa
 *
 * AIST
 *
 * All rights reserved.
 *
 * This program is made available under the terms of the Eclipse Public License
 * v1.0 which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 */
#include "LinearControl.h"

using namespace linear_control_system;
using namespace boost::numeric::ublas;


LinearControl_SISO::LinearControl_SISO(void)
{
  LinearControl_SISO::Initialize();
}

LinearControl_SISO::~LinearControl_SISO()
{
}

void LinearControl_SISO::Initialize(void)
{
  A.clear();
  B.clear();
  C.clear();
  K.clear();
  x.clear();
  y = 0.0;
  u = 0.0;
}

void LinearControl_SISO::update(void)
{
  x = ublas::prod(A, x) + ublas::matrix_column<dmatrix>(B, 0) * u;
  y = ublas::prod(C, x)(0);
}

void LinearControl_SISO::setExtendedSystem(dmatrix& Ax, dmatrix& Bx, dmatrix& Cx)
{
  const unsigned int size = A.size1();
  Ax.resize(size+1, size+1);
  Bx.resize(size+1, 1     );
  Cx.resize(1,      size+1);  
  
  // Extended System

  // Ax = | 1 C*A |
  //      | 0 A   |
  
  matrix_range<dmatrix> Ax11(Ax, range(0, 1     ), range(0, 1     ));
  matrix_range<dmatrix> Ax12(Ax, range(0, 1     ), range(1, size+1));
  matrix_range<dmatrix> Ax21(Ax, range(1, size+1), range(0, 1     ));
  matrix_range<dmatrix> Ax22(Ax, range(1, size+1), range(1, size+1));
  
  Ax11 = didentity(1,1);        Ax12 = ublas::prod(C, A);
  Ax21 = dzeromatrix(size, 1);  Ax22 = A;
  
  // Bx = | C*B |
  //      | B   |
  
  matrix_range<dmatrix> Bx1(Bx, range(0, 1     ), range(0,1));
  matrix_range<dmatrix> Bx2(Bx, range(1, size+1), range(0,1));
  
  Bx1 = ublas::prod(C, B);
  Bx2 = B;
  
  // Cx = | 1 zeros |
  
  matrix_range<dmatrix> Cx1(Cx, range(0, 1), range(0, 1     ));
  matrix_range<dmatrix> Cx2(Cx, range(0, 1), range(1, size+1));
  
  Cx1 = didentity(1, 1);
  Cx2 = dzeromatrix(1, size);
}

bool LinearControl_SISO::solveDiscreteLQRegulator(dmatrix& Kd, dmatrix& P, dmatrix& Ad, dmatrix& Bd, dmatrix& Cd, dmatrix& Qd, double Rd)
{
  const static double Peps = 1.0e-15; // «
  const static int    Nmax = 10000;   // ׻
  
  // solve riccati equation
  const unsigned int size = Ad.size1();
  dmatrix Ppre(dzeromatrix(size, size));
  dvector ddiagP(size);
  
  int n = 0;
  
  do{
    Ppre =  P;
    
    dmatrix AtP(ublas::prod(ublas::trans(Ad), P));
    dmatrix BtP(ublas::prod(ublas::trans(Bd), P));
    dmatrix AtPA(ublas::prod(AtP, Ad));
    dmatrix AtPB(ublas::prod(AtP, Bd));
    dmatrix AtPBBtPA(ublas::prod(AtPB, ublas::trans(AtPB)));
    dmatrix BtPB(ublas::prod(BtP, Bd));
    
    P = Qd + AtPA - AtPBBtPA / (Rd + BtPB(0,0));
    
    for( unsigned int i = 0 ; i < size ; i++ )
      ddiagP(i) = P(i,i) - Ppre(i,i);
    
    n++;
  }while( n < Nmax && ublas::inner_prod(ddiagP,ddiagP) > Peps );
  
  dmatrix BtP(ublas::prod(ublas::trans(Bd), P));
  dmatrix BtPA(ublas::prod(BtP, Ad));
  dmatrix BtPB(ublas::prod(BtP, Bd));
  
  Kd = BtPA / (Rd + BtPB(0,0));
  
  if( n == Nmax )
    return false;
  
  return true;
}
