/*
 * Copyright (c) 2009,
 * @author Mitsuharu Morisawa
 *
 * AIST
 *
 * All rights reserved.
 *
 * This program is made available under the terms of the Eclipse Public License
 * v1.0 which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 */
#include <cstdio>

#include "ModelManager.h"

using namespace motion_generator;


ModelManager::ModelManager(istringstream& strm)
{
  setupModel(strm);
  ModelManager::Initialize();
}

ModelManager::ModelManager(int argc, char* argv[])
{
  setupModel(argc, argv);
  ModelManager::Initialize();
}

ModelManager::ModelManager(const char *model, int argc, char* argv[])
{
  setupModel(model, argc, argv);
  ModelManager::Initialize();
}

ModelManager::ModelManager(BodyPtr& body_)
{
  body = body_;
  setupLinkModel();
  ModelManager::Initialize();
}

void ModelManager::Initialize(void)
{
  for( int i = 0; i < DOF; i++ ){
    pos_save[i] = 0.0;
    joint_max[i] = 0.0;
    joint_min[i] = 0.0;
    vjoint_max[i] = 0.0;
  }
}

ModelManager::~ModelManager()
{
}

void ModelManager::setupLinkModel(void)
{
  waistLink = body->link(WAIST_JOINT);
  ankleLink[RLEG] = body->link(RLEG_END);
  ankleLink[LLEG] = body->link(LLEG_END);
  hipLink[RLEG] = body->link(R_HIP_Y);
  hipLink[LLEG] = body->link(L_HIP_Y);
  kneeLink[RLEG] = body->link(R_KNEE_P);
  kneeLink[LLEG] = body->link(L_KNEE_P);

  for( int i = RLEG; i <= LLEG ; i++ )
    waistToAnklePath[i] = body->getJointPath(waistLink, ankleLink[i]);

#if defined(RARM_END) && defined(LARM_END)
  wristLink[RLEG] = body->link(RARM_END);
  wristLink[LLEG] = body->link(LARM_END);
  for( int i = RLEG; i <= LLEG ; i++ )
    waistToWristPath[i] = body->getJointPath(waistLink, wristLink[i]);
#endif

#if defined(CHEST_JOINT)
  chestLink = body->link(CHEST_JOINT);
  waistToChestPath = body->getJointPath(waistLink, chestLink);
#endif

#if defined(NECK_JOINT)
  neckLink = body->link(NECK_JOINT);
  waistToNeckPath = body->getJointPath(waistLink, neckLink);
#endif

  Total_mass = body->totalMass();

  waistLink->p = 0.0;
  waistLink->R = tvmet::identity<matrix33>();
  for( int i = 0 ; i < body->numJoints() ; i++ ){
    body->joint(i)->q = 0.0;
    body->joint(i)->dq = 0.0;
  }
  body->calcForwardKinematics(true);
  Xg_offset_ini = body->calcCM();

  gsensobj = NULL;
  if( body->numSensors(Sensor::ACCELERATION) )
    gsensobj = body->sensor(Sensor::ACCELERATION, 0);

  const double PowerSupply = 48.0;        // [V]
  setVelocityLimit( PowerSupply );

  Foot_distance = ankleLink[LLEG]->p(Y) - ankleLink[RLEG]->p(Y);
  Leg_length_max = tvmet::norm2(hipLink[RLEG]->p - ankleLink[RLEG]->p);
}

void ModelManager::setupModel(istringstream& strm)
{
#ifdef _WIN32
  string url = "file:///";
#else
  string url = "file://";
#endif
  url+=getenv("OPENHRPHOME");
  url+="/Controller/IOserver/robot/";

  const char *model_url = getenv("ROBOTMODEL");
  if (model_url){
    url+=model_url;
  }else{
    url+=URL;
  }

  body = loadBodyFromModelLoader(url.c_str(), strm);
  if(!body) {
    cerr << ":Warning " << url << " cannot be loaded" << endl;
    throw url;
  }

  setupLinkModel();
}

void ModelManager::setupModel(int argc, char* argv[])
{
#ifdef _WIN32
  string url = "file:///";
#else
  string url = "file://";
#endif
  url+=getenv("OPENHRPHOME");
  url+="/Controller/IOserver/robot/";

  const char *model_url = getenv("ROBOTMODEL");
  if (model_url){
    url+=model_url;
  }else{
    url+=URL;
  }

  body = loadBodyFromModelLoader(url.c_str(), argc, argv);
  if(!body) {
    cerr << ":Warning " << url << " cannot be loaded" << endl;
    throw url;
  }

  setupLinkModel();
}

void ModelManager::setupModel(const char *model, int argc, char* argv[])
{
#ifdef _WIN32
  string url = "file:///";
#else
  string url = "file://";
#endif
  url+= model;

  body = loadBodyFromModelLoader(url.c_str(), argc, argv);
  if(!body) {
    cerr << ":Warning " << url << " cannot be loaded" << endl;
    throw url;
  }

  setupLinkModel();
}

#define MIN_Fz 25.0 //refer to PatternGenerator/server/ModelManager2.cpp
void ModelManager::calcModelZMP(vector3& zmp_model_, double zmp_z, vector3& LinMom_pre, vector3& AngMom_pre, bool& firsttime)
{
  const static double Mg = Total_mass * Gv;

  vector3 LinMom, AngMom;
  body->calcForwardKinematics(true);
  vector3 total_com(body->calcCM());
  body->calcTotalMomentum(LinMom, AngMom);

  vector3 dP, dL;
  // differenciate angular momentum and linear momentum
  if( firsttime == true ){
    dP = 0.0;
    dL = 0.0;
    firsttime = false;
  }
  else{
    dP = (LinMom - LinMom_pre) / DT;
    dL = (AngMom - AngMom_pre) / DT;
  }

  AngMom_pre = AngMom;
  LinMom_pre = LinMom;

  double Fgz = dP(Z) + Mg;
  if( Fgz > MIN_Fz ){
    zmp_model_(X) = (Mg*total_com(X) + zmp_z*dP(X) - dL(Y)) / Fgz;
    zmp_model_(Y) = (Mg*total_com(Y) + zmp_z*dP(Y) + dL(X)) / Fgz;
    zmp_model_(Z) = zmp_z;
  }
  else{
    zmp_model_(X) = total_com(X);
    zmp_model_(Y) = total_com(Y);
    zmp_model_(Z) = zmp_z;
  }
}

void ModelManager::calcModelZMP(vector3& zmp_model_, vector3& total_com_, double zmp_z, vector3& LinMom_pre, vector3& AngMom_pre, bool& firsttime)
{
  const static double Mg = Total_mass * Gv;

  vector3 LinMom, AngMom;
  body->calcForwardKinematics(true);
  total_com_ = body->calcCM();
  body->calcTotalMomentum(LinMom, AngMom);

  vector3 dP, dL;
  // differenciate angular momentum and linear momentum
  if( firsttime == true ){
    dP = 0.0;
    dL = 0.0;
    firsttime = false;
  }
  else{
    dP = (LinMom - LinMom_pre) / DT;
    dL = (AngMom - AngMom_pre) / DT;
  }

  AngMom_pre = AngMom;
  LinMom_pre = LinMom;

  double Fgz = dP(Z) + Mg;
  if( Fgz > MIN_Fz ){
    zmp_model_(X) = (Mg*total_com_(X) + zmp_z*dP(X) - dL(Y)) / Fgz;
    zmp_model_(Y) = (Mg*total_com_(Y) + zmp_z*dP(Y) + dL(X)) / Fgz;
    zmp_model_(Z) = zmp_z;
  }
  else{
    zmp_model_(X) = total_com_(X);
    zmp_model_(Y) = total_com_(Y);
    zmp_model_(Z) = zmp_z;
  }
}

////////////////////////////////////////////////////////////////////////////////
//  refer to PatternGenerator/server/ModelManager2.{h,cpp}
void ModelManager::calcGsensorReference(vector3& gsens, vector3& o_Vgsens_pre, bool& firsttime)
{
  if( gsensobj ){
    vector3 o_Vgsens(gsensobj->link->v + gsensobj->link->R * tvmet::cross(gsensobj->link->w, gsensobj->localPos));

    // acceleration without gravity effect
    // .gsens is used for Kalman filter offset cancelation at stabilizer
    if( firsttime ){
      firsttime = false;
      o_Vgsens_pre = o_Vgsens;
    }

    vector3 o_Agsens((o_Vgsens-o_Vgsens_pre)/DT);
    o_Vgsens_pre = o_Vgsens;

    // data in sensor coordinate
    gsens = tvmet::trans(gsensobj->link->R) * o_Agsens;
  }
}

void ModelManager::checkJointLimits(double t_now, bool& firsttime)
{
  if( firsttime ){
    firsttime = false;
    for(int i = 0 ; i < DOF ; i++ ){
      pos_save[i]  = body->joint(i)->q;
      joint_max[i] = body->joint(i)->q;
      joint_min[i] = body->joint(i)->q;
      vjoint_max[i] = 0.0;
    }
  }
  else {
    for(int i = 0 ; i < DOF ; i++ )
      pos_save[i] += body->joint(i)->dq * DT;
  }

  for(int i = 0 ; i < DOF ; i++){
    // angle limit
    if( !( body->joint(i)->llimit==0.0 && body->joint(i)->ulimit==0.0 ) ) {
      double x = body->joint(i)->q;
      if( x < body->joint(i)->llimit ){
        printf("t= %6.4f : %20s %8.4f < %8.4f (llimit)\n", t_now, body->joint(i)->name.c_str(),x,body->joint(i)->llimit);
      }
      else if( x > body->joint(i)->ulimit ){
        printf("t= %6.4f : %20s %8.4f > %8.4f (ulimit)\n", t_now, body->joint(i)->name.c_str(),x,body->joint(i)->ulimit);
      }
      if( x > joint_max[i] )
        joint_max[i] = x;
      else if( x < joint_min[i] )
        joint_min[i] = x;
    }
    // speed magnitude limit (lvlimit,uvlimit=0 means no information)
    if( !( body->joint(i)->lvlimit==0.0 && body->joint(i)->uvlimit==0.0 ) ){
      double v = fabs( body->joint(i)->dq );
      if ( v > body->joint(i)->uvlimit ){
        printf("t= %6.4f : %20s |%8.4f| > %8.4f (vlimit)\n", t_now, body->joint(i)->name.c_str(), v, body->joint(i)->uvlimit);
      }
      if( v > vjoint_max[i] ) vjoint_max[i] = v;
    }
  }
}

// refer to PatternGenerator/server/ModelManager2.cpp
void ModelManager::setVelocityLimit(double PowerSupply)
{
  for( int i = 0 ; i < DOF ; i++ ){
    if( body->joint(i)->uvlimit == 0.0 && body->joint(i)->lvlimit == 0.0 ){
      // no explicit velocity limit data

      double vlimit = 0.0;
      // voltage to drive the servomotor [V]
      double ke = body->joint(i)->torqueConst * body->joint(i)->gearRatio;       // ke [V/(rad/s)] = kt [Nm/A]
      if( ke != 0.0 )
        vlimit = PowerSupply / ke;      // maximum joint speed

      body->joint(i)->uvlimit =  vlimit;
      body->joint(i)->lvlimit = -vlimit;
    }
  }
}

// refer to PatternGenerator/server/ModelManager2.cpp
void ModelManager::showJointLimits( void )
{
  printf("---------------------------------------------------------\n");
  printf("    Joint name        |   min     max    max_v    pverr\n");
  printf("---------------------------------------------------------\n");
  for( int i  = 0 ; i < DOF ; i++ ){
    double jmax = (joint_max[i]-body->joint(i)->llimit)/( body->joint(i)->ulimit - body->joint(i)->llimit );
    double jmin = (joint_min[i]-body->joint(i)->llimit)/( body->joint(i)->ulimit - body->joint(i)->llimit );
    double vmax = vjoint_max[i] / body->joint(i)->uvlimit;
    double pverr = ToDeg(pos_save[i]-body->joint(i)->q);

    // warning mark
    char jmaxc = (jmax> 1.0)? 'X':' ';
    char jminc = (jmin< 0.0)? 'X':' ';
    char vmaxc = (vmax> 1.0)? 'X':' ';
    char pverrc = (fabs(pverr) > 1.0)? 'X':' ';
    char namec = (jmaxc=='X'||jminc=='X'||vmaxc=='X'||pverrc=='X')? 'X':' ';

    printf("%c%20s | %c%5.1f%% %c%5.1f%% %c%5.1f%% %c%5.2f deg\n",
           namec,body->joint(i)->name.c_str(),jminc,jmin*100,jmaxc,jmax*100,vmaxc,vmax*100,pverrc,pverr);
  }
}
