/*
 * Copyright (c) 2009,
 * @author Mitsuharu Morisawa
 *
 * AIST
 *
 * All rights reserved.
 *
 * This program is made available under the terms of the Eclipse Public License
 * v1.0 which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 */
#include <iostream>
#include <math.h>
#include "MotionInterpolator.h"

using namespace std;
using namespace motion_interpolator;

MotionInterpolator::MotionInterpolator(void)
{
  size = 0;
  mdat = NULL;
  coeff = NULL;
}

MotionInterpolator::MotionInterpolator(interpolator_type ip_type_, int data_num)
{
  InterpolatorType(ip_type_, data_num);
}

void MotionInterpolator::InterpolatorType(interpolator_type ip_type_, int data_num)
{
  ip_type = ip_type_;
  
  switch(ip_type){
  case POINT_TO_POINT:
    size = 2;
    mdat = new MotionData[2];
    coeff = new double[6];    
    for( int i = 0 ; i < 6 ; i++ )
      coeff[i] = 0.0;
    getMotionDataPVA = (getMotionDataPVA_type)&MotionInterpolator::getMotionData_PointToPoint;
    break;
    
  case POINT_TO_POINT_EXP:
    size = 2;
    mdat = new MotionData[2];
    coeff = new double[4];    
    for( int i = 0 ; i < 4 ; i++ )
      coeff[i] = 0.0;
    getMotionDataPVA = (getMotionDataPVA_type)&MotionInterpolator::getMotionData_PointToPoint_Exp;
    break;
    
  case SINGLE_MIDDLE_POINT: 
    size = 3;
    mdat = new MotionData[3];
    coeff = new double[10];
    for( int i = 0 ; i < 10 ; i++ )
      coeff[i] = 0.0;
    getMotionDataPVA = (getMotionDataPVA_type)&MotionInterpolator::getMotionData_SingleMiddlePoint;
    break;
    
  case DOUBLE_MIDDLE_POINT:
    size = 4;
    mdat = new MotionData[4];
    coeff = new double[10];
    for( int i = 0 ; i < 10 ; i++ )
      coeff[i] = 0.0;
    getMotionDataPVA = (getMotionDataPVA_type)&MotionInterpolator::getMotionData_DoubleMiddlePoint;
    break;
    
  case CUBIC_SPLINE:
  case CLAMPED_CUBIC_SPLINE:
    size = data_num;
    mdat = new MotionData[data_num];
    coeff = new double[data_num*2];
    for( int i = 0 ; i < data_num*2 ; i++ )
      coeff[i] = 0.0;
    getMotionDataPVA = (getMotionDataPVA_type)&MotionInterpolator::getMotionData_CubicSpline;
    break;
    
  case TERMINAL_ACC_VEL_ZERO:
    size = 2;
    mdat = new MotionData[2];
    coeff = new double[5];
    for( int i = 0 ; i < 5 ; i++ )
      coeff[i] = 0.0;
    getMotionDataPVA = (getMotionDataPVA_type)&MotionInterpolator::getMotionData_TerminalAccVelZero;
    break;

  case TERMINAL_VEL_ZERO:
    size = 2;
    mdat = new MotionData[2];
    coeff = new double[4];
    for( int i = 0 ; i < 4 ; i++ )
      coeff[i] = 0.0;
    getMotionDataPVA = (getMotionDataPVA_type)&MotionInterpolator::getMotionData_TerminalVelZero;
    break;
    
  case VELOCITY_PROFILE:
    size = 4;
    mdat = new MotionData[4];
    coeff = new double[12];
    for( int i = 0 ; i < 12 ; i++ )
      coeff[i] = 0.0;
    getMotionDataPVA = (getMotionDataPVA_type)&MotionInterpolator::getMotionData_VelocityProfile;
    break;
    
  default:
    size = 0;
    mdat = NULL;
    coeff = NULL;
    getMotionDataPVA = NULL;
  }

  for(int i = 0 ; i < size ; i++){
    mdat[i].n = 0;
    mdat[i].p = 0.0;
    mdat[i].v = 0.0;
    mdat[i].a = 0.0;
  }
}

MotionInterpolator::~MotionInterpolator()
{
  if( mdat  != NULL ) delete [] mdat;
  if( coeff != NULL ) delete [] coeff;
}

void MotionInterpolator::calc_iMT(double t, double iMT[4][4], double &inv_detMT)
{
  const double E  = exp(t);
  const double Em = 1.0 / E;       /* exp(-time) */
  
  iMT[0][0] = -1.0 +Em;
  iMT[0][1] = -1.0 +Em +t;
  iMT[0][2] =  1.0 -Em;
  iMT[0][3] = -Em +1.0 -t*Em;
  iMT[1][0] = -1.0 +E;
  iMT[1][1] =  1.0 -E +t;
  iMT[1][2] =  1.0 -E;
  iMT[1][3] = -1.0 +E -E*t;
  iMT[2][0] = -Em +E;
  iMT[2][1] =  2.0-Em -E;
  iMT[2][2] =      Em -E;
  iMT[2][3] =  2.0-Em -E;
  iMT[3][0] =  2.0-Em -E;
  iMT[3][1] = -Em +E  -2.0*t;
  iMT[3][2] = 2.0 -Em -E -Em*t +E*t;
  iMT[3][3] = Em -E +Em*t +E*t;
  
  inv_detMT = 1.0/(4.0 - 2.0*Em - 2.0*E - Em*t + E*t);     
}

// x : a * x = b
void MotionInterpolator::gauss4(double a[][4], double x[], double b[])
{
  /*   forward elimination   */
  for(int  k = 0; k < 3 ; k++){
    double div = 1.0 / a[k][k];
    for(int i = k + 1 ; i < 4 ; i++){ 
      double multi = a[i][k] * div;
      for(int j = k + 1; j < 4 ; j++)
	a[i][j] = a[i][j] - multi * a[k][j];
      
      b[i] = b[i] - multi * b[k];
    }
  }
  
  /*   backward substitution    */
  x[3] = b[3] / a[3][3];
  for(int i = 2 ; i >= 0 ; i--){
    double sum = b[i];
    for(int j = i + 1 ; j < 4 ; j++)
      sum = sum - a[i][j] * x[j];
    x[i] = sum / a[i][i];
  }
}
  
void MotionInterpolator::calcCoefficient_polynomial(MotionData *dat, double *k)
{
  static double A[4][4]={{0.0}}, B[4]={0.0}, Kd[4]={0.0};
  
  if( (dat[1].n > dat[0].n) && (dat[2].n > dat[1].n) ){
    double t11 = (double)(dat[1].n - dat[0].n)*DT;
    double t12 = t11 * t11;
    double t13 = t11 * t12;
    double t14 = t11 * t13;
    
    double t21 = (double)(dat[1].n - dat[2].n)*DT;
    double t22 = t21 * t21;
    double t23 = t21 * t22;
    double t24 = t21 * t23;
    
    A[0][0] = t13;     A[0][1] = t14;      A[0][2] = 0.0;      A[0][3] = 0.0;
    A[1][0] = 3.0*t12; A[1][1] = 4.0*t13;  A[1][2] = -3.0*t22; A[1][3] = -4.0*t23;
    A[2][0] = 6.0*t11; A[2][1] = 12.0*t12; A[2][2] = -6.0*t21; A[2][3] = -12.0*t22;
    A[3][0] = 0.0;     A[3][1] = 0.0;      A[3][2] = t23;      A[3][3] = t24;
    
    B[0] = dat[1].p - dat[0].p - dat[0].v * t11 - dat[0].a * t12 / 2.0;
    B[1] = dat[2].v + dat[2].a * t21 - dat[0].v - dat[0].a * t11;
    B[2] = dat[2].a - dat[0].a;
    B[3] = dat[1].p - dat[2].p - dat[2].v * t21 - dat[2].a * t22 / 2.0;
    
    //ClapackWrapper::dgesvx(A, Kd, B);
    //Kd = Inverse(A) * B;
    gauss4(A, Kd, B);
    
    k[0] = dat[0].p;  k[1] = dat[0].v;  k[2] = dat[0].a/2.0;  k[3] = Kd[0];  k[4] = Kd[1];
    k[5] = dat[2].p;  k[6] = dat[2].v;  k[7] = dat[2].a/2.0;  k[8] = Kd[2];  k[9] = Kd[3];
  }
  else{
    k[0] = dat[0].p;  k[1] = 0.0;  k[2] = 0.0;  k[3] = 0.0;  k[4] = 0.0;
    k[5] = dat[2].p;  k[6] = 0.0;  k[7] = 0.0;  k[8] = 0.0;  k[9] = 0.0;
  }
}

void MotionInterpolator::makeSplineTable(double qn, double un)
{
  for(int i = 1 ; i < size - 1 ; i++ ){
    double sig = ((double)(mdat[i].n - mdat[i-1].n)) / ((double)(mdat[i+1].n - mdat[i-1].n));
    double p = sig * coeff[i-1] + 2.0;
    coeff[i] = (sig - 1.0) / p;
    coeff[i + size] = (mdat[i+1].p - mdat[i].p) / ((double)(mdat[i+1].n - mdat[i].n)*DT)
      - (mdat[i].p - mdat[i-1].p) / ((double)(mdat[i].n - mdat[i-1].n)*DT);
    coeff[i + size] = (6.0*coeff[i + size] / ((double)(mdat[i+1].n - mdat[i-1].n)*DT) - sig*coeff[i-1 + size]) / p;
  }
  
  coeff[size-1] = (un - qn * coeff[size-2 + size]) / (qn * coeff[size-2] + 1.0);
  for( int j = size - 2 ; j >= 0 ; j-- )
    coeff[j] = coeff[j] * coeff[j+1] + coeff[j + size];
}

void MotionInterpolator::setNaturalSplineCoeff(void)
{
  for( int i = 0 ; i < size ; i++ ){
    coeff[i] = 0.0;
    coeff[i + size] = 0.0;
  }
  double qn = 0.0;
  double un = 0.0;
  
  makeSplineTable(qn, un);
}

void MotionInterpolator::setClampedSplineCoeff(double data_dot_ini, double data_dot_end)
{
  coeff[0] = -0.5;
  coeff[0 + size] = (3.0 / ((double)(mdat[1].n - mdat[0].n)*DT)) *
    ((mdat[1].p - mdat[0].p) / ((double)(mdat[1].n - mdat[0].n)*DT) - data_dot_ini);
  for( int i = 1 ; i < size ; i++ ){
    coeff[i] = 0.0;
    coeff[i + size] = 0.0;
  }
  double qn = 0.5;
  double un = (3.0 / ((double)(mdat[size-1].n - mdat[size-2].n)*DT)) *
    (data_dot_end - (mdat[size-1].p - mdat[size-2].p) / ((double)(mdat[size-1].n - mdat[size-2].n)*DT));
  
  makeSplineTable(qn, un);
}

// POINT_TO_POINT
// Fifth order polynomial with initial and terminal position, velocity, and acceleration conditions
// mdat[0] initial condition : (ni, pi ,vi ,ai)
// mdat[1] termial condition : (nf, pf, vf, af)
void MotionInterpolator::setMotionData(int ni, double pi, double vi, double ai, int nf, double pf, double vf, double af)
{
  mdat[0].n = ni;  mdat[0].p = pi;  mdat[0].v = vi;  mdat[0].a = ai;
  mdat[1].n = nf;  mdat[1].p = pf;  mdat[1].v = vf;  mdat[1].a = af;
  
  if( ip_type == POINT_TO_POINT){
    if( nf > ni ){
      double t1 = (double)(nf-ni)*DT;
      double t2 = t1*t1;
      double t3 = t1*t2;
      double t4 = t1*t3;
      double t5 = t1*t4;
      
      coeff[0] = pi;
      coeff[1] = vi;
      coeff[2] = ai / 2.0;
      coeff[3] = ( 20.0*(pf-pi) - ( 8.0*vf+12.0*vi)*t1 + (    af-3.0*ai)*t2) / (2.0*t3);
      coeff[4] = (-30.0*(pf-pi) + (14.0*vf+16.0*vi)*t1 - (2.0*af-3.0*ai)*t2) / (2.0*t4);
      coeff[5] = ( 12.0*(pf-pi) - 6.0*(vf+vi)*t1 + (af-ai)*t2) / (2.0*t5);
    }
    else{
      coeff[0] = pi;
      coeff[1] = 0.0;
      coeff[2] = 0.0;
      coeff[3] = 0.0;
      coeff[4] = 0.0;
      coeff[5] = 0.0;
    }
  }
  else if(ip_type == POINT_TO_POINT_EXP){
    if( nf > ni ){
      double t = (double)(ni-nf)*DT;
      double iMT[4][4], inv_detMT;
      
      calc_iMT(t, iMT, inv_detMT);
      for(int i = 0; i < 4; i++)
	coeff[i] = (iMT[i][0] * pi + iMT[i][1] * vi + iMT[i][2] * pf + iMT[i][3] * vf) * inv_detMT;
    }
    else{
      coeff[0] = pi;
      coeff[1] = 0.0;
      coeff[2] = 0.0;
      coeff[3] = 0.0;
    }
  }
}

// SINGLE_MIDDLE_POINT
// ,־(1),ü4-4λ¿༰
// mdat[0] initial condition : (ni, pi ,vi ,ai)
// mdat[1] intermidiate condition : (nm, pm)
// mdat[2] termial condition : (nf, pf, vf, af)
void MotionInterpolator::setMotionData(int ni, double pi, double vi, double ai, int nm, double pm, int nf, double pf, double vf, double af)
{
  if( mdat == NULL )
    InterpolatorType(SINGLE_MIDDLE_POINT);
  else if( ip_type != SINGLE_MIDDLE_POINT ){
    delete [] mdat;
    delete [] coeff;
    InterpolatorType(SINGLE_MIDDLE_POINT);
  }
  
  mdat[0].n = ni;  mdat[0].p = pi;  mdat[0].v = vi;   mdat[0].a = ai;
  mdat[1].n = nm;  mdat[1].p = pm;  mdat[1].v = 0.0;  mdat[1].a = 0.0;
  mdat[2].n = nf;  mdat[2].p = pf;  mdat[2].v = vf;   mdat[2].a = af;

  calcCoefficient_polynomial(mdat, coeff);
}

// DOUBLE_MIDDLE_POINT
// ,־(2),ü4-4-4λ¿༰
// mdat[0] initial condition : (ni, pi ,vi ,ai)
// mdat[1] intermidiate condition : (n1, p1)
// mdat[2] intermidiate condition : (n2, p2)
// mdat[3] termial condition : (nf, pf, vf, af)
void MotionInterpolator::setMotionData(int ni, double pi, double vi, double ai, int n1, double p1, int n2, double p2, int nf, double pf, double vf, double af)
{
  if( mdat == NULL )
    InterpolatorType(DOUBLE_MIDDLE_POINT);
  else if( ip_type != DOUBLE_MIDDLE_POINT ){
    delete [] mdat;
    delete [] coeff;
    InterpolatorType(DOUBLE_MIDDLE_POINT);
  }
  
  mdat[0].n = ni;  mdat[0].p = pi;  mdat[0].v = vi;   mdat[0].a = ai;
  mdat[1].n = n1;  mdat[1].p = p1;  mdat[1].v = 0.0;  mdat[1].a = 0.0;
  mdat[2].n = n2;  mdat[2].p = p2;  mdat[2].v = 0.0;  mdat[2].a = 0.0;
  mdat[3].n = nf;  mdat[3].p = pf;  mdat[3].v = vf;   mdat[3].a = af;

  calcCoefficient_polynomial(mdat, coeff); // from mdat[0] to mdat[2]
}

// natural cubic spline & clamped cubic spline
void MotionInterpolator::setMotionData(std::vector<int>& nd, std::vector<double>& pd, bool use_last_data)
{
  for( int i = 0 ; i < size ; i++ ){
    mdat[i].n = nd[i];
    mdat[i].p = pd[i];
  }
  
  if( use_last_data )
    getTerminalData_CubicSpline(mdat[0].p, mdat[0].v, mdat[0].a);
  else{
    mdat[0].v = 0.0;
    mdat[0].a = 0.0;
  }
  
  if( ip_type == CUBIC_SPLINE )
    setNaturalSplineCoeff();
  else
    setClampedSplineCoeff(mdat[0].v, 0.0);
}

// TERMINAL_ACC_VEL_ZERO
void MotionInterpolator::setMotionData(int ni, double pi, double vi, double ai, int nf)
{
  if( mdat == NULL )
    InterpolatorType(TERMINAL_ACC_VEL_ZERO);
  else if( ip_type != TERMINAL_ACC_VEL_ZERO){
    delete [] mdat;
    delete [] coeff;
    InterpolatorType(TERMINAL_ACC_VEL_ZERO);
  }
  
  mdat[0].n = ni;  mdat[0].p = pi;   mdat[0].v = vi;   mdat[0].a = ai;
  mdat[1].n = nf;  mdat[1].p = 0.0;  mdat[1].v = 0.0;  mdat[1].a = 0.0;

  if( nf > ni ){
    double t1 = (double)(nf-ni)*DT;
    double t2 = t1*t1;
    double t3 = t1*t2;

    coeff[0] =  pi;
    coeff[1] =  vi;
    coeff[2] =  ai / 2.0;
    coeff[3] = -(3.0 * vi + 2.0 * ai * t1) / (3.0 * t2);
    coeff[4] =  (2.0 * vi +       ai * t1) / (4.0 * t3);
  }
  else{
    coeff[0] = pi;
    coeff[1] = 0.0;
    coeff[2] = 0.0;
    coeff[3] = 0.0;
    coeff[4] = 0.0;
  }
}

// TERMINAL_VEL_ZERO
void MotionInterpolator::setMotionData(int ni, double pi, double vi, int nf)
{
  if( mdat == NULL )
    InterpolatorType(TERMINAL_VEL_ZERO);
  else if( ip_type != TERMINAL_VEL_ZERO){
    delete [] mdat;
    delete [] coeff;
    InterpolatorType(TERMINAL_VEL_ZERO);
  }
  
  mdat[0].n = ni;  mdat[0].p = pi;   mdat[0].v = vi;   mdat[0].a = 0.0;
  mdat[1].n = nf;  mdat[1].p = 0.0;  mdat[1].v = 0.0;  mdat[1].a = 0.0;
  
  if( nf > ni ){
    double t1 = (double)(nf-ni)*DT;
    
    coeff[0] =  pi;
    coeff[1] =  vi;
    coeff[2] = -vi / t1;
    coeff[3] =  vi / (3.0*t1*t1);
  }
  else{
    coeff[0] = pi;
    coeff[1] = 0.0;
    coeff[2] = 0.0;
    coeff[3] = 0.0;
  }
}

// VELOCITY_PROFILE
void MotionInterpolator::setMotionData(int ni, double pi, double vi, double ai, int n1, int n2, int nf, double pf, double vf, double af)
{
  if( mdat == NULL )
    InterpolatorType(VELOCITY_PROFILE);
  else if( ip_type != VELOCITY_PROFILE){
    delete [] mdat;
    delete [] coeff;
    InterpolatorType(VELOCITY_PROFILE);
  }
  
  mdat[0].n = ni; mdat[0].p = pi;  mdat[0].v = vi;  mdat[0].a = ai;
  mdat[1].n = n1; mdat[1].p = 0.0; mdat[1].v = 0.0; mdat[1].a = 0.0;
  mdat[2].n = n2; mdat[2].p = 0.0; mdat[2].v = 0.0; mdat[2].a = 0.0;
  mdat[3].n = nf; mdat[3].p = pf;  mdat[3].v = vf;  mdat[3].a = af;

  double t10 = (double)(n1-ni)*DT;
  double t20 = (double)(n2-ni)*DT;
  double tf0 = (double)(nf-ni)*DT;

  double t2f   = t20 - tf0;
  double t10_2 = t10 * t10;
  double t2f_2 = t2f * t2f;

  double vmax = (12.0*(pf - pi) + 6.0*(vf*t2f - vi*t10) + af*t2f_2 - ai*t10_2) / (6.0*(tf0 + t20 - t10));
  
  coeff[0] = pi;
  coeff[1] = vi;
  coeff[2] = ai / 2.0;
  coeff[3] = (3.0*(vmax-vi)-2.0*ai*t10) / (3.0*t10_2);
  coeff[4] = (2.0*(vi-vmax)+    ai*t10) / (4.0*t10_2*t10);
  
  coeff[5] = pi + (vi+vmax)*t10/2.0 + ai*t10_2 / 12.0;
  coeff[6] = vmax;

  coeff[7] = pf;
  coeff[8] = vf;
  coeff[9] = af / 2.0;
  coeff[10] = (3.0*(vmax-vf)-2.0*af*t2f) / (3.0*t2f_2);
  coeff[11] = (2.0*(vf-vmax)+    af*t2f) / (4.0*t2f_2*t2f);
}

void MotionInterpolator::getMotionData_PointToPoint(int n_now, double& p, double& v, double& a)
{
  if( n_now < mdat[0].n ) n_now = mdat[0].n;
  if( n_now > mdat[1].n ) n_now = mdat[1].n;
  
  double tn = (double)(n_now - mdat[0].n)*DT;
  
  p = coeff[0] + (coeff[1] + (coeff[2] + (coeff[3] + (coeff[4] + coeff[5]*tn)*tn)*tn)*tn)*tn;
  v = coeff[1] + (2.0*coeff[2] + (3.0*coeff[3] + (4.0*coeff[4] + 5.0*coeff[5]*tn)*tn)*tn)*tn;
  a = 2.0*coeff[2] + (6.0*coeff[3] + (12.0*coeff[4] + 20.0*coeff[5]*tn)*tn)*tn;
}

void MotionInterpolator::getMotionData_PointToPoint_Exp(int n_now, double& p, double& v, double& a)
{
  if( n_now < mdat[0].n ) n_now = mdat[0].n;
  if( n_now > mdat[1].n ) n_now = mdat[1].n;
  
  double tn = (double)(n_now - mdat[1].n)*DT;
  
  if (tn >= -0.1 * DT){
    p = mdat[1].p + mdat[1].v * (tn + DT);
    v = mdat[1].v;
    a = 0.0;
  }
  else{
    tn += DT;
    const double Et  = exp(tn);
    const double Etm = 1.0 / Et;
    p = coeff[0]*Et + coeff[1]*Etm + coeff[2]*tn + coeff[3];
    v = coeff[0]*Et - coeff[1]*Etm + coeff[2];
    a = coeff[0]*Et + coeff[1]*Etm;
  }
}

void MotionInterpolator::getTerminalData_PointToPoint(double& p, double& v, double& a)
{
  p = mdat[1].p;
  v = mdat[1].v;
  a = mdat[1].a;
}

void MotionInterpolator::getMotionData_SingleMiddlePoint(int n_now, double& p, double& v, double& a)
{
  if( n_now < mdat[0].n ) n_now = mdat[0].n;
  if( n_now > mdat[2].n ) n_now = mdat[2].n;

  if( n_now < mdat[1].n ){
    double tn = (double)(n_now - mdat[0].n)*DT;
    p = coeff[0] + (coeff[1] + (coeff[2] + (coeff[3] + coeff[4]*tn)*tn)*tn)*tn;
    v = coeff[1] + (2.0*coeff[2] + (3.0*coeff[3] + 4.0*coeff[4]*tn)*tn)*tn;
    a = 2.0*coeff[2] + (6.0*coeff[3] + 12.0*coeff[4]*tn)*tn;
  }
  else{
    double tn = (double)(n_now - mdat[2].n)*DT;
    p = coeff[5] + (coeff[6] + (coeff[7] + (coeff[8] + coeff[9]*tn)*tn)*tn)*tn;
    v = coeff[6] + (2.0*coeff[7] + (3.0*coeff[8] + 4.0*coeff[9]*tn)*tn)*tn;
    a = 2.0*coeff[7] + (6.0*coeff[8] + 12.0*coeff[9]*tn)*tn;
  }
}

void MotionInterpolator::getTerminalData_SingleMiddlePoint(double& p, double& v, double& a)
{
  p = mdat[2].p;
  v = mdat[2].v;
  a = mdat[2].a;
}

void MotionInterpolator::getMotionData_DoubleMiddlePoint(int n_now, double& p, double& v, double& a)
{
  if( n_now < mdat[0].n ) n_now = mdat[0].n;
  if( n_now > mdat[3].n ) n_now = mdat[3].n;

  if( n_now <= mdat[1].n ){
    double tn = (double)(n_now - mdat[0].n)*DT;
    p = coeff[0] + (coeff[1] + (coeff[2] + (coeff[3] + coeff[4]*tn)*tn)*tn)*tn;
    v = coeff[1] + (2.0*coeff[2] + (3.0*coeff[3] + 4.0*coeff[4]*tn)*tn)*tn;
    a = 2.0*coeff[2] + (6.0*coeff[3] + 12.0*coeff[4]*tn)*tn;
    
    if( n_now == mdat[1].n){
      mdat[1].v = v;
      mdat[1].a = a;
      calcCoefficient_polynomial(&mdat[1], coeff);
    }
  }
  else if( n_now <= mdat[2].n ){
    double tn = (double)(n_now - mdat[1].n)*DT;
    p = coeff[0] + (coeff[1] + (coeff[2] + (coeff[3] + coeff[4]*tn)*tn)*tn)*tn;
    v = coeff[1] + (2.0*coeff[2] + (3.0*coeff[3] + 4.0*coeff[4]*tn)*tn)*tn;
    a = 2.0*coeff[2] + (6.0*coeff[3] + 12.0*coeff[4]*tn)*tn;
  }
  else{
    double tn = (double)(n_now - mdat[3].n)*DT;
    p = coeff[5] + (coeff[6] + (coeff[7] + (coeff[8] + coeff[9]*tn)*tn)*tn)*tn;
    v = coeff[6] + (2.0*coeff[7] + (3.0*coeff[8] + 4.0*coeff[9]*tn)*tn)*tn;
    a = 2.0*coeff[7] + (6.0*coeff[8] + 12.0*coeff[9]*tn)*tn;
  }
}

void MotionInterpolator::getTerminalData_DoubleMiddlePoint(double& p, double& v, double& a)
{
  p = mdat[3].p;
  v = mdat[3].v;
  a = mdat[3].a;
}

void MotionInterpolator::getMotionData_CubicSpline(int n_now,double& p, double& v, double& a)
{
  if( n_now < mdat[0].n      ) n_now = mdat[0].n;
  if( n_now > mdat[size-1].n ) n_now = mdat[size-1].n;
  
  int segment = 0;
  while( segment < size-2 && n_now > mdat[segment+1].n )
    segment++;
  
  double h = (double)(mdat[segment+1].n - mdat[segment].n)*DT;
  if( h < 0.0 )
    throw("Bad input to routine splint");
  double dhi = (double)(mdat[segment+1].n - n_now)*DT / h;
  double dlo = (double)(n_now - mdat[segment].n)*DT / h;
  
  p = dhi*mdat[segment].p + dlo*mdat[segment+1].p + ((dhi*dhi*dhi-dhi)*coeff[segment] + (dlo*dlo*dlo-dlo)*coeff[segment+1])*(h*h)/6.0;
  v = (-mdat[segment].p + mdat[segment+1].p)/h + ((-3.0*dhi*dhi+1.0)*coeff[segment] + (3.0*dlo*dlo-1.0)*coeff[segment+1])*h/6.0;
  a = dhi*coeff[segment] + dlo*coeff[segment+1];
}

void MotionInterpolator::getTerminalData_CubicSpline(double& p, double& v, double& a)
{
  double h = (double)(mdat[size-1].n - mdat[size-2].n)*DT;
  if( h < 0.0 )
    throw("Bad input to routine splint");
  double dlo = (double)(mdat[size-1].n - mdat[size-2].n)*DT / h;
  
  p = dlo*mdat[size-1].p + ((dlo*dlo*dlo-dlo)*coeff[size-1])*(h*h)/6.0;
  v = (-mdat[size-2].p + mdat[size-1].p)/h + (coeff[size-2] + (3.0*dlo*dlo-1.0)*coeff[size-1])*h/6.0;
  a = dlo*coeff[size-1];
}

void MotionInterpolator::getMotionData_TerminalAccVelZero(int n_now, double& p, double& v, double& a)
{
  if( n_now < mdat[0].n ) n_now = mdat[0].n;
  if( n_now > mdat[1].n ) n_now = mdat[1].n;
  
  double tn = (double)(n_now - mdat[0].n)*DT;
  p = coeff[0] + (coeff[1] + (coeff[2] + (coeff[3] + coeff[4]*tn)*tn)*tn)*tn;
  v = coeff[1] + (2.0*coeff[2] + (3.0*coeff[3] + 4.0*coeff[4]*tn)*tn)*tn;
  a = 2.0*coeff[2] + (6.0*coeff[3] + 12.0*coeff[4]*tn)*tn;
}

void MotionInterpolator::getMotionData_TerminalVelZero(int n_now, double& p, double& v, double& a)
{
  if( n_now < mdat[0].n ) n_now = mdat[0].n;
  if( n_now > mdat[1].n ) n_now = mdat[1].n;
  
  double tn = (double)(n_now - mdat[0].n)*DT;
  p = coeff[0] + (coeff[1] + (coeff[2] + coeff[3]*tn)*tn)*tn;
  v = coeff[1] + (2.0*coeff[2] + 3.0*coeff[3]*tn)*tn;
  a = 2.0*coeff[2] + 6.0*coeff[3]*tn;
}

void MotionInterpolator::getMaximumData(double& p_max, double& p_min, double& v_max)
{
  double tn = (double)(mdat[1].n - mdat[0].n)*DT;

  if( ip_type == TERMINAL_ACC_VEL_ZERO ){
  
    double tp_max, tv_max;
    double d = 2.0 * mdat[0].v + mdat[0].a * tn;
    if( fabs(d) < 1.0e-10 ){
      tp_max = 0.0;
      tv_max = 0.0;
    }
    else{
      tp_max = -(mdat[0].v * tn) / d; 
      tv_max =  (mdat[0].a * tn * tn) / (3.0 * d);
      
      if( tp_max > tn )
	tp_max = tn;
      if( tv_max > tn )
	tv_max = tn;
      
      if( tp_max < 0.0 )
	tp_max = 0.0;
      if( tv_max < 0.0 )
	tv_max = 0.0;
    }
    
    double pf = (((coeff[4] * tn + coeff[3]) * tn + coeff[2]) * tn + coeff[1]) * tn + coeff[0];
    p_max = (((coeff[4] * tp_max + coeff[3]) * tp_max + coeff[2]) * tp_max + coeff[1]) * tp_max + coeff[0];
    p_min = p_max;
    
    if( p_max < pf )
      p_max = pf;
    if( p_max < mdat[0].p )
      p_max = mdat[0].p;
    
    if( p_min > pf )
      p_min = pf;
    if( p_min > mdat[0].p )
      p_min = mdat[0].p;
  
    v_max = ((4.0*coeff[4] * tv_max + 3.0*coeff[3]) * tv_max + 2.0*coeff[2]) * tv_max + coeff[1];    
    if( fabs(v_max) < fabs(mdat[0].v) )
      v_max = mdat[0].v;
  
  }
  else if( ip_type == TERMINAL_VEL_ZERO ){
    double pterm = mdat[0].p + mdat[0].v * tn / 3.0;

    if( mdat[0].p > pterm ){
      p_max = mdat[0].p;
      p_min = pterm;
    }
    else{
      p_max = pterm;
      p_min = mdat[0].p;
    }

    v_max = mdat[0].v;
  }
  else{
    p_max = 0.0;
    p_min = 0.0;
    v_max = 0.0;
  }
}

void MotionInterpolator::getMotionData_VelocityProfile(int n_now, double& p, double& v, double& a)
{
  if( n_now <= mdat[1].n ){
    if( n_now < mdat[0].n )
      n_now = mdat[0].n;

    double tn = (double)(n_now - mdat[0].n)*DT;
    p = coeff[0] + (coeff[1] + (coeff[2] + (coeff[3] + coeff[4]*tn)*tn)*tn)*tn;
    v = coeff[1] + (2.0*coeff[2] + (3.0*coeff[3] + 4.0*coeff[4]*tn)*tn)*tn;
    a = 2.0*coeff[2] + (6.0*coeff[3] + 12.0*coeff[4]*tn)*tn;    
  }
  else if( n_now >= mdat[2].n ){
    if( n_now > mdat[3].n )
      n_now = mdat[3].n;

    double tn = (double)(n_now - mdat[3].n)*DT;
    p = coeff[7] + (coeff[8] + (coeff[9] + (coeff[10] + coeff[11]*tn)*tn)*tn)*tn;
    v = coeff[8] + (2.0*coeff[9] + (3.0*coeff[10] + 4.0*coeff[11]*tn)*tn)*tn;
    a = 2.0*coeff[9] + (6.0*coeff[10] + 12.0*coeff[11]*tn)*tn;    
  }
  else{
    double tn = (double)(n_now - mdat[1].n)*DT;
    p = coeff[5] + coeff[6] * tn;
    v = coeff[6];
    a = 0.0;
  }
}

bool MotionInterpolator::isFinish(int n_now)
{
  switch(ip_type){
  case POINT_TO_POINT:
  case POINT_TO_POINT_EXP:
    if( mdat[1].n <= n_now ) return true;
    break;
    
  case SINGLE_MIDDLE_POINT: 
    if( mdat[2].n <= n_now ) return true;
    break;
    
  case DOUBLE_MIDDLE_POINT:
    if( mdat[3].n <= n_now ) return true;
    break;
    
  case CUBIC_SPLINE:
  case CLAMPED_CUBIC_SPLINE:
    if( mdat[size-1].n <= n_now ) return true;
    break;
    
  case TERMINAL_ACC_VEL_ZERO:
    if( mdat[1].n <= n_now ) return true;
    break;
    
  case TERMINAL_VEL_ZERO:
    if( mdat[1].n <= n_now ) return true;
    break;
    
  case VELOCITY_PROFILE:
    if( mdat[3].n <= n_now ) return true;
    break;
  }

  return false;
}
