/*
 * Copyright (c) 2009,
 * @author Mitsuharu Morisawa
 *
 * AIST
 *
 * All rights reserved.
 *
 * This program is made available under the terms of the Eclipse Public License
 * v1.0 which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 */
#ifndef __MINTERPOLATOR_H__
#define __MINTERPOLATOR_H__

#include <vector>
#include "mphysics.h"

namespace motion_interpolator{
  
  typedef struct{
    int n;
    double p;
    double v;
    double a;
  } MotionData;
  
  typedef enum{
    POINT_TO_POINT,
    POINT_TO_POINT_EXP,
    SINGLE_MIDDLE_POINT,
    DOUBLE_MIDDLE_POINT,
    CUBIC_SPLINE,
    CLAMPED_CUBIC_SPLINE,
    TERMINAL_ACC_VEL_ZERO,
    TERMINAL_VEL_ZERO,
    VELOCITY_PROFILE
  } interpolator_type;
  
  class MotionInterpolator{
    interpolator_type  ip_type;       // interpolation type
    int                size;          // number of motion data
    MotionData         *mdat;         // motion data
    double             *coeff;        // coefficients of polynomial
    
    void calc_iMT(double t, double iMT[4][4], double &inv_detMT);
    void gauss4(double a[][4], double x[], double b[]);
    void calcCoefficient_polynomial(MotionData *dat, double *k);
    void makeSplineTable(double qn, double un);
    void setNaturalSplineCoeff(void);
    void setClampedSplineCoeff(double data_dot_ini, double data_dot_end);
    void getMotionData_PointToPoint(int n_now, double& p, double& v, double& a);
    void getMotionData_PointToPoint_Exp(int n_now, double& p, double& v, double& a);
    void getMotionData_SingleMiddlePoint(int n_now, double& p, double& v, double& a);
    void getMotionData_DoubleMiddlePoint(int n_now, double& p, double& v, double& a);
    void getMotionData_CubicSpline(int n_now, double& p, double& v, double& a);
    void getMotionData_TerminalAccVelZero(int n_now, double& p, double& v, double& a);
    void getMotionData_TerminalVelZero(int n_now, double& p, double& v, double& a);
    void getMotionData_VelocityProfile(int n_now, double& p, double& v, double& a);
    
    void getTerminalData_PointToPoint(double& p, double& v, double& a);
    void getTerminalData_SingleMiddlePoint(double& p, double& v, double& a);
    void getTerminalData_DoubleMiddlePoint(double& p, double& v, double& a);
    void getTerminalData_CubicSpline(double& p, double& v, double& a);
    
  public:
    MotionInterpolator(void);
    MotionInterpolator(interpolator_type ip_type_, int data_num=2);
    ~MotionInterpolator();
    
    void InterpolatorType(interpolator_type ip_type_, int data_num=2);
    
    typedef void (MotionInterpolator::*getMotionDataPVA_type)(int n_now, double& p, double& v, double& a);
    typedef void (MotionInterpolator::*getMotionDataPV_type)(int n_now, double& p, double& v);
    typedef void (MotionInterpolator::*getMotionDataP_type)(int n_now, double& p);
    
    getMotionDataPVA_type getMotionDataPVA;
    void getMotionData(int n_now, double& p, double& v, double& a){ (this->*getMotionDataPVA)(n_now, p, v, a); };
    void getMotionData(int n_now, double& p, double& v){ double a = 0.0; (this->*getMotionDataPVA)(n_now, p, v, a); };
    void getMotionData(int n_now, double& p){ double a = 0.0, v = 0.0; (this->*getMotionDataPVA)(n_now, p, v, a); };
    
    // POINT_TO_POINT (POLYNOMIAL, EXPONENTIAL)
    inline void setMotionData(int ni, int nf, double pf){
      double pi, vi, ai;
      getTerminalData_PointToPoint(pi, vi, ai);
      setMotionData(ni, pi, vi, ai, nf, pf, 0.0, 0.0);
    }
    inline void setMotionData(int ni, int nf, double pf, double vf, double af){
      double pi, vi, ai;
      getTerminalData_PointToPoint(pi, vi, ai);
      setMotionData(ni, pi, vi, ai, nf, pf, vf, af);
    }
    inline void setMotionData(int ni, double pi, int nf, double pf){
      setMotionData(ni, pi, 0.0, 0.0, nf, pf, 0.0, 0.0);
    }
    void setMotionData(int ni, double pi, double vi, double ai, int nf, double pf, double vf, double af);
    
    // SINGLE_MIDDLE_POINT
    inline void setMotionData(int ni, int nm, double pm, int nf, double pf){
      double pi, vi, ai;
      getTerminalData_SingleMiddlePoint(pi, vi, ai);
      setMotionData(ni, pi, vi, ai, nm, pm, nf, pf, 0.0, 0.0);
    }
    inline void setMotionData(int ni, int nm, double pm, int nf, double pf, double vf, double af){
      double pi, vi, ai;
      getTerminalData_SingleMiddlePoint(pi, vi, ai);
      setMotionData(ni, pi, vi, ai, nm, pm, nf, pf, vf, af);
    }
    inline void setMotionData(int ni, double pi, int nm, double pm, int nf, double pf){
      setMotionData(ni, pi, 0.0, 0.0, nm, pm, nf, pf, 0.0, 0.0);
    }
    void setMotionData(int ni, double pi, double vi, double ai, int nm, double pm, int nf, double pf, double vf, double af);
    
    // DOUBLE_MIDDLE_POINT
    inline void setMotionData(int ni, int n1, double p1, int n2, double p2, int nf, double pf){
      double pi, vi, ai;
      getTerminalData_DoubleMiddlePoint(pi, vi, ai);
      setMotionData(ni, pi, vi, ai, n1, p1, n2, p2, nf, pf, 0.0, 0.0);
    }
    inline void setMotionData(int ni, int n1, double p1, int n2, double p2, int nf, double pf, double vf, double af){
      double pi, vi, ai;
      getTerminalData_DoubleMiddlePoint(pi, vi, ai);
      setMotionData(ni, pi, vi, ai, n1, p1, n2, p2, nf, pf, vf, af);
    }
    inline void setMotionData(int ni, double pi, int n1, double p1, int n2, double p2, int nf, double pf){
      setMotionData(ni, pi, 0.0, 0.0, n1, p1, n2, p2, nf, pf, 0.0, 0.0);
    }
    void setMotionData(int ni, double pi, double vi, double ai, int n1, double p1, int n2, double p2, int nf, double pf, double vf, double af);
    
    // CUBIC_SPLINE
    // CLAMPED_CUBIC_SPLINE
    void setMotionData(std::vector<int>& nd, std::vector<double>& pd, bool use_last_data = false);
    
    // TERMINAL_ACC_VEL_ZERO
    void setMotionData(int ni, double pi, double vi, double ai, int nf);
    
    // TERMINAL_VEL_ZERO
    void setMotionData(int ni, double pi, double vi, int nf);
    
    // VELOCITY_PROFILE
    void setMotionData(int ni, double pi, double vi, double ai, int n1, int n2, int nf, double pf, double vf, double af);
    
    void getMaximumData(double& pmax, double& pmin, double& vmax);
    
    bool isFinish(int n_now);
  };
  
}; // end of namespace motion_interpolator

#endif // __MINTERPOLATOR_H__
