/*
 * Copyright (c) 2009,
 * @author Mitsuharu Morisawa
 *
 * AIST
 *
 * All rights reserved.
 *
 * This program is made available under the terms of the Eclipse Public License
 * v1.0 which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 */
#include "MwalkGenerator.h"

using namespace mwalk_pattern_generator;


MwalkGenerator::MwalkGenerator(istringstream& strm) :
  ModelManager(strm),
  FootStepManager(ModelManager::Foot_distance)
{
  MwalkGenerator::Initialize();
  ANcog.setFootDistance(ModelManager::Foot_distance);
  wbmg = new WholeBodyMotionGenerator(body);
}

MwalkGenerator::MwalkGenerator(int argc, char* argv[]) :
  ModelManager(argc, argv),
  FootStepManager(ModelManager::Foot_distance)
{
  MwalkGenerator::Initialize();
  ANcog.setFootDistance(ModelManager::Foot_distance);
  wbmg = new WholeBodyMotionGenerator(body);
}

MwalkGenerator::MwalkGenerator(const char *model, int argc, char* argv[]) :
  ModelManager(model, argc, argv),
  FootStepManager(ModelManager::Foot_distance)
{
  MwalkGenerator::Initialize();
  ANcog.setFootDistance(ModelManager::Foot_distance);
  wbmg = new WholeBodyMotionGenerator(body);
}

MwalkGenerator::~MwalkGenerator()
{
  delete wbmg;
}

void MwalkGenerator::Initialize(void)
{
  verbose = false;
  
  n_clock = 0;
  n_period = 0;
  n_period_rest = 0;
  
  flg_status.reset();
  
  STEP_YAW_RATIO  =  0.2;
  WAIST_YAW_RATIO =  0.0;
  ARM_SWING_RATIO = ToRad(10.0)/0.3;
  
  AbsPosOrg = 0.0;
  AbsRpyOrg = 0.0;
  AbsPosOrgNext = 0.0;
  AbsRpyOrgNext = 0.0;
  AbsPosOrg_prev = 0.0;
  AbsRpyOrg_prev = 0.0;
  
  COG_cmd = 0.0;
  VCOG_cmd = 0.0;
  VCOGdot_cmd = 0.0;
  Zmp_cmd = 0.0;
  VZmp_cmd = 0.0;

  COG_shp = 0.0;
  VCOG_shp = 0.0;
  VCOGdot_shp = 0.0;
  Zmp_shp = 0.0;
  VZmp_shp = 0.0;
  
  COG_prev = 0.0;
  VCOG_prev = 0.0;
  Zmp_prev = 0.0;
  VZmp_prev = 0.0;
  
  COG_dfilter = 0.0;
  VCOG_dfilter = 0.0;
  VCOGdot_dfilter = 0.0;
  
  COG_cmp = 0.0;
  VCOG_cmp = 0.0;
  VCOGdot_cmp = 0.0;
  
  Zmp_model = 0.0;
  Zmp_ideal = 0.0;
  Zmp_model2 = 0.0;
  
  Xg_offset = 0.0;
  Vg_offset = 0.0;
  INIT_COG_OFFSET_GAIN = 50.0;
  UPDATE_COG_OFFSET_GAIN = 0.0;
  
  VCOGdot_res = 0.0;
  
  for( int i = RLEG ; i <= LLEG ; i++ ){
    Foot_cmd[i].Initialize(i, ModelManager::Foot_distance);
    Foot_next_cmd[i].Initialize(i, ModelManager::Foot_distance);
    Ankle_cmd[i].Initialize(i, ModelManager::Foot_distance);
    Shoulder[i].InterpolatorType(POINT_TO_POINT_EXP);
  }
  WaistHeight.InterpolatorType(SINGLE_MIDDLE_POINT);
  
  Waist_cmd.Initialize();
  Pivot_cmd[RLEG] = 0.0;
  Pivot_cmd[LLEG] = 0.0;
  TR_ankle_ds_ratio[0] = 0.5;
  TR_ankle_ds_ratio[1] = 1.0 - TR_ankle_ds_ratio[0];
  
  THmiddle_ratio = 0.5;
  XHmiddle_ratio = 0.0;
  
  total_com = 0.0;
  total_inr = 0.0;
  
  lin_mom_pre = 0.0;
  ang_mom_pre = 0.0;
  lin_mom_pre2 = 0.0;
  ang_mom_pre2 = 0.0;
  
  Gsensor_offset = 0.0;  
  Vgsens_pre = 0.0;
  
  for( int i = 0 ; i <= Ncog ; i++ ){
    zmp_des[i] = 0.0;
    vzmp_des[i] = 0.0;
    Tm[i] = 0.0;
  }
  
  for( int i = 0 ; i < Ncog ; i++ ){
    cog_height[i] = DEFAULT_COG_HEIGHT;
    omega[i] = sqrt(Gv / cog_height[i]);
  }
  
  Aginv      = dzeromatrix(Ncog*2+6, Ncog*2+6);
  Aginv_prev = dzeromatrix(Ncog*2+6, Ncog*2+6);
  for( int i = 0 ; i < 2 ; i++ ){
    Bg[i] = dzerovector(Ncog*2+6);
  }
  for( int i = 0 ; i < 3 ; i++ ){
    for( int j = 0 ; j < 7 ; j++ ){
      for( int k = 0 ; k < Ncog ; k++ )
	Kg[k*7+j][i] = 0.0;
    }
  }
  
  Reset();
  
  q_ini = 0.0;
  q_res = 0.0;
  q_res_pre = 0.0;
  q_cmd = 0.0;
  q_cmd_pre = 0.0;
  qdot_cmd = 0.0;
  qddot_cmd = 0.0;
  q_cmd2 = 0.0;
  q_cmd_pre2 = 0.0;
  qdot_cmd2 = 0.0;
  q_cmd_org = 0.0;
  
  flg_firsttime[JCHECK] = true;
  ModelManager::Initialize();
  
  Trunk_roll_fixed_height = 1.0;

#ifdef USE_CHEST_RMC
  Vctrl_cmd.resize(6+DOF_CHEST);
#else
  Vctrl_cmd.resize(6);
#endif
  PL_cmd.resize(6);
}

void MwalkGenerator::Reset(void)
{
  mmode  = INITCONF;
  n_comp = 0;
  
  flg_firsttime[ZMP] = true;
  flg_firsttime[ZMP2] = true;
  flg_firsttime[GSENS] = true;
  flg_firsttime[JOINT] = true;
  
  flg_status[READY_INITCONF] = false;
#if defined WITH_TOE_JOINT
  flg_status[ONE_STEP] = true;
#endif
  
  ws = STAND_ANKLE;
  ws_next = STAND_ANKLE;
  ws_pre = STAND_ANKLE;  
  ws_prev = STAND_ANKLE;
#ifdef WITH_TOE_JOINT
  ws_last = STAND_ANKLE;
  ws_last_prev = STAND_ANKLE;
#endif
  
  Pcog_dzmp[X].reset();
  Pcog_dzmp[Y].reset();
  waistP_pbuf.clear();
  waistV_pbuf.clear();
  waistR_pbuf.clear();
  waistW_pbuf.clear();
  for(int i = RLEG ; i <= LLEG ; i++){
    ankleP_pbuf[i].clear();
    ankleR_pbuf[i].clear();
  }
  q_pbuf.clear();
  zzmp_pbuf.clear();
}

void MwalkGenerator::setOriginForFirstStep(int leg)
{
  AbsPosOrgNext = AbsPosOrg + rotFromRpy(AbsRpyOrg) * (ANcog.Foot_ini[leg] - ANcog.Foot[leg]);
  AbsRpyOrgNext = AbsRpyOrg;  
  
  AbsPosOrg = AbsPosOrgNext;
  AbsRpyOrg = AbsRpyOrgNext;
}

void MwalkGenerator::setLandingPosition(State& Leg_next)
{
  const unsigned int FQsize = FootQueue.size();

  if( FQsize == 1 ){
    AbsPosOrgNext = AbsPosOrg + rotFromRpy(AbsRpyOrg) * FootQueue[0].p / 2.0;
    AbsRpyOrgNext = AbsRpyOrg + FootQueue[0].rpy;
    
    Leg_next.P     = AbsPosOrg + rotFromRpy(AbsRpyOrg) * FootQueue[0].p + rotFromRpy(AbsRpyOrgNext) * Leg_next.offset;
    Leg_next.R_rpy = AbsRpyOrgNext + FootQueue[0].frpy;
  }
  else if( FQsize > 1 ){
    AbsPosOrgNext = AbsPosOrg + rotFromRpy(AbsRpyOrg) * FootQueue[0].p;
    AbsRpyOrgNext = AbsRpyOrg + FootQueue[0].rpy;
    
    Leg_next.P     = AbsPosOrgNext + rotFromRpy(AbsRpyOrgNext) * Leg_next.offset;
    Leg_next.R_rpy = AbsRpyOrgNext + FootQueue[0].frpy;
  }
  else{
    AbsPosOrgNext = AbsPosOrg;
    AbsRpyOrgNext = AbsRpyOrg;
    
    Leg_next.P     = AbsPosOrgNext + Leg_next.offset;
    Leg_next.R_rpy = AbsRpyOrgNext + FootQueue[0].frpy;
  }
  
  //cout << "AbsPosOrgNext = " << AbsPosOrgNext << endl;
  //cout << "AbsRpyOrgNext = " << AbsRpyOrgNext << endl;
}

void MwalkGenerator::setLandingPosition(State& SwingLeg_next, State& SupportLeg)
{
  const unsigned int FQsize = FootQueue.size();
  
  if( FQsize == 1 ){
    SwingLeg_next.P     = SupportLeg.P + rotFromRpy(SupportLeg.R_rpy) * FootQueue[0].p / 2.0;
    SwingLeg_next.R_rpy = SupportLeg.R_rpy + FootQueue[0].rpy;
    
    AbsPosOrgNext       = SwingLeg_next.P - rotFromRpy(SwingLeg_next.R_rpy) * SwingLeg_next.offset;
    AbsRpyOrgNext       = SwingLeg_next.R_rpy;
  }
  else if( FQsize > 1 ){
    SwingLeg_next.P     = SupportLeg.P + rotFromRpy(SupportLeg.R_rpy) * FootQueue[0].p;
    SwingLeg_next.R_rpy = SupportLeg.R_rpy + FootQueue[0].rpy;
    
    AbsPosOrgNext       = SwingLeg_next.P - rotFromRpy(SwingLeg_next.R_rpy) * SwingLeg_next.offset;
    AbsRpyOrgNext       = SwingLeg_next.R_rpy;
  }
}

void MwalkGenerator::setWaistHeightParam(double hip_z, double step_length, int n_start, int n_period)
{
  WaistHeight.setMotionData(
    n_start,
    Waist_cmd.P(Z),
    n_start + (int)((double)n_period * THmiddle_ratio),
    Waist_cmd.P(Z) + hip_z * THmiddle_ratio + fabs(step_length) * XHmiddle_ratio,
    n_start + n_period,
    Waist_cmd.P(Z) + hip_z
  );
}

void MwalkGenerator::setWaistHeightParam(FootPlace& FP, State& foot_, State& foot_next_, int n_clock)
{
  int n_start, n_period;
  if( FP.hip_z > 0.0 ){
    n_start  = n_clock + NPhip_updown_ini(FP.nss+FP.nds, 0);
    n_period = NPhip_updown_end(FP.nss+FP.nds, 0);
  }
  else{
    n_start  = n_clock + NPhip_updown_ini(FP.nss, 1);
    n_period = NPhip_updown_end(FP.nss, 1);
  }
  
  setWaistHeightParam(FP.hip_z, foot_next_.P(X)-foot_.P(X), n_start, n_period);
}

void MwalkGenerator::setSwingArmParam(WalkingState& ws_next_, unsigned int FQsize, double step_length, int n_start, int n_period)
{
#ifdef WITH_ARM_JOINT
  if( FQsize == 1 ){
    Shoulder[RLEG].setMotionData(n_start, q_cmd[R_SHOULDER_P], n_start+n_period, q_ini[R_SHOULDER_P]);
    Shoulder[LLEG].setMotionData(n_start, q_cmd[L_SHOULDER_P], n_start+n_period, q_ini[L_SHOULDER_P]);
  }
  else{
    WalkingState sup_leg = getSupportLeg(ws_next_);
    if( sup_leg == LLEG ){
      Shoulder[RLEG].setMotionData(n_start, n_start+n_period,  step_length*ARM_SWING_RATIO+q_ini[R_SHOULDER_P]);
      Shoulder[LLEG].setMotionData(n_start, n_start+n_period, -step_length*ARM_SWING_RATIO+q_ini[L_SHOULDER_P]);
    }
    else if( sup_leg == RLEG ){
      Shoulder[RLEG].setMotionData(n_start, n_start+n_period, -step_length*ARM_SWING_RATIO+q_ini[R_SHOULDER_P]);
      Shoulder[LLEG].setMotionData(n_start, n_start+n_period,  step_length*ARM_SWING_RATIO+q_ini[L_SHOULDER_P]);
    }
    else{
      Shoulder[RLEG].setMotionData(n_start, q_cmd[R_SHOULDER_P], n_start+n_period, q_ini[R_SHOULDER_P]);
      Shoulder[LLEG].setMotionData(n_start, q_cmd[L_SHOULDER_P], n_start+n_period, q_ini[L_SHOULDER_P]);
    }
  }
#endif
}

void MwalkGenerator::adjustSupportTime(vector3& Xgi, vector3& Vgi, vector3& Zmpi, vector3& VZmpi, vector3& Xgf, vector3& Vgf, vector3& Zmpf, vector3& VZmpf, double foot_angle, int& n_comp_)
{
  // ѹϰ֤Ȥɸ
  double cth = cos(foot_angle);
  double sth = sin(foot_angle);
  
  double xgi_ =  (Xgi(X)-Zmpi(X))  * cth + (Xgi(Y)-Zmpi(Y))  * sth;
  double xgf_ =  (Xgf(X)-Zmpf(X))  * cth + (Xgf(Y)-Zmpf(Y))  * sth;
  double vgi_ =  (Vgi(X)-VZmpi(X)) * cth + (Vgi(Y)-VZmpi(Y)) * sth;
  double vgf_ =  (Vgf(X)-VZmpf(X)) * cth + (Vgf(Y)-VZmpf(Y)) * sth;
  
  if( verbose ){
    //cout << "Zmpi = " << Zmpi << endl;
    //cout << "Zmpf = " << Zmpf << endl;

    cout << "th = " << ToDeg(foot_angle) << "[deg]" << endl;
    cout << "Xgi_ = " << xgi_ << ",Vgi_ = " << vgi_ << endl;
    cout << "Xgf_ = " << xgf_ << ",Vgf_ = " << vgf_ << endl;
  }
  
  double t_comp = 0.0;
  
#if 1
  //if( fabs(omega*xgi_ + vgi_) > 1.0e-4 || fabs(omega*xgf_ + vgf_) > 1.0e-4 ){
  //if( fabs(omega*xgf_ - vgf_) > 1.0e-4 ){
  //double cog_ratio = (omega*xgi_ - vgi_) / (omega*xgf_ - vgf_);
  if( fabs(omega[2]*xgi_ + vgi_) > 1.0e-8 ){
    double cog_ratio = (omega[2]*xgf_ + vgf_) / (omega[2]*xgi_ + vgi_);
    
    if( cog_ratio > 0.0 )
      t_comp = log(cog_ratio) / omega[2];
    else
      t_comp = (double)(Nss+Nds)*DT;
    //t_comp = fabs(log(fabs(cog_ratio))) / omega[2];
    
    //t_comp = log(fabs(cog_ratio)) / omega[2];
    
    if( t_comp + (double)(FootQueue[0].nss)*DT <= 0.5 ){
      t_comp = 0.5 - (double)(FootQueue[0].nss)*DT;
      
      if( verbose )
	cout << "lower limit of time compenstation" << endl;
    }
    else if( t_comp > (double)(Nss+Nds)*DT/2.0 )
      t_comp = (double)(Nss+Nds)*DT/2.0;
    //else if( t_comp > (double)(Nss+Nds)*DT )
    //t_comp = (double)(Nss+Nds)*DT;
    
    if( verbose ){
      cout << "cog_ratio = " << cog_ratio << endl;
      
      ////double cog_ratio2 = (omega*xgi_ - vgi_) / (omega*xgf_ - vgf_);
      //double cog_ratio2 = (omega*xgf_ + vgf_) / (omega*xgi_ + vgi_);
      //double t_comp2 = log(fabs(cog_ratio2)) / omega;
      //cout << "Tcomp2 = " << t_comp2 << ", cog_ratio2 = " << cog_ratio2 << endl;
    }
  }
#endif
#if 0
  else if( fabs(omega*xgf_ - vgf_) > 1.0e-4 ){
    double cog_ratio = (omega*xgi_ - vgi_) / (omega*xgf_ - vgf_);
    
    if( cog_ratio > 0.0 ){
      t_comp = log(cog_ratio) / omega;
      
      if( t_comp + FootQueue[0].nss <= 0.3 )
	t_comp = 0.3 - FootQueue[0].nss;
    }
  }
#endif
  
  if( t_comp >= 0.0 ) n_comp_ = (int)(t_comp/DT+0.5);
  else                n_comp_ = (int)(t_comp/DT-0.5);
}

void MwalkGenerator::calcZMPTrackingControl(const double zmp[3], const double accel[3])
{
  if( flg_status[REACTIVE_MOTION] ){
    Zmp_res = zmp[0], zmp[1], zmp[2];

    VCOGdot_res = accel[0], accel[1], accel[2];
    
#if 0  
    vector3 Zmp_err(0.0);
    Zmp_err(X) = threshold(Zmp_res(X)-Zmp_ref(X), 0.05, -0.05);
    //Zmp_err(Y) = threshold(Zmp_res(Y)-Zmp_ref(Y), 0.05, -0.05);
    
#if 1
    //VCOGdot_cmp(X) = Kzmp_track * (Zmp_res(X) - Zmp_ref(X));
    VCOGdot_cmp(X) = Gv / (COG_cmd(Z) - Zmp_cmd(Z)) * Zmp_err(X);
    VCOGdot_cmp(Y) = Gv / (COG_cmd(Z) - Zmp_cmd(Z)) * Zmp_err(Y);
    VCOGdot_cmp(Z) = 0.0;
    //VCOGdot_cmpI(X) = Gv / (COG_cmd(Z) - Zmp_cmd(Z)) * (Zmp_res(X) - Zmp_ref(X)) + Zmp_err(X);
    //Zmp_err(X) += Kzmp_err * (Zmp_res(X) - Zmp_ref(X));
#endif
    COG_cmp += VCOG_cmp * DT;
    VCOG_cmp += VCOGdot_cmp * DT;
    
#if 0
    cout << "ZMP_res = " << Zmp_res(X) << "," << Zmp_res(Y) << endl;
    cout << "ZMP_ref = " << Zmp_ref(X) << "," << Zmp_ref(Y) << endl;
    cout << "COG_cmp = " << COG_cmp(X) << "," << COG_cmp(Y) << endl;
#endif
    
#endif
  }
}

void MwalkGenerator::setFootPlace(int& n_comp_)
{
  if( !FootQueueNext.empty() ){
    if( verbose ){
      std::cout << "Change Foot Place! (t=" << (double)n_clock*DT << "[s])" << endl;
      std::cout << "Previous : (" << FootQueue[0].p(X) << "," << FootQueue[0].p(Y)     << "," << FootQueue[0].p(Z)     << "," << FootQueue[0].rpy(Z) << ")" << endl;
      std::cout << "Next : (" << FootQueueNext[0].p(X) << "," << FootQueueNext[0].p(Y) << "," << FootQueueNext[0].p(Z) << "," << FootQueueNext[0].rpy(Z) << ")" << endl;
    }
    
    double foot_angle = atan2( FootQueueNext[0].p(Y)-FootQueue[0].p(Y), FootQueueNext[0].p(X)-FootQueue[0].p(X) );
    
    if( !FootQueue.empty() ){
      FootQueueNext.back().support_leg = FootQueue[0].support_leg;
      FootQueueNext.back().nss = FootQueue[0].nss;
      FootQueueNext.back().nds = FootQueue[0].nds;
      FootQueue.pop_front();
    }
    FootQueue.push_front(FootQueueNext.back());
    FootQueueNext.pop_back();
    
    if( ws != STAND_ANKLE && flg_status[TIME_ADJUSTMENT] == true ){
      vector3 Xgi, Vgi, Vgdoti, Zmpi, VZmpi;
      ANcog.getMotion(Kg, Xgi, Vgi, Vgdoti, Zmpi, VZmpi, Tm, (double)n_clock*DT, Ncog);
      
      vector3 Xgf, Vgf, Zmpf, VZmpf;
      popCOGCoef(Xgf, Vgf, Zmpf, VZmpf, 0);
      
      adjustSupportTime(Xgi, Vgi, Zmpi, VZmpi, Xgf, Vgf, Zmpf, VZmpf, foot_angle, n_comp_); // (from, to) : 
      
      if( verbose ){
	//std::cout << "Now : Xgi = " << Xgi(X) << "," << Xgi(Y) << "," << Xgi(Z) << ",";
	//std::cout << "Vgi = " << Vgi(X) << "," << Vgi(Y) << "," << Vgi(Z) << ",";
	//std::cout << "Zmpi = " << Zmpi(X) << "," << Zmpi(Y) << "," << Zmpi(Z) << endl;
	//std::cout << "Mod : Xgf = " << Xgf(X) <<"," << Xgf(Y) << "," << Xgf(Z) << ",";
	//std::cout << "Vgf = " << Vgf(X) << "," << Vgf(Y) << "," << Vgf(Z) << ",";
	//std::cout << "Zmpf = " << Zmpf(X) << "," << Zmpf(Y) << "," << Zmpf(Z) << endl;
	std::cout << "adjust time = " << (double)n_comp_*DT << "[s]\n" << endl;
      }
    }
    
    flg_status[FOOT_MODIFICATION] = true;
  }
  else{
    n_comp_ = 0;
    flg_status[FOOT_MODIFICATION] = false;
  }
}

void MwalkGenerator::updateFootPlace(State& Leg, State& Leg_next)
{
  FootQueue.pop_front();
  
  Leg.P     = Leg_next.P;
  Leg.V     = Leg_next.V;
  Leg.Vdot  = Leg_next.Vdot;
  Leg.R_rpy = Leg_next.R_rpy;
  Leg.W     = Leg_next.W;
  Leg.Wdot  = Leg_next.Wdot;
  
  AbsPosOrg = AbsPosOrgNext;
  AbsRpyOrg = AbsRpyOrgNext;
}

void MwalkGenerator::makeUpperBodyMotion(vector3& waist_p, vector3& waist_v, vector3& waist_rpy, vector3& waist_rpy_dot, vector<double>& angle, vector<double>& angle_dot)
{
  for( int i = DOF_LEG ; i < DOF ; i++ ){
    q_cmd[i] = angle[i];
    qdot_cmd[i] = angle_dot[i];
  }
  
  Waist_cmd.P(Z) = waist_p(Z);
  Waist_cmd.V(Z) = waist_v(Z);
  
  const matrix33 Rfoot_local(rotFromRpy(0.0, 0.0, (Ankle_cmd[RLEG].R_rpy(2) + Ankle_cmd[LLEG].R_rpy(2)) / 2.0));
  const vector3  Xfoot_local(trans(Rfoot_local) * (Foot_cmd[LLEG].P - Foot_cmd[RLEG].P));
  double waist_angle = atan2(Xfoot_local(X), Xfoot_local(Y));
  waist_rpy(2) = -waist_angle * STEP_YAW_RATIO + (Ankle_cmd[RLEG].R_rpy(2) + Ankle_cmd[LLEG].R_rpy(2)) / 2.0;
  Waist_cmd.R_rpy = waist_rpy;
  Waist_cmd.R = rotFromRpy(Waist_cmd.R_rpy);
  
  const double dfx = Foot_cmd[LLEG].V(X)-Foot_cmd[RLEG].V(X);
  const double dfy = Foot_cmd[LLEG].V(Y)-Foot_cmd[RLEG].V(Y);
  double waist_angle_dot =
    STEP_YAW_RATIO * (Xfoot_local(Y)*dfx - Xfoot_local(X)*dfy) / (Xfoot_local(X)*Xfoot_local(X) + Xfoot_local(Y)*Xfoot_local(Y))
    + (Ankle_cmd[RLEG].W(2) + Ankle_cmd[LLEG].W(2)) / 2.0;  
  vector3 waist_cmd_drpy(0.0, 0.0, waist_angle_dot);
  Waist_cmd.W = omegaFromRpyAngularVelocity(Waist_cmd.R_rpy, waist_cmd_drpy);
  
#ifdef WITH_CHEST_JOINT
  q_cmd[CHEST_Y] += waist_angle * STEP_YAW_RATIO;
#endif
  
  for( int i = DOF_LEG ; i < DOF ; i++ )
    body->joint(i)->q = q_cmd[i];
}

void MwalkGenerator::makeUpperBodyMotion(void)
{
  const matrix33 Rfoot_local_trans(rotFromRpy(0.0, 0.0, -(Ankle_cmd[RLEG].R_rpy(2) + Ankle_cmd[LLEG].R_rpy(2)) / 2.0));
  const vector3  Xfoot_local(Rfoot_local_trans * (Foot_cmd[LLEG].P - Foot_cmd[RLEG].P));
  double waist_angle = atan2(Xfoot_local(X), Xfoot_local(Y));
  
#if 1
  vector3 R_rpy_cmd_pre(Waist_cmd.R_rpy);
  if( flg_status[TRUNK_ROLL] && Trunk_roll_fixed_height != 0.0 )
    Waist_cmd.R_rpy(0) = atan2(Waist_cmd.P(Y), Trunk_roll_fixed_height);
  Waist_cmd.R_rpy(2) = -waist_angle * STEP_YAW_RATIO + (Ankle_cmd[RLEG].R_rpy(2) + Ankle_cmd[LLEG].R_rpy(2)) / 2.0;
  Waist_cmd.R = rotFromRpy(Waist_cmd.R_rpy);
  
  vector3 waist_cmd_drpy((Waist_cmd.R_rpy - R_rpy_cmd_pre)/DT);
  Waist_cmd.W = omegaFromRpyAngularVelocity(Waist_cmd.R_rpy, waist_cmd_drpy);
#endif
  
#ifdef WITH_CHEST_JOINT
  q_cmd[CHEST_Y] = q_ini[CHEST_Y] + waist_angle * (STEP_YAW_RATIO + WAIST_YAW_RATIO);
  q_cmd[CHEST_P] = q_ini[CHEST_P];
  body->joint(CHEST_Y)->q = q_cmd[CHEST_Y];
  body->joint(CHEST_P)->q = q_cmd[CHEST_P];
#endif
  
#ifdef WITH_ARM_JOINT  
  Shoulder[RLEG].getMotionData(n_clock, q_cmd[R_SHOULDER_P]);
  Shoulder[LLEG].getMotionData(n_clock, q_cmd[L_SHOULDER_P]);
  body->joint(R_SHOULDER_P)->q  = q_cmd[R_SHOULDER_P];
  body->joint(L_SHOULDER_P)->q  = q_cmd[L_SHOULDER_P];
#endif
}

void MwalkGenerator::makeCOGMotion(void)
{
  ANcog.getMotion(Kg, COG_cmd, VCOG_cmd, VCOGdot_cmd, Zmp_cmd, VZmp_cmd, Tm, (double)n_clock*DT, Ncog);
  WaistHeight.getMotionData(n_clock, Waist_cmd.P(Z), Waist_cmd.V(Z), Waist_cmd.Vdot(Z));
  
  if( UPDATE_COG_OFFSET_GAIN > 0.0 ){
    waistLink->p = Waist_cmd.P;
    waistLink->R = Waist_cmd.R;
    for( int i = 0 ; i < DOF ; i++ )
      body->joint(i)->q = q_cmd[i];
    body->calcForwardKinematics();
    vector3 COG_model(body->calcCM());
    vector3 Xg_offset_cmd(tvmet::trans(Waist_cmd.R) * (COG_model - Waist_cmd.P));
    vector3 Xg_offset_diff((Xg_offset_cmd - Xg_offset_ini) * UPDATE_COG_OFFSET_GAIN);
    Xg_offset_ini += Xg_offset_diff * DT;
    
    Xg_offset = Waist_cmd.R * Xg_offset_ini;
    Vg_offset = cross(Waist_cmd.W, Xg_offset) + Waist_cmd.R * Xg_offset_diff;
  }
  else{
    Xg_offset = Waist_cmd.R * Xg_offset_ini;
    Vg_offset = cross(Waist_cmd.W, Xg_offset);
  }
  
  COG_cmd(Z)  = Waist_cmd.P(Z) + Xg_offset(Z);
  VCOG_cmd(Z) = Waist_cmd.V(Z) + Vg_offset(Z);
  
  Waist_cmd.P(X) = COG_cmd(X)  + COG_cmp(X) - COG_shp(X)  - Xg_offset(X);
  Waist_cmd.P(Y) = COG_cmd(Y)  + COG_cmp(Y) - COG_shp(Y)  - Xg_offset(Y);
  Waist_cmd.V(X) = VCOG_cmd(X) + VCOG_cmp(X) - VCOG_shp(X) - Vg_offset(X);
  Waist_cmd.V(Y) = VCOG_cmd(Y) + VCOG_cmp(Y) - VCOG_shp(Y) - Vg_offset(Y);
}

void MwalkGenerator::makeWholeBodyMotion(void)
{
  ANcog.getMotion(Kg, COG_cmd, VCOG_cmd, VCOGdot_cmd, Zmp_cmd, VZmp_cmd, Tm, (double)n_clock*DT, Ncog);

#if 0
  const matrix33 Rfoot_local_trans(rotFromRpy(0.0, 0.0, -(Ankle_cmd[RLEG].R_rpy(2) + Ankle_cmd[LLEG].R_rpy(2)) / 2.0));
  const vector3  Xfoot_local(Rfoot_local_trans * (Foot_cmd[LLEG].P - Foot_cmd[RLEG].P));
  double waist_angle = atan2(Xfoot_local(X), Xfoot_local(Y));
  
  static vector3 Waist_rpy_ref(Waist_cmd.R_rpy);
  vector3 Waist_rpy_pre(Waist_rpy_ref);
  if( flg_status[TRUNK_ROLL] && Trunk_roll_fixed_height != 0.0 )
    Waist_rpy_ref(0) = atan2(Waist_cmd.P(Y), Trunk_roll_fixed_height);
  Waist_rpy_ref(2) = -waist_angle * STEP_YAW_RATIO + (Ankle_cmd[RLEG].R_rpy(2) + Ankle_cmd[LLEG].R_rpy(2)) / 2.0;
  vector3 Waist_drpy_ref((Waist_rpy_ref - Waist_rpy_pre)/DT);
#endif
  
  if( UPDATE_COG_OFFSET_GAIN > 0.0 ){
    waistLink->p = Waist_cmd.P;
    waistLink->R = Waist_cmd.R;
    for( int i = 0 ; i < DOF ; i++ )
      body->joint(i)->q = q_cmd[i];
    body->calcForwardKinematics();
    vector3 COG_model(body->calcCM());
    vector3 Xg_offset_cmd(tvmet::trans(Waist_cmd.R) * (COG_model - Waist_cmd.P));
    vector3 Xg_offset_diff((Xg_offset_cmd - Xg_offset_ini) * UPDATE_COG_OFFSET_GAIN);
    Xg_offset_ini += Xg_offset_diff * DT;
    
    Xg_offset = Waist_cmd.R * Xg_offset_ini;
    Vg_offset = cross(Waist_cmd.W, Xg_offset) + Waist_cmd.R * Xg_offset_diff;
  }
  else{
    Xg_offset = Waist_cmd.R * Xg_offset_ini;
    Vg_offset = cross(Waist_cmd.W, Xg_offset);
  }
  
  const double RpyGain = 10.0;
#ifdef USE_CHEST_RMC
  dvector Vctrl_ref(6+DOF_CHEST);
#else
  dvector Vctrl_ref(6);
#endif
  for( int i = X ; i <= Z ; i++ ){
    //Vctrl_ref(i) = VCOG_cmd(i); // better than zero (low noize of waist velocity)
    Vctrl_ref(i) = 10.0*(COG_cmd(i) - Xg_offset(i) - Waist_cmd.P(i)) + VCOG_cmd(i) - Vg_offset(i);
  }
  
  vector3 Waist_rpy_cmd(rpyFromRot(Waist_cmd.R));
  for( int i = X ; i <= Z ; i++ )
    //Vctrl_ref(i+3) = RpyGain*(Waist_rpy_ref(i) - Waist_rpy_cmd(i)) + Waist_drpy_ref(i);
    Vctrl_ref(i+3) = -RpyGain*Waist_rpy_cmd(i);
  
#ifdef USE_CHEST_RMC
  for( int i = 0 ; i < DOF_CHEST ; i++ )
    Vctrl_ref(6+i) = -10.0 * q_cmd[DOF_LEG+i];
#endif
  
  static bool firsttime = true;
  if( firsttime ){
    firsttime = false;
    
    WaistHeight.getMotionData(n_clock, Waist_cmd.P(Z), Waist_cmd.V(Z), Waist_cmd.Vdot(Z));
    COG_cmd(Z)  = Waist_cmd.P(Z) + Xg_offset(Z);
    VCOG_cmd(Z) = Waist_cmd.V(Z) + Vg_offset(Z);
    
    Waist_cmd.P(X) = COG_cmd(X) - Xg_offset(X);
    Waist_cmd.P(Y) = COG_cmd(Y) - Xg_offset(Y);
    Waist_cmd.V(X) = VCOG_cmd(X) - Vg_offset(X);
    Waist_cmd.V(Y) = VCOG_cmd(Y) - Vg_offset(Y);
  }

#if 1
#ifndef USE_CHEST_RMC
  wbmg->makeWholeBodyMotion(Waist_cmd, COG_cmd, VCOG_cmd, Ankle_cmd, Zmp_cmd, Vctrl_ref);
#else
  wbmg->makeWholeBodyMotion(Waist_cmd, q_cmd, qdot_cmd, COG_cmd, VCOG_cmd, Ankle_cmd, Zmp_cmd, Vctrl_ref);
#endif

#else

#ifndef USE_CHEST_RMC
  wbmg->makeWholeBodyMotion(Vctrl_cmd, PL_cmd, Waist_cmd, COG_cmd, VCOG_cmd, Ankle_cmd, Zmp_cmd, Vctrl_ref);
#else
  wbmg->makeWholeBodyMotion(Vctrl_cmd, PL_cmd, Waist_cmd, q_cmd, qdot_cmd, COG_cmd, VCOG_cmd, Ankle_cmd, Zmp_cmd, Vctrl_ref);
#endif
#endif
}

void MwalkGenerator::updateWholeBodyMotion(void)
{
#ifndef USE_CHEST_RMC
  wbmg->updateWholeBodyMotion(Waist_cmd, Vctrl_cmd);
#else
  wbmg->updateWholeBodyMotion(Waist_cmd, q_cmd, qdot_cmd, Vctrl_cmd);
#endif
  //wbmg->calcMomentum(PL_cmd, Waist_cmd, q_cmd, qdot_cmd);
}

void MwalkGenerator::pushCOGCoef(int n_period_, int n_clock_)
{
  n_prev = n_clock_;
  n_period_prev = n_period_;
#ifdef WITH_TOE_JOINT
  ws_last_prev = ws_last;
#endif
  ws_prev = ws;
  FootQueuePrev.push_front(FootQueue.front());
  AbsPosOrg_prev = AbsPosOrg;
  AbsRpyOrg_prev = AbsRpyOrg;
  COG_prev  = COG_cmd;
  VCOG_prev = VCOG_cmd;
  Zmp_prev  = Zmp_cmd;
  VZmp_prev = VZmp_cmd;
  Aginv_prev = Aginv;
}

void MwalkGenerator::popCOGCoef(vector3&  Xgi_cmd, vector3& Vgi_cmd, vector3& Zmpi_cmd, vector3& VZmpi_cmd, int n_aft)
{
  // 򥻥å
  FootQueue.push_front(FootQueuePrev.front());
  FootQueuePrev.pop_front();

  double Tm_prev[Ncog+1];
#ifndef WITH_TOE_JOINT
  ANcog.predictZMPPath(Tm_prev, zmp_des, vzmp_des, Zmp_prev, VZmp_prev, AbsPosOrg_prev, AbsRpyOrg_prev, FootQueue, ws_prev, n_period_prev, n_prev);
#else
  ANcog.predictZMPPath(Tm_prev, zmp_des, vzmp_des, Zmp_prev, VZmp_prev, AbsPosOrg_prev, AbsRpyOrg_prev, FootQueue, ws_prev, ws_last_prev, n_period_prev, n_prev);
#endif
  
#ifndef WITH_TOE_JOINT
  for( int i = 0 ; i < Ncog ; i++ )
    cog_height[i] = Waist_cmd.P(Z) + Xg_offset_ini(Z) + (zmp_des[i+1](Z) - zmp_des[i](Z)) / 2.0;
#else
  for( unsigned int i = 0 ; i <= Nstep ; i++ ){
    for( int j = 0 ; j < 4 ; i++ )
      cog_height[i+j] = Waist_cmd.P(Z) + Xg_offset_ini(Z) + (zmp_des[i*4+3](Z) - zmp_des[i*4](Z)) / 2.0;
  }
#endif // WITH_TOE_JOINT 
  
  ANcog.setCOGHeight(cog_height);
  ANcog.makeCoefParam(Bg, Tm_prev, zmp_des, vzmp_des, COG_prev, VCOG_prev);
  ANcog.makeCoefVector(Kg, Aginv_prev, Bg, Tm_prev, zmp_des, vzmp_des);
  ANcog.getMotion(Kg, Xgi_cmd, Vgi_cmd, Zmpi_cmd, VZmpi_cmd, Tm_prev, (double)(n_clock+n_aft)*DT, Ncog);

  FootQueue.pop_front();
}

void MwalkGenerator::makeCOGCoef(vector3& Xgi_cmd_, vector3& Vgi_cmd_, vector3& Zmpi_cmd_, vector3& VZmpi_cmd_, int n_period_, int n_clock_)
{
#ifndef FAST_COG_TRAJECTORY
  if( n_period_ > 1 ){
#endif

#ifndef WITH_TOE_JOINT
    ANcog.predictZMPPath(Tm, zmp_des, vzmp_des, Zmpi_cmd_, VZmpi_cmd_, AbsPosOrg, AbsRpyOrg, FootQueue, ws, n_period_, n_clock_);
    for( int i = 0 ; i < Ncog ; i++ )
      cog_height[i] = Waist_cmd.P(Z) + Xg_offset_ini(Z) + (zmp_des[i+1](Z) - zmp_des[i](Z)) / 2.0;
#else
    ANcog.predictZMPPath(Tm, zmp_des, vzmp_des, Zmpi_cmd_, VZmpi_cmd_, AbsPosOrg, AbsRpyOrg, FootQueue, ws, ws_last, n_period_, n_clock_);
    for( unsigned int i = 0 ; i < Nstep ; i++ ){
      for( int j = 0 ; j < 4 ; j++ )
	cog_height[i*4+j] = Waist_cmd.P(Z) + Xg_offset_ini(Z) + (zmp_des[i*4+3](Z) - zmp_des[i*4](Z)) / 2.0;
    }
    for( int j = 0 ; j < 3 ; j++ )
      cog_height[Nstep*4+j] = Waist_cmd.P(Z) + Xg_offset_ini(Z) + (zmp_des[Nstep*4+3](Z) - zmp_des[Nstep*4](Z)) / 2.0;
#endif // WITH_TOE_JOINT 
    
    ANcog.setCOGHeight(cog_height);
    
#ifdef FAST_COG_TRAJECTORY
    dmatrix ainv(Ncog,2);
    matrix22 A[Ncog], An[Ncog];
    matrix24 B[Ncog], Bn[Ncog];
    vector2  C[Ncog];
    
    ANcog.setCoefMatrix(ainv, A, B, C, An, Bn, Tm, ws);
    ANcog.setBoundaryCondition(Kg, ainv, A, B, C, An, Bn, Tm, zmp_des, vzmp_des, Xgi_cmd_, Vgi_cmd_);
#else
    ANcog.makeCoefMatrix(Aginv, Tm); 
    ANcog.makeCoefParam(Bg, Tm, zmp_des, vzmp_des, Xgi_cmd_, Vgi_cmd_);
    ANcog.makeCoefVector(Kg, Aginv, Bg, Tm, zmp_des, vzmp_des);
#endif
  
    if( flg_status[ZMP_SHAPING_PREVIEW] == true ){
      Pcog[X].n_start = n_clock_;
      Pcog[Y].n_start = n_clock_;
      ANcog.makeZMPError(zmp_err, zmp_des, Kg, Tm);
    }
    
    if( flg_status[TIME_ADJUSTMENT] == true )
      pushCOGCoef(n_period_, n_clock_);
    
#ifndef FAST_COG_TRAJECTORY
  }
#endif
}

void MwalkGenerator::makeCOGCoefReactive(vector3& Xgi_reac_, vector3& Vgi_reac_, vector3& Zmpi_reac_, vector3& VZmpi_reac_, int n_period_, int n_clock_)
{
  ANcog_reac.setCOGHeight(cog_height);
  
  dmatrix  ainv(Ncog,2);
  matrix22 A[Ncog], An[Ncog];
  matrix24 B[Ncog], Bn[Ncog];
  vector2  C[Ncog];
  
  const int Nstep_mod = 1;
  
  dmatrix M(2, Nstep_mod+2);
  dvector s_mod[2];
  s_mod[X].resize(M.size2());
  s_mod[X] = dzerovector(M.size2());
  s_mod[Y].resize(M.size2());
  s_mod[Y] = dzerovector(M.size2());  
  
  ANcog_reac.setCoefMatrix2(M, A, B, C, An, Bn, Tm, ws);
  ANcog_reac.setBoundaryCondition2(Kg_reac, M, s_mod, A, B, C, An, Bn, Tm, zmp_reac, vzmp_reac, Xgi_reac_, Vgi_reac_);
  
  int FQsize = FootQueue.size();
  if( FQsize > Nstep_mod )
    FQsize = Nstep_mod;
  for( int i = 0 ; i < FQsize ; i++ ){
    FootQueue[i].p(X) += s_mod[X](i);
    FootQueue[i].p(Y) += s_mod[Y](i);
  }
  flg_status[FOOT_MODIFICATION] = true;
}

void MwalkGenerator::makeReactiveMotion(void)
{
  vector3 Zmp_err(0.0);
  Zmp_err(X) = Zmp_res(X) - Zmp_ref(X);
  //Zmp_err(X) = threshold(Zmp_res(X)-Zmp_ref(X), 0.05, -0.05);
  //Zmp_err(Y) = threshold(Zmp_res(Y)-Zmp_ref(Y), 0.05, -0.05);
  
  VCOGdot_cmp(X) = Gv / (COG_cmd(Z) - Zmp_cmd(Z)) * Zmp_err(X);
  VCOGdot_cmp(Y) = Gv / (COG_cmd(Z) - Zmp_cmd(Z)) * Zmp_err(Y);
  VCOGdot_cmp(Z) = 0.0;
  
  if( COG_cmp(X) + VCOG_cmp(X) * DT < -0.2 || COG_cmp(X) + VCOG_cmp(X) * DT > 0.2 )
    VCOG_cmp = 0.0;
  
  COG_cmp  += VCOG_cmp * DT;
  VCOG_cmp += VCOGdot_cmp * DT;
  
  FootQueue[0].p += 2.0 * VCOG_cmp * DT;
  
  if( FootQueue[0].p(X) > 0.2 )
    FootQueue[0].p(X) = 0.2;
  else if( FootQueue[0].p(X) < -0.2 )
    FootQueue[0].p(X) = -0.2;
  
  //FootQueue[0].p = COG_cmp;
  flg_status[FOOT_MODIFICATION] = true;
}

void MwalkGenerator::calcZMPErrorOfAnalyticalSolution(int n_period_rest)
{
  vector3 zmp_err_(0.0);
  static vector3 zmp_err_max(0.0);
  if( verbose && isSingleSupport(ws) ){
    if( isLeftSupport(ws) )
      zmp_err_ = Zmp_model - Foot_cmd[LLEG].P;
    else if( isRightSupport(ws) )
      zmp_err_ = Zmp_model - Foot_cmd[RLEG].P;
    
    if( fabs(zmp_err_(X)) > fabs(zmp_err_max(X)) )
      zmp_err_max(X) = zmp_err_(X);
    if( fabs(zmp_err_(Y)) > fabs(zmp_err_max(Y)) )
      zmp_err_max(Y) = zmp_err_(Y);
  }
  
  if( n_period_rest == 0 ){
    if( verbose && ws != STAND_ANKLE )
      cout << "T=" << (double)n_clock*DT << "(" << n_clock << ")" << ", Zmp_err Max = " << zmp_err_max(X) << "," << zmp_err_max(Y) << endl;
    zmp_err_max = 0.0;
  }
}

void MwalkGenerator::setZMPShaping(void)
{
  if( flg_status[ZMP_SHAPING_PREVIEW] == true ){
    if( ws != STAND_ANKLE ){
      if( (int)zmp_err[X].size()-Pcog[X].Nprev >= n_clock-Pcog[X].n_start){
	for( int i = 0 ; i < Pcog[X].Nprev ; i++ ){
	  Pcog[X].setReference(zmp_err[X](i+n_clock-Pcog[X].n_start), i+n_clock+1);
	  Pcog[Y].setReference(zmp_err[Y](i+n_clock-Pcog[Y].n_start), i+n_clock+1);
	}
      }
      else{
	for( int i = 0 ; i < Pcog[X].Nprev ; i++ ){
	  Pcog[X].ref = dzerovector(Pcog[X].ref.size());
	  Pcog[Y].ref = dzerovector(Pcog[Y].ref.size());
	}
	
	//if( verbose )
	//cerr << "Timeout of preview control = " << (double)n_clock*DT << endl;
      }
      
      Pcog[X].update(COG_shp(X), VCOG_shp(X), VCOGdot_shp(X), Zmp_shp(X), n_clock);
      Pcog[Y].update(COG_shp(Y), VCOG_shp(Y), VCOGdot_shp(Y), Zmp_shp(Y), n_clock);
    }
  }
  else{
    COG_shp  = 0.0;
    VCOG_shp = 0.0;
    VCOGdot_shp = 0.0;
    Zmp_shp  = 0.0;
    VZmp_shp = 0.0;
  }
}

void MwalkGenerator::walkingState_STAND(WalkingState& ws_next, int& n_period, int& n_period_rest)
{
  static WalkingState ws_next_ = STAND_ANKLE;
  
  const unsigned int FQsize = FootQueue.size();  
  if( n_period_rest == 0 ){
    setFootPlace(n_comp);
    if( FootQueue.empty() ){
      ws_next_ = STAND_ANKLE;
      
      if( ws_pre != STAND_ANKLE )
	n_period = Nss;
      else
	n_period = 1;
      
      setWaistHeightParam(0.0, Foot_cmd[RLEG].P(X)-Foot_cmd[LLEG].P(X), n_clock, 1);
      setSwingArmParam(ws_next, FQsize, 0.0, n_clock, 1);
    }
    else{
      n_period = FootQueue[0].nss;
      
      int sup_leg = FootQueue[0].support_leg;
      int swg_leg = (sup_leg == RLEG) ? LLEG : RLEG;
      
      if( FootQueue[0].support_leg != RLEG ){
	ws_next_ = CENTER_TO_LEFT_ANKLE;
	setWaistHeightParam(0.0, Foot_cmd[LLEG].P(X)-Foot_cmd[RLEG].P(X), n_clock, n_period);
      }
      else{
	ws_next_ = CENTER_TO_RIGHT_ANKLE;
	setWaistHeightParam(0.0, Foot_cmd[RLEG].P(X)-Foot_cmd[LLEG].P(X), n_clock, n_period);
      }
      setSwingArmParam(ws_next, FQsize, FootQueue[0].p(X), n_clock, n_period);
      
#ifdef ORIGIN_PATH
      FootMotionGenerator::setPivotParam(FootQueue[0].p, swg_leg, sup_leg, ws, n_clock, n_period); 
#else
      vector3 foot_step(tvmet::trans(Foot_cmd[sup_leg].R)*(Foot_cmd[sup_leg].P - Foot_cmd[swg_leg].P) + FootQueue[0].p);
      FootMotionGenerator::setPivotParam(foot_step, swg_leg, sup_leg, ws, n_start, n_period); 
#endif
      
      n_period += n_comp;
    }
    n_period_rest = n_period;
    
    makeCOGCoef(COG_cmd, VCOG_cmd, Zmp_cmd, VZmp_cmd, n_period, n_clock);

    double yaw_ini = 0.0;
    vector3 step_ini(0.0);
    FootMotionGenerator::setSwingRpyParam(step_ini, yaw_ini, RLEG, FQsize, n_clock, n_period);
    FootMotionGenerator::setSwingRpyParam(step_ini, yaw_ini, LLEG, FQsize, n_clock, n_period);
    FootMotionGenerator::setAnkleParam(Foot_cmd, Foot_next_cmd, n_clock, n_period);
    FootMotionGenerator::Reset(n_clock, n_period);
  }
  else
    makeCOGCoef(COG_cmd, VCOG_cmd, Zmp_cmd, VZmp_cmd, n_period_rest, n_clock);
  
  vector3 Foot_distance(Foot_cmd[LLEG].P - Foot_cmd[RLEG].P);
  ANcog.setInitialFootDistance(Foot_distance);
  
  n_period_rest--;
  if( n_period_rest == 0 ){
    if( ws_next_ == CENTER_TO_LEFT_ANKLE )
      setOriginForFirstStep(LLEG);
    else if( ws_next_ == CENTER_TO_RIGHT_ANKLE )
      setOriginForFirstStep(RLEG);    
    
    ws_next = ws_next_;
#ifdef WITH_TOE_JOINT
    ws_last = ws;
#endif
  }
}

void MwalkGenerator::walkingState_TO_LEFT(WalkingState& ws_next, int& n_period, int& n_period_rest)
{
  if( n_period_rest == 0 ){
    n_period = FootQueue[0].nds;
    n_period_rest = n_period;
  }
  
  if( n_period_rest == static_cast<int>(TR_ankle_ds_ratio[1]*FootQueue[0].nds) && flg_status[ANKLE_DS_ROTATION] ){
    unsigned int FQsize = FootQueue.size();
    if( FQsize >= 1 ){
      int n_start_ankle = n_clock;
      int n_period_ankle = n_period_rest + FootQueue[0].nss;

      if( FQsize > 1 )
	n_period_ankle += static_cast<int>(TR_ankle_ds_ratio[0]*FootQueue[1].nds);
      else
	n_period_ankle += Nds;
      
#ifdef ORIGIN_PATH
      setLandingPosition(Foot_next_cmd[RLEG]);
#else
      setLandingPosition(Foot_next_cmd[RLEG], Foot_cmd[LLEG]);
#endif
      double  foot_yaw = Foot_next_cmd[RLEG].R_rpy(2) - Foot_cmd[RLEG].R_rpy(2);
      vector3 foot_step(Foot_next_cmd[RLEG].P - Foot_cmd[RLEG].P);
      FootMotionGenerator::setPivotParam(foot_step, LLEG, RLEG, ws, n_start_ankle, n_period_ankle);
      FootMotionGenerator::setSwingRpyParam(foot_step, foot_yaw, RLEG, FQsize, n_start_ankle, n_period_ankle);
    }
  }
  
  //if( !flg_status[ONE_STEP] )
  //makeCOGCoef(COG_cmd, VCOG_cmd, Zmp_cmd, VZmp_cmd, n_period_rest, n_clock);
  
  n_period_rest--;
  if( n_period_rest == 0 ){
#ifndef WITH_TOE_JOINT
    ws_next = LEFT_SUPPORT_ANKLE;
#else
    
#ifdef ORIGIN_PATH
    if( FootQueue[0].p(X) >= 0.0 )
      ws_next = LEFT_SUPPORT_ANKLE;
    else
      ws_next = LEFT_SUPPORT_TOE;
#else
    vector3 foot_step(Foot_cmd[LLEG].R*FootQueue[0].p);
    if( foot_step(X) >= 0.0 )
      ws_next = LEFT_SUPPORT_ANKLE;
    else
      ws_next = LEFT_SUPPORT_TOE;
#endif

    ws_last = ws;
#endif
  }
}

void MwalkGenerator::walkingState_TO_RIGHT(WalkingState& ws_next, int& n_period, int& n_period_rest)
{
  if( n_period_rest == 0 ){
    n_period = FootQueue[0].nds;
    n_period_rest = n_period;
  }
  
  if( n_period_rest == static_cast<int>(TR_ankle_ds_ratio[1]*FootQueue[0].nds) && flg_status[ANKLE_DS_ROTATION] ){
    unsigned int FQsize = FootQueue.size();
    if( FQsize >= 1 ){
      int n_start_ankle = n_clock;
      int n_period_ankle = n_period_rest + FootQueue[0].nss;
      
      if( FQsize > 1 )
	n_period_ankle += static_cast<int>(TR_ankle_ds_ratio[0]*FootQueue[1].nds);
      else
	n_period_ankle += Nds;
      
#ifdef ORIGIN_PATH
      setLandingPosition(Foot_next_cmd[LLEG]);
#else
      setLandingPosition(Foot_next_cmd[LLEG], Foot_cmd[RLEG]);
#endif
      double  foot_yaw = Foot_next_cmd[LLEG].R_rpy(2) - Foot_cmd[LLEG].R_rpy(2);
      vector3 foot_step(Foot_next_cmd[LLEG].P - Foot_cmd[LLEG].P);
      FootMotionGenerator::setPivotParam(foot_step, RLEG, LLEG, ws, n_start_ankle, n_period_ankle);
      FootMotionGenerator::setSwingRpyParam(foot_step, foot_yaw, LLEG, FQsize, n_start_ankle, n_period_ankle);
    }
  }
  
  //if( !flg_status[ONE_STEP] )
  //makeCOGCoef(COG_cmd, VCOG_cmd, Zmp_cmd, VZmp_cmd, n_period_rest, n_clock);
  
  n_period_rest--;
  if( n_period_rest == 0 ){
#ifndef WITH_TOE_JOINT
    ws_next = RIGHT_SUPPORT_ANKLE;
#else
    
#ifdef ORIGIN_PATH
    if( FootQueue[0].p(X) >= 0.0 )
      ws_next = RIGHT_SUPPORT_ANKLE;
    else
      ws_next = RIGHT_SUPPORT_TOE;
#else
    vector3 foot_step(Foot_cmd[RLEG].R*FootQueue[0].p);
    if( foot_step(X) >= 0.0 )
      ws_next = RIGHT_SUPPORT_ANKLE;
    else
      ws_next = RIGHT_SUPPORT_TOE;
#endif
    
    ws_last = ws;
#endif
  }
}

void MwalkGenerator::walkingState_TO_CENTER(WalkingState& ws_next, int& n_period, int& n_period_rest)
{
  if( n_period_rest == 0 ){
    n_period = Nds;
    n_period_rest = n_period;
  }
  
  //if( !flg_status[ONE_STEP] )
  //makeCOGCoef(COG_cmd, VCOG_cmd, Zmp_cmd, VZmp_cmd, n_period_rest, n_clock);
  
  n_period_rest--;
  if( n_period_rest == 0 ){
    if( FootQueue.empty() )
      ws_next = STAND_ANKLE;
    else if( FootQueue[0].support_leg != RLEG )
      ws_next = CENTER_TO_LEFT_ANKLE;
    else
      ws_next = CENTER_TO_RIGHT_ANKLE;
#ifdef WITH_TOE_JOINT
    ws_last = ws_pre;
#endif
  }
}

#ifndef WITH_TOE_JOINT

void MwalkGenerator::walkingState_LEFT_SUPPORT(WalkingState& ws_next, int& n_period, int& n_period_rest)
{
  const unsigned int FQsize = FootQueue.size();
  if( n_period_rest == 0 ){
    setFootPlace(n_comp);
    FootQueue[0].nss += n_comp;
    n_period = FootQueue[0].nss;
    n_period_rest = n_period;
    
    makeCOGCoef(COG_cmd, VCOG_cmd, Zmp_cmd, VZmp_cmd, n_period, n_clock);
    
    if( flg_status[REACTIVE_MOTION] )
      makeReactiveMotion();
    
#ifdef ORIGIN_PATH
    setLandingPosition(Foot_next_cmd[RLEG]);
#else
    setLandingPosition(Foot_next_cmd[RLEG], Foot_cmd[LLEG]);
#endif
    setWaistHeightParam(FootQueue[0], Foot_cmd[LLEG], Foot_cmd[RLEG], n_clock);
    setSwingArmParam(ws_next, FQsize, FootQueue[0].p(X), n_clock, n_period);
    FootMotionGenerator::setSwingPosParam(Foot_cmd[RLEG], Foot_next_cmd[RLEG], n_clock, n_period, FootQueue[0].nss);
    
    if( !flg_status[ANKLE_DS_ROTATION] ){
      double foot_yaw = Foot_next_cmd[RLEG].R_rpy(2) - Foot_cmd[RLEG].R_rpy(2);
      vector3 foot_step(Foot_next_cmd[RLEG].P - Foot_cmd[RLEG].P);
      FootMotionGenerator::setPivotParam(foot_step, LLEG, RLEG, ws, n_clock, n_period); 
      FootMotionGenerator::setSwingRpyParam(foot_step, foot_yaw, RLEG, FQsize, n_clock, n_period);
    }
    
    FootMotionGenerator::setAnkleParam(Foot_cmd, Foot_next_cmd, n_clock, n_period);
  }
  else if( n_period_rest >= FootQueue[0].nss - Nchange(FootQueue[0].nss) ){
    if( !flg_status[ONE_STEP] ){
      setFootPlace(n_comp);
      FootQueue[0].nss += n_comp;
      n_period += n_comp;
      n_period_rest += n_comp;
      
      makeCOGCoef(COG_cmd, VCOG_cmd, Zmp_cmd, VZmp_cmd, n_period_rest, n_clock);

      if( flg_status[REACTIVE_MOTION] )
	makeReactiveMotion();
    }
    
#ifdef ORIGIN_PATH
    setLandingPosition(Foot_next_cmd[RLEG]);
#else
    setLandingPosition(Foot_next_cmd[RLEG], Foot_cmd[LLEG]);
#endif

    FootMotionGenerator::setSwingVerticalPosParam(Foot_cmd[RLEG], Foot_next_cmd[RLEG], n_clock, n_period_rest, FootQueue[0].nss);
    if( flg_status[FOOT_MODIFICATION] )
      FootMotionGenerator::setSwingHorizontalPosParam(Foot_cmd[RLEG], Foot_next_cmd[RLEG], n_clock, n_period_rest);
  }
  
#if 0
  else{
    if( !flg_status[ONE_STEP] )
      makeCOGCoef(COG_cmd, VCOG_cmd, Zmp_cmd, VZmp_cmd, n_period_rest, n_clock);
  }
#endif
  
  n_period_rest--;
  if( n_period_rest == 0 ){
    updateFootPlace(Foot_cmd[RLEG], Foot_next_cmd[RLEG]);
    
    if( FootQueue.empty() )
      ws_next = LEFT_TO_CENTER_ANKLE;
    else if( FootQueue[0].support_leg != RLEG )
      ws_next = KEEP_LEFT_ANKLE;
    else
      ws_next = LEFT_TO_RIGHT_ANKLE;
  }
}

void MwalkGenerator::walkingState_RIGHT_SUPPORT(WalkingState& ws_next, int& n_period, int& n_period_rest)
{
  const unsigned int FQsize = FootQueue.size();
  if( n_period_rest == 0 ){
    setFootPlace(n_comp);
    FootQueue[0].nss += n_comp;    
    n_period = FootQueue[0].nss;
    n_period_rest = n_period;
    
    makeCOGCoef(COG_cmd, VCOG_cmd, Zmp_cmd, VZmp_cmd, n_period, n_clock);
    
    if( flg_status[REACTIVE_MOTION] )
      makeReactiveMotion();
    
#ifdef ORIGIN_PATH
    setLandingPosition(Foot_next_cmd[LLEG]);
#else
    setLandingPosition(Foot_next_cmd[LLEG], Foot_cmd[RLEG]);
#endif
    setWaistHeightParam(FootQueue[0], Foot_cmd[RLEG], Foot_cmd[LLEG], n_clock);
    setSwingArmParam(ws_next, FQsize, FootQueue[0].p(X), n_clock, n_period);
    FootMotionGenerator::setSwingPosParam(Foot_cmd[LLEG], Foot_next_cmd[LLEG], n_clock, n_period, FootQueue[0].nss);
    
    if( !flg_status[ANKLE_DS_ROTATION] ){
      double foot_yaw = Foot_next_cmd[LLEG].R_rpy(2) - Foot_cmd[LLEG].R_rpy(2);
      vector3 foot_step(Foot_next_cmd[LLEG].P - Foot_cmd[LLEG].P);
      FootMotionGenerator::setPivotParam(foot_step, RLEG, LLEG, ws, n_clock, n_period); 
      FootMotionGenerator::setSwingRpyParam(foot_step, foot_yaw, LLEG, FQsize, n_clock, n_period);
    }
    
    FootMotionGenerator::setAnkleParam(Foot_cmd, Foot_next_cmd, n_clock, n_period);
  }
  else if( n_period_rest >= FootQueue[0].nss - Nchange(FootQueue[0].nss) ){
    if( !flg_status[ONE_STEP] ){
      setFootPlace(n_comp);
      FootQueue[0].nss += n_comp;
      n_period += n_comp;
      n_period_rest += n_comp;
      
      makeCOGCoef(COG_cmd, VCOG_cmd, Zmp_cmd, VZmp_cmd, n_period_rest, n_clock);
      
      if( flg_status[REACTIVE_MOTION] )
	makeReactiveMotion();
    }
    
#ifdef ORIGIN_PATH
    setLandingPosition(Foot_next_cmd[LLEG]);
#else
    setLandingPosition(Foot_next_cmd[LLEG], Foot_cmd[RLEG]);
#endif
    
    FootMotionGenerator::setSwingVerticalPosParam(Foot_cmd[LLEG], Foot_next_cmd[LLEG], n_clock, n_period_rest, FootQueue[0].nss);
    if( flg_status[FOOT_MODIFICATION] )
      FootMotionGenerator::setSwingHorizontalPosParam(Foot_cmd[LLEG], Foot_next_cmd[LLEG], n_clock, n_period_rest);
  }

#if 0
  else{
    if( !flg_status[ONE_STEP] )
      makeCOGCoef(COG_cmd, VCOG_cmd, Zmp_cmd, VZmp_cmd, n_period_rest, n_clock);
  }
#endif
  
  n_period_rest--;
  if( n_period_rest == 0 ){
    updateFootPlace(Foot_cmd[LLEG], Foot_next_cmd[LLEG]);
    
    if( FootQueue.empty() )
      ws_next = RIGHT_TO_CENTER_ANKLE;
    else if( FootQueue[0].support_leg != RLEG )
      ws_next = RIGHT_TO_LEFT_ANKLE;
    else
      ws_next = KEEP_RIGHT_ANKLE;
  }
}

#else // #ifndef WITH_TOE_JOINT

void MwalkGenerator::walkingState_LEFT_SUPPORT_ANKLE(WalkingState& ws_next, int& n_period, int& n_period_rest)
{
  const unsigned int FQsize = FootQueue.size();      
  if( ws_last == LEFT_SUPPORT_TOE_TO_ANKLE ){
    if( n_period_rest == 0 ){
      n_period = FootQueue[0].nss - NPzmp_toe_to_ankle(FootQueue[0].nss, 1);
      n_period_rest = n_period;
      makeCOGCoef(COG_cmd, VCOG_cmd, Zmp_cmd, VZmp_cmd, n_period, n_clock);  
    }
    
    if( n_period_rest >= FootQueue[0].nss - Nchange(FootQueue[0].nss) )
      FootMotionGenerator::setSwingPosParam(Foot_cmd[RLEG], Foot_next_cmd[RLEG], n_clock, n_period_rest, FootQueue[0].nss);
  }
  else{
    if( n_period_rest == 0 ){
      n_period = NPzmp_ankle_to_toe(FootQueue[0].nss, 0);
      n_period_rest = n_period;
      
      makeCOGCoef(COG_cmd, VCOG_cmd, Zmp_cmd, VZmp_cmd, n_period, n_clock);
#ifdef ORIGIN_PATH
      setLandingPosition(Foot_next_cmd[RLEG]);
#else
      setLandingPosition(Foot_next_cmd[RLEG], Foot_cmd[LLEG]);
#endif
      setWaistHeightParam(FootQueue[0], Foot_cmd[LLEG], Foot_cmd[RLEG], n_clock);
      setSwingArmParam(ws_next, FQsize, FootQueue[0].p(X), n_clock, n_period);
      FootMotionGenerator::setSwingPosParam(Foot_cmd[RLEG], Foot_next_cmd[RLEG], n_clock, FootQueue[0].nss, FootQueue[0].nss);
      FootMotionGenerator::setToeParam(LLEG, RLEG, FQsize, n_clock, FootQueue[0].nss);
      
      if( !flg_status[ANKLE_DS_ROTATION] ){
	double foot_yaw = Foot_next_cmd[RLEG].R_rpy(2) - Foot_cmd[RLEG].R_rpy(2);
	vector3 foot_step(Foot_next_cmd[RLEG].P - Foot_cmd[RLEG].P);
	FootMotionGenerator::setPivotParam(foot_step, LLEG, RLEG, ws, n_clock, n_period);
	FootMotionGenerator::setSwingRpyParam(foot_step, foot_yaw, RLEG, FQsize, n_clock, FootQueue[0].nss);
      }
      
      FootMotionGenerator::setAnkleParam(Foot_cmd, Foot_next_cmd, n_clock, FootQueue[0].nss);
    }
    else if( n_period_rest >= NPzmp_ankle_to_toe(FootQueue[0].nss, 0) - Nchange(FootQueue[0].nss) )
      FootMotionGenerator::setSwingPosParam(Foot_cmd[RLEG], Foot_next_cmd[RLEG], n_clock, n_period_rest + FootQueue[0].nss - NPzmp_ankle_to_toe(FootQueue[0].nss, 0), FootQueue[0].nss);
  }
  
  n_period_rest--;
  if( n_period_rest == 0 ){
    if( ws_last == LEFT_SUPPORT_TOE_TO_ANKLE ){
      updateFootPlace(Foot_cmd[RLEG], Foot_next_cmd[RLEG]);
      
      if( FootQueue.empty() )
	ws_next = LEFT_TO_CENTER_ANKLE;
      else if( FootQueue[0].support_leg != RLEG )
        ws_next = KEEP_LEFT_ANKLE;
      else
	ws_next = LEFT_TO_RIGHT_ANKLE;
    }
    else{
      ws_next = LEFT_SUPPORT_ANKLE_TO_TOE;
    }
    
    ws_last = ws;
  }
}

void MwalkGenerator::walkingState_LEFT_SUPPORT_ANKLE_TO_TOE(WalkingState& ws_next, int& n_period, int& n_period_rest)
{
  if( n_period_rest == 0 ){
    n_period = NPzmp_ankle_to_toe(FootQueue[0].nss, 1) - NPzmp_ankle_to_toe(FootQueue[0].nss, 0);
    n_period_rest = n_period;
  }
  
  if( n_period_rest >= NPzmp_ankle_to_toe(FootQueue[0].nss, 1) - Nchange(FootQueue[0].nss) )
    FootMotionGenerator::setSwingPosParam(Foot_cmd[RLEG], Foot_next_cmd[RLEG], n_clock, n_period_rest + FootQueue[0].nss - NPzmp_ankle_to_toe(FootQueue[0].nss, 1), FootQueue[0].nss);
  
  //if( !flg_status[ONE_STEP] )
  //makeCOGCoef(COG_cmd, VCOG_cmd, Zmp_cmd, VZmp_cmd, n_period_rest, n_clock);
  
  n_period_rest--;
  if( n_period_rest == 0 ){
    ws_next = LEFT_SUPPORT_TOE;
    ws_last = ws;
  }
}

void MwalkGenerator::walkingState_LEFT_SUPPORT_TOE_TO_ANKLE(WalkingState& ws_next, int& n_period, int& n_period_rest)
{
  if( n_period_rest == 0 ){
    n_period = NPzmp_toe_to_ankle(FootQueue[0].nss, 1) - NPzmp_toe_to_ankle(FootQueue[0].nss, 0);
    n_period_rest = n_period;
  }
  
  if( n_period_rest >= NPzmp_toe_to_ankle(FootQueue[0].nss, 1) - Nchange(FootQueue[0].nss) )
    FootMotionGenerator::setSwingPosParam(Foot_cmd[RLEG], Foot_next_cmd[RLEG], n_clock, n_period_rest + FootQueue[0].nss - NPzmp_toe_to_ankle(FootQueue[0].nss, 1), FootQueue[0].nss);
  
  //if( !flg_status[ONE_STEP] )
  //makeCOGCoef(COG_cmd, VCOG_cmd, Zmp_cmd, VZmp_cmd, n_period_rest, n_clock);
  
  n_period_rest--;
  if( n_period_rest == 0 ){
    ws_next = LEFT_SUPPORT_ANKLE;
    ws_last = ws;
  }
}

void MwalkGenerator::walkingState_LEFT_SUPPORT_TOE(WalkingState& ws_next, int& n_period, int& n_period_rest)
{
  const unsigned int FQsize = FootQueue.size();
  if( ws_last == LEFT_SUPPORT_ANKLE_TO_TOE ){
    if( n_period_rest == 0 ){
      n_period = FootQueue[0].nss - NPzmp_ankle_to_toe(FootQueue[0].nss, 1);
      n_period_rest = n_period;
      makeCOGCoef(COG_cmd, VCOG_cmd, Zmp_cmd, VZmp_cmd, n_period, n_clock);
    }
    
    if( n_period_rest >= FootQueue[0].nss - Nchange(FootQueue[0].nss) )
      FootMotionGenerator::setSwingPosParam(Foot_cmd[RLEG], Foot_next_cmd[RLEG], n_clock, n_period_rest, FootQueue[0].nss);
  }
  else{
    if( n_period_rest == 0 ){
      n_period = NPzmp_toe_to_ankle(FootQueue[0].nss, 0);
      n_period_rest = n_period;
      makeCOGCoef(COG_cmd, VCOG_cmd, Zmp_cmd, VZmp_cmd, n_period, n_clock);
#ifdef ORIGIN_PATH
      setLandingPosition(Foot_next_cmd[RLEG]);
#else
      setLandingPosition(Foot_next_cmd[RLEG], Foot_cmd[LLEG]);
#endif
      setWaistHeightParam(FootQueue[0], Foot_cmd[LLEG], Foot_cmd[RLEG], n_clock);
      setSwingArmParam(ws_next, FQsize, FootQueue[0].p(X), n_clock, n_period);
      FootMotionGenerator::setSwingPosParam(Foot_cmd[RLEG], Foot_next_cmd[RLEG], n_clock, FootQueue[0].nss, FootQueue[0].nss);
      FootMotionGenerator::setToeParam(LLEG, RLEG, FQsize, n_clock, FootQueue[0].nss);
      
      if( !flg_status[ANKLE_DS_ROTATION] ){
	double foot_yaw = Foot_next_cmd[RLEG].R_rpy(2) - Foot_cmd[RLEG].R_rpy(2);
	vector3 foot_step(Foot_next_cmd[RLEG].P - Foot_cmd[RLEG].P);
	FootMotionGenerator::setPivotParam(foot_step, LLEG, RLEG, ws, n_clock, n_period);
	FootMotionGenerator::setSwingRpyParam(foot_step, foot_yaw, RLEG, FQsize, n_clock, FootQueue[0].nss);
      }
      
      FootMotionGenerator::setAnkleParam(Foot_cmd, Foot_next_cmd, n_clock, FootQueue[0].nss);
    }
    else if( n_period_rest >= NPzmp_toe_to_ankle(FootQueue[0].nss, 0) - Nchange(FootQueue[0].nss) )
      FootMotionGenerator::setSwingPosParam(Foot_cmd[RLEG], Foot_next_cmd[RLEG], n_clock, n_period_rest + FootQueue[0].nss - NPzmp_toe_to_ankle(FootQueue[0].nss, 0), FootQueue[0].nss);
  }
  
  n_period_rest--;
  if( n_period_rest == 0 ){
    if( ws_last == LEFT_SUPPORT_ANKLE_TO_TOE ){
      updateFootPlace(Foot_cmd[RLEG], Foot_next_cmd[RLEG]);
      
      if( FootQueue.empty() )
	ws_next = LEFT_TO_CENTER_ANKLE;
      else if( FootQueue[0].support_leg != RLEG )
	ws_next = KEEP_LEFT_ANKLE;
      else
	ws_next = LEFT_TO_RIGHT_ANKLE;
    }
    else{
      ws_next = LEFT_SUPPORT_TOE_TO_ANKLE;
    }

    ws_last = ws;
  }
}

void MwalkGenerator::walkingState_RIGHT_SUPPORT_ANKLE(WalkingState& ws_next, int& n_period, int& n_period_rest)
{
  const unsigned int FQsize = FootQueue.size();
  if( ws_last == RIGHT_SUPPORT_TOE_TO_ANKLE ){
    if( n_period_rest == 0 ){
      n_period = FootQueue[0].nss - NPzmp_toe_to_ankle(FootQueue[0].nss, 1);
      n_period_rest = n_period;
      makeCOGCoef(COG_cmd, VCOG_cmd, Zmp_cmd, VZmp_cmd, n_period, n_clock);  
    }
    
    if( n_period_rest >= FootQueue[0].nss - Nchange(FootQueue[0].nss) )
      FootMotionGenerator::setSwingPosParam(Foot_cmd[LLEG], Foot_next_cmd[LLEG], n_clock, n_period_rest, FootQueue[0].nss);
  }
  else{
    if( n_period_rest == 0 ){
      n_period = NPzmp_ankle_to_toe(FootQueue[0].nss, 0);
      n_period_rest = n_period;

      makeCOGCoef(COG_cmd, VCOG_cmd, Zmp_cmd, VZmp_cmd, n_period, n_clock);
#ifdef ORIGIN_PATH
      setLandingPosition(Foot_next_cmd[LLEG]);
#else
      setLandingPosition(Foot_next_cmd[LLEG], Foot_cmd[RLEG]);
#endif
      setWaistHeightParam(FootQueue[0], Foot_cmd[RLEG], Foot_cmd[LLEG], n_clock);
      setSwingArmParam(ws_next, FQsize, FootQueue[0].p(X), n_clock, n_period);
      FootMotionGenerator::setSwingPosParam(Foot_cmd[LLEG], Foot_next_cmd[LLEG], n_clock, FootQueue[0].nss, FootQueue[0].nss);
      FootMotionGenerator::setToeParam(RLEG, LLEG, FQsize, n_clock, FootQueue[0].nss);
      
      if( !flg_status[ANKLE_DS_ROTATION] ){
	double foot_yaw = Foot_next_cmd[LLEG].R_rpy(2) - Foot_cmd[LLEG].R_rpy(2);
	vector3 foot_step(Foot_next_cmd[LLEG].P - Foot_cmd[LLEG].P);
	FootMotionGenerator::setPivotParam(foot_step, RLEG, LLEG, ws, n_clock, n_period); 
	FootMotionGenerator::setSwingRpyParam(foot_step, foot_yaw, LLEG, FQsize, n_clock, FootQueue[0].nss);
      }
      
      FootMotionGenerator::setAnkleParam(Foot_cmd, Foot_next_cmd, n_clock, FootQueue[0].nss);
    }
    else if( n_period_rest >= NPzmp_ankle_to_toe(FootQueue[0].nss, 0) - Nchange(FootQueue[0].nss) )
      FootMotionGenerator::setSwingPosParam(Foot_cmd[LLEG], Foot_next_cmd[LLEG], n_clock, n_period_rest + FootQueue[0].nss - NPzmp_ankle_to_toe(FootQueue[0].nss, 0), FootQueue[0].nss);
  }
  
  n_period_rest--;
  if( n_period_rest == 0 ){
    if( ws_last == RIGHT_SUPPORT_TOE_TO_ANKLE ){
      updateFootPlace(Foot_cmd[LLEG], Foot_next_cmd[LLEG]);
      
      if( FootQueue.empty() )
	ws_next = RIGHT_TO_CENTER_ANKLE;
      else if( FootQueue[0].support_leg != LLEG )
        ws_next = KEEP_RIGHT_ANKLE;
      else
	ws_next = RIGHT_TO_LEFT_ANKLE;
    }
    else{
      ws_next = RIGHT_SUPPORT_ANKLE_TO_TOE;
    }
    
    ws_last = ws;
  }
}

void MwalkGenerator::walkingState_RIGHT_SUPPORT_ANKLE_TO_TOE(WalkingState& ws_next, int& n_period, int& n_period_rest)
{
  if( n_period_rest == 0 ){
    n_period = NPzmp_ankle_to_toe(FootQueue[0].nss, 1) - NPzmp_ankle_to_toe(FootQueue[0].nss, 0);
    n_period_rest = n_period;
  }
  
  if( n_period_rest >= NPzmp_ankle_to_toe(FootQueue[0].nss, 1) - Nchange(FootQueue[0].nss) )
    FootMotionGenerator::setSwingPosParam(Foot_cmd[LLEG], Foot_next_cmd[LLEG], n_clock, n_period_rest + FootQueue[0].nss - NPzmp_ankle_to_toe(FootQueue[0].nss, 1), FootQueue[0].nss);
  
  //if( !flg_status[ONE_STEP] )
  //makeCOGCoef(COG_cmd, VCOG_cmd, Zmp_cmd, VZmp_cmd, n_period_rest, n_clock);
  
  n_period_rest--;
  if( n_period_rest == 0 ){
    ws_next = RIGHT_SUPPORT_TOE;
    ws_last = ws;
  }
}

void MwalkGenerator::walkingState_RIGHT_SUPPORT_TOE_TO_ANKLE(WalkingState& ws_next, int& n_period, int& n_period_rest)
{
  if( n_period_rest == 0 ){
    n_period = NPzmp_toe_to_ankle(FootQueue[0].nss, 1) - NPzmp_toe_to_ankle(FootQueue[0].nss, 0); 
    n_period_rest = n_period;
  }
  
  if( n_period_rest >= NPzmp_toe_to_ankle(FootQueue[0].nss, 1) - Nchange(FootQueue[0].nss) )
    FootMotionGenerator::setSwingPosParam(Foot_cmd[LLEG], Foot_next_cmd[LLEG], n_clock, n_period_rest + FootQueue[0].nss - NPzmp_toe_to_ankle(FootQueue[0].nss, 1), FootQueue[0].nss);
  
  //if( !flg_status[ONE_STEP] )
  //makeCOGCoef(COG_cmd, VCOG_cmd, Zmp_cmd, VZmp_cmd, n_period_rest, n_clock);  
  
  n_period_rest--;
  if( n_period_rest == 0 ){
    ws_next = RIGHT_SUPPORT_ANKLE;
    ws_last = ws;
  }
}

void MwalkGenerator::walkingState_RIGHT_SUPPORT_TOE(WalkingState& ws_next, int& n_period, int& n_period_rest)
{
  const unsigned int FQsize = FootQueue.size();
  if( ws_last == RIGHT_SUPPORT_ANKLE_TO_TOE ){
    if( n_period_rest == 0 ){
      n_period = FootQueue[0].nss - NPzmp_ankle_to_toe(FootQueue[0].nss, 1);
      n_period_rest = n_period;
      makeCOGCoef(COG_cmd, VCOG_cmd, Zmp_cmd, VZmp_cmd, n_period, n_clock);
    }
    
    if( n_period >= FootQueue[0].nss - Nchange(FootQueue[0].nss) )
      FootMotionGenerator::setSwingPosParam(Foot_cmd[LLEG], Foot_next_cmd[LLEG], n_clock, n_period, FootQueue[0].nss);
  }
  else{
    if( n_period_rest == 0 ){
      n_period = NPzmp_toe_to_ankle(FootQueue[0].nss, 0);
      n_period_rest = n_period;
      makeCOGCoef(COG_cmd, VCOG_cmd, Zmp_cmd, VZmp_cmd, n_period, n_clock);
#ifdef ORIGIN_PATH
      setLandingPosition(Foot_next_cmd[LLEG]);
#else
      setLandingPosition(Foot_next_cmd[LLEG], Foot_cmd[RLEG]);
#endif
      setWaistHeightParam(FootQueue[0], Foot_cmd[RLEG], Foot_cmd[LLEG], n_clock);
      setSwingArmParam(ws_next, FQsize, FootQueue[0].p(X), n_clock, n_period);
      FootMotionGenerator::setSwingPosParam(Foot_cmd[LLEG], Foot_next_cmd[LLEG], n_clock, FootQueue[0].nss, FootQueue[0].nss);
      FootMotionGenerator::setToeParam(RLEG, LLEG, FQsize, n_clock, FootQueue[0].nss);
      
      if( !flg_status[ANKLE_DS_ROTATION] ){
	double foot_yaw = Foot_next_cmd[LLEG].R_rpy(2) - Foot_cmd[LLEG].R_rpy(2);
	vector3 foot_step(Foot_next_cmd[LLEG].P - Foot_cmd[LLEG].P);
	FootMotionGenerator::setPivotParam(foot_step, RLEG, LLEG, ws, n_clock, n_period); 
	FootMotionGenerator::setSwingRpyParam(foot_step, foot_yaw, LLEG, FQsize, n_clock, FootQueue[0].nss);
      }
      
      FootMotionGenerator::setAnkleParam(Foot_cmd, Foot_next_cmd, n_clock, FootQueue[0].nss);
    }
    else if( n_period_rest >= NPzmp_toe_to_ankle(FootQueue[0].nss, 0) - Nchange(FootQueue[0].nss) )
      FootMotionGenerator::setSwingPosParam(Foot_cmd[LLEG], Foot_next_cmd[LLEG], n_clock, n_period_rest + FootQueue[0].nss - NPzmp_toe_to_ankle(FootQueue[0].nss, 0), FootQueue[0].nss);
  }
  
  n_period_rest--;
  if( n_period_rest == 0 ){
    if( ws_last == RIGHT_SUPPORT_ANKLE_TO_TOE ){
      updateFootPlace(Foot_cmd[LLEG], Foot_next_cmd[LLEG]);
      
      if( FootQueue.empty() )
	ws_next = RIGHT_TO_CENTER_ANKLE;
      else if( FootQueue[0].support_leg != LLEG )
	ws_next = KEEP_RIGHT_ANKLE;
      else
	ws_next = RIGHT_TO_LEFT_ANKLE;
    }
    else{
      ws_next = RIGHT_SUPPORT_TOE_TO_ANKLE;
    }
    
    ws_last = ws;
  }
}

#endif


void MwalkGenerator::updateWalkingState(WalkingState& ws_next, int& n_period, int& n_period_rest)
{
  ws_pre = ws;
  ws     = ws_next;
  
  if( verbose && ws != ws_pre ){
    cout << "\nT = " << (double)n_clock*DT << " State : " << getWalkingStateString(ws) << endl;
    //cout <<"COG = " << COG_cmd << ", ZMP = " << Zmp_cmd << endl;
    
    if( isLeftSupport(ws_pre) || isRightSupport(ws_pre) )
      cout << "AbsPos(x,y,th) = " << AbsPosOrg(X) << "," << AbsPosOrg(Y) << "," << ToDeg(AbsRpyOrg(Z)) << endl;
  }
  
  switch( ws ){
  case STAND_ANKLE:
    walkingState_STAND(ws_next, n_period, n_period_rest);
    break;
    
  case CENTER_TO_LEFT_ANKLE:
  case RIGHT_TO_LEFT_ANKLE:
  case KEEP_LEFT_ANKLE:
    walkingState_TO_LEFT(ws_next, n_period, n_period_rest);
    break;
    
  case CENTER_TO_RIGHT_ANKLE:
  case LEFT_TO_RIGHT_ANKLE:
  case KEEP_RIGHT_ANKLE:
    walkingState_TO_RIGHT(ws_next, n_period, n_period_rest);
    break;
    
  case LEFT_TO_CENTER_ANKLE:
  case RIGHT_TO_CENTER_ANKLE:
    walkingState_TO_CENTER(ws_next, n_period, n_period_rest);
    break;
    
#ifndef WITH_TOE_JOINT
  case LEFT_SUPPORT_ANKLE:
    walkingState_LEFT_SUPPORT(ws_next, n_period, n_period_rest);
    break;
    
  case RIGHT_SUPPORT_ANKLE:
    walkingState_RIGHT_SUPPORT(ws_next, n_period, n_period_rest);
    break;
#else
  case LEFT_SUPPORT_ANKLE:
    walkingState_LEFT_SUPPORT_ANKLE(ws_next, n_period, n_period_rest);
    break;
    
  case LEFT_SUPPORT_ANKLE_TO_TOE:
  case KEEP_LEFT_ANKLE_TO_TOE:
    walkingState_LEFT_SUPPORT_ANKLE_TO_TOE(ws_next, n_period, n_period_rest);
    break;
    
  case LEFT_SUPPORT_TOE:
    walkingState_LEFT_SUPPORT_TOE(ws_next, n_period, n_period_rest);
    break;
    
  case LEFT_SUPPORT_TOE_TO_ANKLE:
  case KEEP_LEFT_TOE_TO_ANKLE:
    walkingState_LEFT_SUPPORT_TOE_TO_ANKLE(ws_next, n_period, n_period_rest);
    break;
    
  case RIGHT_SUPPORT_ANKLE:
    walkingState_RIGHT_SUPPORT_ANKLE(ws_next, n_period, n_period_rest);
    break;
    
  case RIGHT_SUPPORT_ANKLE_TO_TOE:
  case KEEP_RIGHT_ANKLE_TO_TOE:
    walkingState_RIGHT_SUPPORT_ANKLE_TO_TOE(ws_next, n_period, n_period_rest);
    break;
    
  case RIGHT_SUPPORT_TOE:
    walkingState_RIGHT_SUPPORT_TOE(ws_next, n_period, n_period_rest);
    break;
    
  case RIGHT_SUPPORT_TOE_TO_ANKLE:
  case KEEP_RIGHT_TOE_TO_ANKLE:
    walkingState_RIGHT_SUPPORT_TOE_TO_ANKLE(ws_next, n_period, n_period_rest);
    break;
#endif
    
  default:
    cerr << "Warning!!! Invalid walking state!!!" << endl;
    break;
  }
  
  setZMPShaping();
  calcZMPErrorOfAnalyticalSolution(n_period_rest);
  
  n_clock++;
}

/* --- MwalkGenerator::initConfiguration ----------------------------------- */

void
MwalkGenerator::initConfiguration(
  const double basePos[3], const double baseAtt[9], const double angle[DOF],
  double zmp[3])
{
  // start walking from current posture
  for( int i = 0 ; i < DOF ; i++ ){
    q_cmd[i] = angle[i];
    q_ini[i] = angle[i];
  }
  
  getVector3(Waist_cmd.P, basePos);
  getMatrix33FromRowMajorArray(Waist_cmd.R, baseAtt);
  Waist_cmd.R_rpy = rpyFromRot(Waist_cmd.R);
  waist_yaw_offset = Waist_cmd.R_rpy(2);
  Waist_cmd.R_rpy(2) = 0.;
  Waist_cmd.R = rotFromRpy(Waist_cmd.R_rpy);

  waistLink->p = Waist_cmd.P;
  waistLink->R = Waist_cmd.R;

  for( int i = 0 ; i < DOF ; i++ )
    body->joint(i)->q = q_cmd[i];
  body->calcForwardKinematics();
  
  Foot_cmd[RLEG].P  = ankleLink[RLEG]->p - ankleLink[RLEG]->R * FootToAnkle;
  Foot_cmd[LLEG].P  = ankleLink[LLEG]->p - ankleLink[LLEG]->R * FootToAnkle;
  
  vector3 Foot_ctr((Foot_cmd[LLEG].P + Foot_cmd[RLEG].P) / 2.0);
  AbsPosOrg = Foot_ctr(X), Foot_ctr(Y), 0.0;
  AbsRpyOrg = Waist_cmd.R_rpy;

  Waist_cmd.P(Z) -= Foot_ctr(Z);
  
  waistLink->p = Waist_cmd.P;
  waistLink->R = Waist_cmd.R;
  body->calcForwardKinematics();
  
  COG_cmd  = body->calcCM();
  
  vector3 Xg_offset_cmd(tvmet::trans(Waist_cmd.R) * (COG_cmd - Waist_cmd.P));
  vector3 Xg_offset_diff(Xg_offset_cmd - Xg_offset_ini);
  Xg_offset_ini += Xg_offset_diff * INIT_COG_OFFSET_GAIN * DT;
  
  flg_status[READY_INITCONF] = tvmet::dot(Xg_offset_diff, Xg_offset_diff) < 1.0e-4*1.0e-4 ? true : false;
  
  for( int i = RLEG ; i <= LLEG ; i++ ){
    Ankle_cmd[i].R     = ankleLink[i]->R;
    Ankle_cmd[i].R_rpy = rpyFromRot(Ankle_cmd[i].R);
    Ankle_cmd[i].P     = ankleLink[i]->p;
    
    Foot_cmd[i].R      = Ankle_cmd[i].R;
    Foot_cmd[i].R_rpy  = Ankle_cmd[i].R_rpy;
    Foot_cmd[i].P      = Ankle_cmd[i].P - Ankle_cmd[i].R * FootToAnkle;
    
    Foot_next_cmd[i].R     = Foot_cmd[i].R;
    Foot_next_cmd[i].P     = Foot_cmd[i].P;
    Foot_next_cmd[i].R_rpy = Foot_cmd[i].R_rpy;
  }
  
  vector3 Zmp_ref(0.0);
  getVector3(Zmp_ref, zmp);
  Zmp_cmd = Waist_cmd.R * Zmp_ref + Waist_cmd.P;
  VZmp_cmd = 0.0;
  
  for( int i = 0 ; i < Ncog ; i++ ){
    cog_height[i] = COG_cmd(Z) - Zmp_cmd(Z);
    omega[i] = sqrt(Gv / cog_height[i]);
  }
  ANcog.setCOGHeight(cog_height);
}

void
MwalkGenerator::initState(
  vector3& waist_p, vector3& waist_v, vector3& waist_rpy, vector3& waist_rpy_dot,
  vector<double>& angle, vector<double>& angle_dot, vector<double>& angle_pre,
  vector3& zmp, vector3& vzmp, WalkingState& ws_next, vector3& lin_mom_pre_, vector3& ang_mom_pre_,
  vector3& vgsens_pre_, int n_clock_, bool first_step_)
{
  n_clock = n_clock_;
  lin_mom_pre = lin_mom_pre_;
  ang_mom_pre = ang_mom_pre_;
  Vgsens_pre = vgsens_pre_;
  
  for( int i = 0 ; i < DOF ; i++ ){
    q_cmd[i] = angle[i];
    qdot_cmd[i] = angle_dot[i];
    q_cmd_pre[i] = angle_pre[i];
  }
  
  Waist_cmd.P = waist_p;
  Waist_cmd.V = waist_v;
  Waist_cmd.R = rotFromRpy(waist_rpy);
  Waist_cmd.R_rpy = waist_rpy;
  Waist_cmd.W = omegaFromRpyAngularVelocity(waist_rpy, waist_rpy_dot);
  
  waistLink->p = Waist_cmd.P;
  waistLink->v = Waist_cmd.V;
  waistLink->R = Waist_cmd.R;
  waistLink->w = Waist_cmd.W;
  for( int i = 0 ; i < DOF ; i++ )
    body->joint(i)->q = q_cmd[i];
  body->calcForwardKinematics();

  if( first_step_ ){
    for( int i = 0 ; i < DOF ; i++ )
      q_ini[i] = angle[i];
    Xg_offset_ini = tvmet::trans(Waist_cmd.R)*(body->calcCM() - Waist_cmd.P);
  }
  
  for( int i = RLEG ; i <= LLEG ; i++ ){
    Ankle_cmd[i].R     = ankleLink[i]->R;
    Ankle_cmd[i].R_rpy = rpyFromRot(Ankle_cmd[i].R);
    Ankle_cmd[i].P     = ankleLink[i]->p;
    
    Foot_cmd[i].R      = Ankle_cmd[i].R;
    Foot_cmd[i].R_rpy  = Ankle_cmd[i].R_rpy;
    Foot_cmd[i].P      = Ankle_cmd[i].P - Ankle_cmd[i].R * FootToAnkle;
    
    Foot_next_cmd[i].R     = Foot_cmd[i].R;
    Foot_next_cmd[i].P     = Foot_cmd[i].P;
    Foot_next_cmd[i].R_rpy = Foot_cmd[i].R_rpy;
  }
  
  Xg_offset = Waist_cmd.R * Xg_offset_ini;
  Vg_offset = cross(Waist_cmd.W, Xg_offset); 
  
  COG_cmd  = Waist_cmd.P + Xg_offset;
  VCOG_cmd = Waist_cmd.V + Vg_offset;
  
  Zmp_cmd  = zmp;
  VZmp_cmd = vzmp;
  
  if( first_step_ ){
    ws = STAND_ANKLE;
    ws_next = STAND_ANKLE;
  }
  
  if( FootQueue[0].support_leg == RLEG ){
    AbsPosOrg = Foot_cmd[RLEG].P - Foot_cmd[RLEG].R * Foot_cmd[RLEG].offset;
    AbsRpyOrg = Foot_cmd[RLEG].R_rpy;
  }
  else{
    AbsPosOrg = Foot_cmd[LLEG].P - Foot_cmd[LLEG].R * Foot_cmd[LLEG].offset;
    AbsRpyOrg = Foot_cmd[LLEG].R_rpy;
  }
}


/* --- MwalkGenerator::setStateForZmpfix ----------------------------------- */
void
MwalkGenerator::setStateForZmpfix(
  double basePos[3], double baseAtt[9], double *angle, double zmp[3])
{
  for( int i = 0 ; i < DOF ; i++ )
    q_cmd[i] = angle[i];
  
  getVector3(Waist_cmd.P, basePos);
  getMatrix33FromRowMajorArray(Waist_cmd.R, baseAtt);
  Waist_cmd.R_rpy = rpyFromRot(Waist_cmd.R);
  waist_yaw_offset = Waist_cmd.R_rpy(2);
  Waist_cmd.R_rpy(2) = 0.;
  Waist_cmd.R = rotFromRpy(Waist_cmd.R_rpy);
  
  waistLink->p = Waist_cmd.P;
  waistLink->R = Waist_cmd.R;
  for( int i = 0 ; i < DOF ; i++ )
    body->joint(i)->q = q_cmd[i];
  body->calcForwardKinematics();
  
  for( int i = RLEG ; i <= LLEG ; i++ ){
    Ankle_cmd[i].R = ankleLink[i]->R;
    Ankle_cmd[i].R_rpy = rpyFromRot(Ankle_cmd[i].R);
    Ankle_cmd[i].P = ankleLink[i]->p;
    Foot_cmd[i].R  = Ankle_cmd[i].R;
    Foot_cmd[i].R_rpy  = Ankle_cmd[i].R_rpy;
    Foot_cmd[i].P  = Ankle_cmd[i].P - Ankle_cmd[i].R * FootToAnkle;
    Foot_next_cmd[i].R     = Foot_cmd[i].R;
    Foot_next_cmd[i].P     = Foot_cmd[i].P;
    Foot_next_cmd[i].R_rpy = Foot_cmd[i].R_rpy;
  }
  
  vector3 Zmp_ref(0.0);
  getVector3(Zmp_ref, zmp);
  Zmp_cmd = Waist_cmd.R * Zmp_ref + Waist_cmd.P;
}

void MwalkGenerator::setCOGHeight(deque<vector3>& waist_p, deque<vector3>& waist_rpy)
{
  unsigned int FQsize = FootQueue.size();
  if( FQsize > Nstep )
    FQsize = Nstep;

  int total_step = 0;
  vector3 xg_offset(Xg_offset_ini);
  for( unsigned int i = 0 ; i < FQsize ; i++ ){
    xg_offset = rotFromRpy(waist_rpy[total_step]) * Xg_offset_ini;
    cog_height[i*2] = waist_p[total_step](Z) + xg_offset(Z) - Zmp_cmd(Z);
    omega[i*2] = sqrt(Gv / cog_height[i*2]);
    total_step += FootQueue[i].nss;

    xg_offset = rotFromRpy(waist_rpy[total_step]) * Xg_offset_ini;
    cog_height[i*2+1] = waist_p[total_step](Z) + xg_offset(Z) - Zmp_cmd(Z);
    omega[i*2+1] = sqrt(Gv / cog_height[i*2+1]);
    total_step += FootQueue[i].nds;
  }
  
  for( unsigned int i = FQsize ; i < Nstep ; i++ ){
    xg_offset = rotFromRpy(waist_rpy[total_step]) * Xg_offset_ini;
    cog_height[i*2] = waist_p[total_step](Z) + xg_offset(Z) - Zmp_cmd(Z);
    omega[i*2] = sqrt(Gv / cog_height[i*2]);
    
    xg_offset = rotFromRpy(waist_rpy[total_step]) * Xg_offset_ini;
    cog_height[i*2+1] = waist_p[total_step](Z) + xg_offset(Z) - Zmp_cmd(Z);
    omega[i*2+1] = sqrt(Gv / cog_height[i*2+1]);
  }
  
  xg_offset = rotFromRpy(waist_rpy[total_step]) * Xg_offset_ini;
  cog_height[Nstep*2] = waist_p[total_step](Z) + xg_offset(Z) - Zmp_cmd(Z);
  omega[Nstep*2] = sqrt(Gv / cog_height[Nstep*2]);
  
  ANcog.setCOGHeight(cog_height);
}

void MwalkGenerator::setJointCommand(
  vector3& waist_p, matrix33& waist_R, vector3 *ankle_p, matrix33 *ankle_R,
  vector3& waist_v, vector3& waist_w, Jvector& q_cmd_, Jvector& q_cmd_pre_, Jvector& qdot_cmd_)
{
  for( int leg = RLEG ; leg <= LLEG ; leg++){
    if ( !waistToAnklePath[leg]->calcInverseKinematics(waist_p, waist_R, ankle_p[leg], ankle_R[leg]) && verbose){
      cerr << "IK failed(" << leg << ")" << endl;
      
      //cout << "Waist pos = " << Waist_cmd.P << endl;
      //cout << "ankle pos = " << Ankle_cmd[leg].P << endl;
      //cout << "ankle att = " << Ankle_cmd[leg].R << endl;
    }
  }
  
  if( flg_firsttime[JOINT] ){
    flg_firsttime[JOINT] = false;
    for( int i = 0 ; i < DOF_LEG ; i++ )
      q_cmd_pre_[i] = body->joint(i)->q;
  }

  for( int i = 0 ; i < DOF_LEG ; i++ )
    q_cmd_[i] = body->joint(i)->q;
  
  for( int i = 0 ; i < DOF ; i++ ){
    qdot_cmd_[i]  = (q_cmd_[i] - q_cmd_pre_[i]) / DT;
    q_cmd_pre_[i] = q_cmd_[i];
  }
  
  waistLink->v = waist_v;
  waistLink->w = waist_w;
  
  for( int i = 0 ; i < DOF ; i++ )
    body->joint(i)->dq = qdot_cmd_[i];
}

////////////////////////////////////////////////////////////////////////////////
void MwalkGenerator::calcIdealZMPCommand( vector3& zmp_ideal_ )
{
#ifndef WITH_TOE_JOINT
  if( mmode == WALKING ){
    if( isSingleSupport(ws) ){
      int sup_leg = getSupportLeg(ws);
      vector3 zmp_ini(Foot_cmd[sup_leg].P + Foot_cmd[sup_leg].R * ANcog.Zmp_ankle_ini[sup_leg]);
      zmp_ideal_ = zmp_des[1] + (zmp_ini - zmp_des[1]) * interp_func3((double)n_period_rest/(double)n_period);
    }
    else
      zmp_ideal_ = Zmp_cmd;
  }
  else
    zmp_ideal_ = Zmp_cmd;
#else
  if( mmode == WALKING ){
    if( isSingleSupport(ws) ){
      int sup_leg = getSupportLeg(ws);
      if( isAnkleSupport(ws) ){
	vector3 zmp_ini(Foot_cmd[sup_leg].P + Foot_cmd[sup_leg].R * ANcog.Zmp_ankle_ini[sup_leg]);
	zmp_ideal_ = zmp_des[1] + (zmp_ini - zmp_des[1]) * interp_func3((double)n_period_rest/(double)n_period);
      }
      else if( isToeSupport(ws) ){
	vector3 zmp_ini(Foot_cmd[sup_leg].P + Foot_cmd[sup_leg].R * ANcog.Zmp_toe[sup_leg]);
	zmp_ideal_ = zmp_des[1] + (zmp_ini - zmp_des[1]) * interp_func3((double)n_period_rest/(double)n_period);
      }
      else
	zmp_ideal_ = Zmp_cmd;
    }
    else
      zmp_ideal_ = Zmp_cmd;
  }
  else
    zmp_ideal_ = Zmp_cmd;
#endif
}

void MwalkGenerator::setMotionCommandToBuffer(
  vector3& waist_pos_, matrix33& waist_att_, Jvector& q_cmd_, vector3& zmp_model_,
  double basePos[3], double baseAtt[9], double *angle, double zmp[3], double accel[3])
{
  matrix33 xwaist = waist_att_;
  vector3 xrpy = rpyFromRot(xwaist);
  xrpy(2) += waist_yaw_offset;
  xwaist = rotFromRpy(xrpy);

  setVector3(waist_pos_, basePos);
  setMatrix33ToRowMajorArray(xwaist, baseAtt);
  for( int i = 0 ; i < DOF ; i++ )
    angle[i] = q_cmd_[i];
  
  Zmp_ref = tvmet::trans(waist_att_) * (zmp_model_ - waist_pos_);
  setVector3(Zmp_ref, zmp);
  
  ModelManager::calcGsensorReference(Gsensor_offset, Vgsens_pre, flg_firsttime[GSENS]);
  setVector3(Gsensor_offset, accel);
}

void MwalkGenerator::setMotionCommandWithPreviewControl(
  double basePos[3], double baseAtt[9], double *angle, double zmp[3], double accel[3])
{
  if( waistP_pbuf.empty() == true ){
    Pcog_dzmp[X].n_start = n_clock;
    Pcog_dzmp[Y].n_start = n_clock;
  }
  
  waistP_pbuf.push_back(Waist_cmd.P);
  waistV_pbuf.push_back(Waist_cmd.V);
  waistR_pbuf.push_back(Waist_cmd.R);
  waistW_pbuf.push_back(Waist_cmd.W);
  for(int i = RLEG ; i <= LLEG ; i++){
    ankleP_pbuf[i].push_back(Ankle_cmd[i].P);
    ankleR_pbuf[i].push_back(Ankle_cmd[i].R);
  }
  q_pbuf.push_back(q_cmd);
  zzmp_pbuf.push_back(Zmp_model(Z));
  
  // insert zmp error to the last of buffer
  Pcog_dzmp[X].setReference(Zmp_model(X)-Zmp_ideal(X), n_clock);
  Pcog_dzmp[Y].setReference(Zmp_model(Y)-Zmp_ideal(Y), n_clock);
  
  vector3  waistP_cur( waistP_pbuf.front() );
  vector3  waistV_cur( waistV_pbuf.front() );
  matrix33 waistR_cur( waistR_pbuf.front() );
  vector3  waistW_cur( waistW_pbuf.front() );
  Jvector  q_cur(q_pbuf.front());
  
  if( (int)waistP_pbuf.size() >= Pcog_dzmp[X].Nprev ){
    Pcog_dzmp[X].update(COG_dfilter(X), VCOG_dfilter(X), VCOGdot_dfilter(X), n_clock);
    Pcog_dzmp[Y].update(COG_dfilter(Y), VCOG_dfilter(Y), VCOGdot_dfilter(Y), n_clock);
    
    waistP_cur -= COG_dfilter;
    waistV_cur -= VCOG_dfilter;
  }
  
#ifdef WITH_TOE_JOINT
  body->joint(R_TOE_P)->q = q_cur[R_TOE_P];
  body->joint(L_TOE_P)->q = q_cur[L_TOE_P];
#endif
  for( int i = DOF_LEG ; i < DOF ; i++ )
    q_cmd2[i] = q_cur[i];
  
  vector3  ankleP[2] = {ankleP_pbuf[RLEG].front(), ankleP_pbuf[LLEG].front()};
  matrix33 ankleR[2] = {ankleR_pbuf[RLEG].front(), ankleR_pbuf[LLEG].front()};
  setJointCommand( waistP_cur, waistR_cur, ankleP, ankleR, waistV_cur, waistW_cur, q_cmd2, q_cmd_pre2, qdot_cmd2);
  
  ModelManager::calcModelZMP(Zmp_model2, zzmp_pbuf.front(), lin_mom_pre2, ang_mom_pre2, flg_firsttime[ZMP2]);
  setMotionCommandToBuffer(waistP_cur, waistR_cur, q_cmd2, Zmp_model2, basePos, baseAtt, angle, zmp, accel);
  
  if( (int)waistP_pbuf.size() >= Pcog_dzmp[X].Nprev ){
    waistP_pbuf.pop_front();
    waistV_pbuf.pop_front();
    waistR_pbuf.pop_front();
    waistW_pbuf.pop_front();
    for( int i = RLEG ; i <= LLEG ; i++ ){
      ankleP_pbuf[i].pop_front();
      ankleR_pbuf[i].pop_front();
    }
    q_pbuf.pop_front();
    zzmp_pbuf.pop_front();
  }
}

/* --- MwalkGenerator::setMotionCommand ------------------------------------ */
void
MwalkGenerator::setMotionCommand(
  double basePos[3], double baseAtt[9], double *angle, double zmp[3], double accel[3])
{
  vector3  ankleP[2] = {Ankle_cmd[RLEG].P, Ankle_cmd[LLEG].P};
  matrix33 ankleR[2] = {Ankle_cmd[RLEG].R, Ankle_cmd[LLEG].R};
  setJointCommand(Waist_cmd.P, Waist_cmd.R, ankleP, ankleR, Waist_cmd.V, Waist_cmd.W, q_cmd, q_cmd_pre, qdot_cmd);
  
  calcIdealZMPCommand(Zmp_ideal);
  ModelManager::calcModelZMP(Zmp_model, Zmp_cmd(Z), lin_mom_pre, ang_mom_pre, flg_firsttime[ZMP]);
  
  if( !flg_status[DFILTER_PREVIEW] )
    setMotionCommandToBuffer(Waist_cmd.P, Waist_cmd.R, q_cmd, Zmp_model, basePos, baseAtt, angle, zmp, accel);
  else
    setMotionCommandWithPreviewControl(basePos, baseAtt, angle, zmp, accel);
  
  if( flg_status[CHECK_JOINT_STATUS] )
    checkJointLimits((double)n_clock*DT, flg_firsttime[JCHECK]);
  
  if( flg_status[READY_INITCONF] && flg_status[WALK_REQUEST] ){
    mmode = WALKING;
    flg_status[READY_INITCONF] = false;
    flg_status[WALK_REQUEST] = false;
    flg_status[STOP_WALKING] = true;
    flg_firsttime[JCHECK] = true;
  }
  else if( flg_status[READY_INITCONF] && flg_status[ZMPFIX_REQUEST] ){
    mmode = ZMPFIX;
    flg_status[READY_INITCONF] = false;
    flg_status[ZMPFIX_REQUEST] = false;
  }
  else{
    vector3 dist(COG_cmd(X) - Zmp_cmd(X), COG_cmd(Y) - Zmp_cmd(Y), 0.0);    
    if( FootQueue.empty() && ws == STAND_ANKLE && tvmet::dot(dist,dist) < 1.0e-4*1.0e-4 && tvmet::dot(waistLink->v, waistLink->v) < 1.0e-4*1.0e-4 ){
      if( flg_status[STOP_WALKING] ){ 
	flg_status[STOP_WALKING] = false;
	mmode = INITCONF;
	Reset();
	
	if( verbose )
	  cout << "stop walking" << endl;
      }
      else if( flg_status[STOP_ZMPFIX] ){
	flg_status[STOP_ZMPFIX] = false;
	mmode = INITCONF;
	Reset();
	
	if( verbose )
	  cout << "stop zmpfix" << endl;
      }
    }
  }
}

void MwalkGenerator::setMotionCommand(
  vector3& basePos_, vector3& basePos_dot_, vector3& baseAtt_rpy_, vector3& baseAtt_rpy_dot_,
  vector<double>& angle_, vector<double>& angle_dot_, vector3& zmp_, vector3& accel_)
{
  double   basePos[3];
  double   baseAtt[9];
  double   angle[DOF];
  double   zmp[3];
  double   accel[3];
  matrix33 baseAtt_;

  setMotionCommand(basePos, baseAtt, angle, zmp, accel);
  getVector3(basePos_, basePos);
  getMatrix33FromRowMajorArray(baseAtt_, baseAtt);
  baseAtt_rpy_ = rpyFromRot(baseAtt_);
  
  basePos_dot_ = waistLink->v;
  baseAtt_rpy_dot_ = rpyAngularVelocityFromOmega(baseAtt_rpy_, waistLink->w);
  
  for(int i = 0 ; i < DOF ; i++ ){
    angle_[i] = angle[i];
    angle_dot_[i] = body->joint(i)->dq;
  }
  
  getVector3(zmp_, zmp);
  getVector3(accel_, accel);
}
