/*
 * Copyright (c) 2009,
 * @author Mitsuharu Morisawa
 *
 * AIST
 *
 * All rights reserved.
 *
 * This program is made available under the terms of the Eclipse Public License
 * v1.0 which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 */
#ifndef __MWALK_GENERATOR_H__
#define __MWALK_GENERATOR_H__

#include <bitset>
#include <queue>
#include "mwalk_common.h"
#include "WholeBodyMotionGenerator.h"
#include "FootMotionGenerator.h"
#include "FootStepManager.h"
#include "State.h"
#include "AnalyticalSolution.h"
#include "PreviewTrajectory.h"
#include "MotionInterpolator.h"

using namespace motion_generator;
using namespace motion_interpolator;
using namespace linear_control_system;


namespace mwalk_pattern_generator{
  
  enum FLAG_FIRSTTIME{
    ZMP,
    ZMP2,
    GSENS,
    JOINT,
    JCHECK,
    FLAG_FIRSTTIME_NUM
  };
  
  enum FLAG_STATUS{
    READY_INITCONF,
    WALK_REQUEST,
    STOP_WALKING,
    FOOT_MODIFICATION,
    ZMPFIX_REQUEST,
    STOP_ZMPFIX,
    TIME_ADJUSTMENT,
    ZMP_SHAPING_PREVIEW,
    DFILTER_PREVIEW,
    REACTIVE_MOTION,
    CHECK_JOINT_STATUS,
    ONE_STEP,
    ANKLE_DS_ROTATION,
    TRUNK_ROLL,
    FLAG_STATUS_NUM
  };
  
  class MwalkGenerator : public ModelManager, public FootMotionGenerator, public FootStepManager{
  public:
    /// permit output messages(std::cout)
    bool verbose;
    
    /// request to initialize variables
    bool flg_firsttime[FLAG_FIRSTTIME_NUM];

    /// walking options
    bitset<FLAG_STATUS_NUM>    flg_status;
    
    /// absolute time from starting walk
    int n_clock;
    
    /// rest time of current walking phase
    int n_period_rest;
    
    /// period of current walking phase
    int n_period;    
    
    int n_prev;
    int n_period_prev;
    int n_comp;
    
    /// current motion mode
    MotionMode mmode;
    
    /// current walking state
    WalkingState ws;
    
    /// next walking state
    WalkingState ws_next; 
    
    /// previous walking state
    WalkingState ws_pre;
    WalkingState ws_prev;
#ifdef WITH_TOE_JOINT
    WalkingState ws_last;
    WalkingState ws_last_prev;
#endif
    
    double STEP_YAW_RATIO;
    double WAIST_YAW_RATIO;
    double ARM_SWING_RATIO;
    
    double THmiddle_ratio;
    double XHmiddle_ratio;
    
    double INIT_COG_OFFSET_GAIN;
    double UPDATE_COG_OFFSET_GAIN;
    
    vector3 AbsPosOrg, AbsRpyOrg;
    vector3 AbsPosOrgNext, AbsRpyOrgNext;
    vector3 AbsPosOrg_prev, AbsRpyOrg_prev;
    
    WholeBodyMotionGenerator *wbmg;
    AnalyticalSolution ANcog, ANcog_reac;
    PreviewTrajectory  Pcog[2];
    PreviewTrajectoryType1 Pcog_dzmp[2];
    
    vector3 COG_cmd, VCOG_cmd, VCOGdot_cmd, Zmp_cmd, VZmp_cmd;
    vector3 COG_shp, VCOG_shp, VCOGdot_shp, Zmp_shp, VZmp_shp;
    vector3 COG_prev, VCOG_prev, Zmp_prev, VZmp_prev;
    vector3 COG_dfilter, VCOG_dfilter, VCOGdot_dfilter;
    vector3 COG_cmp, VCOG_cmp, VCOGdot_cmp;
    
    vector3 VCOGdot_res;
    vector3 Gsensor_offset;
    vector3 Vgsens_pre;
    
    State   Foot_cmd[2];
    State   Foot_next_cmd[2];
    State   Ankle_cmd[2];
    State   Waist_cmd;
    double  waist_yaw_offset;
    vector3 Pivot_cmd[2];
    
    deque<vector3>   waistP_pbuf;
    deque<vector3>   waistV_pbuf;
    deque<matrix33>  waistR_pbuf;
    deque<vector3>   waistW_pbuf;
    deque<vector3>   ankleP_pbuf[2];
    deque<matrix33>  ankleR_pbuf[2];
    deque<Jvector>   q_pbuf;
    deque<double>    zzmp_pbuf;
    
    MotionInterpolator WaistHeight;
    MotionInterpolator Shoulder[2];
    double Trunk_roll_fixed_height;
    double TR_ankle_ds_ratio[2];
    
    deque<FootPlace> FootQueue;
    deque<FootPlace> FootQueueNext;
    deque<FootPlace> FootQueuePrev;
    
    double cog_height[Ncog];
    double omega[Ncog];
    
    // for walking pattern by analytical solution
    double Tm[Ncog+1];
    double Kg[Ncog*7][3], Kg_reac[Ncog*7][3];
    vector3 zmp_des[Ncog+1], vzmp_des[Ncog+1];
    vector3 zmp_reac[Ncog+1], vzmp_reac[Ncog+1];
    dvector zmp_err[2];
    ANmatrix Aginv, Aginv_prev;
    ANvector Bg[2];
    
    vector3 total_com, total_inr;
    vector3 lin_mom_pre,  ang_mom_pre;
    vector3 lin_mom_pre2, ang_mom_pre2;
    
    vector3 Xg_offset;
    vector3 Vg_offset;
    vector3 Xg_offset_ini;
    
    vector3 Zmp_model;
    vector3 Zmp_ideal;
    vector3 Zmp_model2;
    vector3 Zmp_ref;
    vector3 Zmp_res;
    
    Jvector q_ini;
    Jvector q_res;
    Jvector q_res_pre;
    Jvector qdot_res;
    Jvector q_cmd, q_cmd2;
    Jvector q_cmd_pre, q_cmd_pre2;
    Jvector qdot_cmd, qdot_cmd2;
    Jvector qddot_cmd;
    Jvector q_cmd_org;

    dvector Vctrl_cmd;
    dvector PL_cmd;
    
  protected:
    void setFootPlace(int& n_comp_);
    void updateFootPlace(State& Leg, State& Leg_next);
    void setOriginForFirstStep(int leg);
    void setLandingPosition(State& Leg_next);
    void setLandingPosition(State& SwingLeg_next, State& SupportLeg);
    void setWaistHeightParam(double hip_z, double step_length, int n_start, int n_period);
    void setWaistHeightParam(FootPlace& FP, State& foot_, State& foot_next_, int n_clock);
    void setSwingArmParam(WalkingState& ws_next_, unsigned int FQsize, double step_length, int n_start, int n_period);
    void adjustSupportTime(vector3& Xgi, vector3& Vgi, vector3& Zmpi, vector3& VZmpi,
			   vector3& Xgf, vector3& Vgf, vector3& Zmpf, vector3& VZmpf, double foot_angle, int& n_comp_);

    void pushCOGCoef(int n_period, int n_clock);
    void popCOGCoef(vector3& Xgi_cmd, vector3&  Vgi_cmd, vector3& Zmpi_cmd, vector3& VZmpi_cmd, int n_aft);
    void makeCOGCoef(vector3& Xgi_cmd, vector3& Vgi_cmd, vector3& Zmpi_cmd, vector3& VZmpi_cmd, int n_period, int n_clock);
    void makeCOGCoefReactive(vector3& Xgi_reac_, vector3& Vgi_reac_, vector3& Zmpi_reac_, vector3& VZmpi_reac_, int n_period_, int n_clock_);
    void makeReactiveMotion(void);
    void makeCOGComp(int n_clock);

    /*! @brief This function is only used for monitoring maximum fluctuation of ZMP from analytical solution.
     * @param[in] n_period_rest
     */
    void calcZMPErrorOfAnalyticalSolution(int n_period_rest);
    
    /*! @brief Suppress ZMP fluctuation by preview control
     */
    void setZMPShaping(void);
    
    void walkingState_STAND(WalkingState& ws_next, int& n_period, int& n_period_rest);
    void walkingState_TO_LEFT(WalkingState& ws_next, int& n_period, int& n_period_rest);
    void walkingState_TO_RIGHT(WalkingState& ws_next, int& n_period, int& n_period_rest);
    void walkingState_TO_CENTER(WalkingState& ws_next, int& n_period, int& n_period_rest);
#ifndef WITH_TOE_JOINT
    void walkingState_LEFT_SUPPORT(WalkingState& ws_next, int& n_period, int& n_period_rest);
    void walkingState_RIGHT_SUPPORT(WalkingState& ws_next, int& n_period, int& n_period_rest);
#else
    void walkingState_LEFT_SUPPORT_ANKLE(WalkingState& ws_next, int& n_period, int& n_period_rest);
    void walkingState_LEFT_SUPPORT_ANKLE_TO_TOE(WalkingState& ws_next, int& n_period, int& n_period_rest);
    void walkingState_LEFT_SUPPORT_TOE_TO_ANKLE(WalkingState& ws_next, int& n_period, int& n_period_rest);
    void walkingState_LEFT_SUPPORT_TOE(WalkingState& ws_next, int& n_period, int& n_period_rest);
    void walkingState_RIGHT_SUPPORT_ANKLE(WalkingState& ws_next, int& n_period, int& n_period_rest);
    void walkingState_RIGHT_SUPPORT_ANKLE_TO_TOE(WalkingState& ws_next, int& n_period, int& n_period_rest);
    void walkingState_RIGHT_SUPPORT_TOE_TO_ANKLE(WalkingState& ws_next, int& n_period, int& n_period_rest);
    void walkingState_RIGHT_SUPPORT_TOE(WalkingState& ws_next, int& n_period, int& n_period_rest);
#endif

    /*! @brief calculate leg joint angles from waist and feet position and angular velocity of all joints
     * @param[in]     waist_p    waist position
     * @param[in]     waist_R    waist attitude
     * @param[in]     angle_p    ankle position[2]
     * @param[in]     angle_R    ankle attitude[2]
     * @param[in]     waist_v    waist velocity (set to body pointer)
     * @param[in]     waist_w    waist angular velicity (set to body pointer)
     * @param[in,out] q_cmd_     current joint angles[DOF]
     * @param[in]     q_cmd_pre_ previous joint angles[DOF]
     * @param[out]    qdot_cmd_  joint anglular velocity[DOF]
     */
    void setJointCommand(vector3& waist_p, matrix33& waist_R, vector3 *ankle_p, matrix33 *ankle_R,
			 vector3& waist_v, vector3& waist_w, Jvector& q_cmd_, Jvector& q_cmd_pre_, Jvector& qdot_cmd_);
    void calcIdealZMPCommand(vector3& zmp_ideal_);
    void setMotionCommandWithPreviewControl(double basePos[3], double baseAtt[9], double *angle,
					    double zmp[3], double accel[3]);
    void setMotionCommandToBuffer(vector3& waist_pos_, matrix33& waist_att_, Jvector& q_cmd_,
				  vector3& zmp_model_, double basePos[3], double baseAtt[9],
				  double *angle, double zmp[3], double accel[3]);
    
  public:
    /*! @brief Under development for reactive motion
     * @param[in] zmp 
     * @param[in] accel
     */
    void calcZMPTrackingControl(const double zmp[3], const double accel[3]);

    /*! @brief Set current robot configuration as the starting configuration
     * @param[in]  basePos waist position (do not need to set)
     * @param[in]  baseAtt waist attitude
     * @param[in]  angle   all of joint angles
     */
    void initConfiguration(const double basePos[3], const double baseAtt[9],
			   const double angle[DOF], double zmp[3]);
    void initState(vector3& waist_p, vector3& waist_v, vector3& waist_rpy, vector3& waist_rpy_dot,
		   vector<double>& angle, vector<double>& angle_dot, vector<double>& angle_pre,
		   vector3& zmp, vector3& vzmp, WalkingState& ws_next, vector3& lin_mom_pre_,
		   vector3& ang_mom_pre_, vector3& vgsens_pre_, int n_clock_, bool first_step);
    void setStateForZmpfix(double basePos[3], double baseAtt[9], double *angle,
			   double zmp[3]);
    void setCOGHeight(deque<vector3>& waist_p, deque<vector3>& waist_rpy);
    
    /*! @brief Generate arm and waist swing
     */
    void makeUpperBodyMotion(void);
    void makeUpperBodyMotion(vector3& waist_p, vector3& waist_v, vector3& waist_rpy, vector3& waist_rpy_dot,
			     vector<double>& angle, vector<double>& angle_dot);
    
    /*! @brief Generate COG motion and estimate distance between the COG and the waist position
     */
    void makeCOGMotion(void);
    void makeWholeBodyMotion(void);
    void updateWholeBodyMotion(void);
    
    /*! @brief Fill user buffers with the current motor command
     * @param[out]  basePos waist position (do not need to set)
     * @param[out]  baseAtt waist attitude
     * @param[out]  angle   all of joint angles
     * @param[out]  zmp     zmp position at waist coordinate
     * @param[out]  accel   ideal acceleration at sensor position used by kalman filter
     */
    void setMotionCommand(double basePos[3], double baseAtt[9], double *angle,
			  double zmp[3], double accel[3]);
    void setMotionCommand(vector3& basePos_, vector3& basePos_dot_, vector3& baseAtt_rpy_,
			  vector3& baseAtt_rpy_dot_, vector<double>& angle_,
			  vector<double>& angle_dot_, vector3& zmp_, vector3& accel_);
    
    /*! @brief Manage transition of walking state
     * @param[in,out]  ws_next next walking state
     * @param[in,out]  n_period a period of current walking state
     * @param[in,out]  n_period_rest a rest time of current walking state
     */
    void updateWalkingState(WalkingState& ws_next, int& n_period, int& n_period_rest);
    
    /*! @brief Manage transition of walking state
     */
    inline void updateWalkingState(void){
      updateWalkingState(ws_next, n_period, n_period_rest);
    }
#ifndef WITH_TOE_JOINT
    /*! @brief Generate swing leg trajectory and foot rotation 
     */
    inline void makeSwingLegMotion(void){
      FootMotionGenerator::makeSwingLegMotion(Foot_cmd, Ankle_cmd, Pivot_cmd, ws, n_clock);
    }
#else
    /*! @brief Generate swing leg trajectory and foot rotation
     */
    inline void makeSwingLegMotion(void){
      FootMotionGenerator::makeSwingLegMotion(Foot_cmd, Ankle_cmd,
					      body->joint(R_TOE_P)->q, body->joint(L_TOE_P)->q,
					      Pivot_cmd, ws, n_clock);
    }
#endif
    /*! @brief Initialize flags and variables which are necessary to initialize only once
     */
    void Initialize(void);

    /*! @brief Reset flags and variables which are necessary to initialize for restarting to walk
     */
    void Reset(void);
    
    MwalkGenerator(istringstream& strm);
    MwalkGenerator(int argc, char* argv[]);
    MwalkGenerator(const char *model, int argc, char* argv[]);
    virtual ~MwalkGenerator();
    
    /*! @brief To get walking phase
     * @return walking pahse
     */
    inline WalkingState m_getWalkingPhase(void){ return ws; }

    /*! @brief To get a period of walking phase
     * @return a period of walking pahse
     */
    inline int m_getWalkingPhasePeriod(void){ return n_period; }
    
    /*! @brief To get a current period of each walking phase
     * @return a current period of each walking phase
     */
    inline int m_getWalkingPhaseCurrent(void){ return n_period-n_period_rest-1; }

    /*! @brief To get current joint command
     * @param[out] q_cmd_ current joint command [DOF]
     */
    inline void m_getJointCommand(Jvector& q_cmd_){ q_cmd_ = q_cmd; }
    
    /*! @brief To get position and attitude of feet
     * @param[out] ankle_p ankle position [2]
     * @param[out] ankle_R ankle attitude [2]
     */
    inline void m_getFootCommand(vector3 *ankle_p, matrix33 *ankle_R){
      ankle_p[RLEG] = Foot_cmd[RLEG].P;
      ankle_p[LLEG] = Foot_cmd[LLEG].P;
      ankle_R[RLEG] = Foot_cmd[RLEG].R;
      ankle_R[LLEG] = Foot_cmd[LLEG].R;
    };
    
    /*! @brief To get waist position and attitude
     * @param[out] waist_p waist position
     * @param[out] waist_R waist attitude
     */
    inline void m_getWaistCommand(vector3& waist_p, matrix33& waist_R){
      waist_p = Waist_cmd.P;
      waist_R = Waist_cmd.R;
    }
    
    /*! @brief To get command of COG position
     * @param[out] cog_p center of gravity position
     */
    inline void m_getCOGCommand(vector3& cog_p){ cog_p = COG_cmd; }
  };
  
}  // end of namespace mwalk_pattern_generator


#endif // __MWALK_GENERATOR_H__
