/*
 * Copyright (c) 2009,
 * @author Mitsuharu Morisawa
 *
 * AIST
 *
 * All rights reserved.
 *
 * This program is made available under the terms of the Eclipse Public License
 * v1.0 which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 */
#include <iostream>
#include "PreviewControl.h"
#include "ReadText.h"

using namespace linear_control_system;
using namespace read_text_parameter;
using namespace boost::numeric::ublas;

PreviewControl_base::PreviewControl_base(void)
{
  PreviewControl_base::Initialize();
}

PreviewControl_base::~PreviewControl_base()
{
}

void PreviewControl_base::Initialize(void)
{
  LinearControl_SISO::Initialize();
  
  Nprev = 0;
  n_start = 0;
  
  F.clear();
  ref.clear();
}

bool PreviewControl_base::loadGains(const char *filename, int debug_level)
{
  ifstream istrm(filename);
  if (!istrm.is_open()) {
    std::cerr << "file not found (" << filename << ")" << endl;
    return false;
  }
  else{
    std::cout << "load preview gain (" << filename << ")" << endl;
    
    string buf;
    while( GetString(istrm, buf) ){
      if((buf[0] == '%') || (buf[0] == '#'))
	SkipToEOL(istrm);       // skip commnent line
      else if( buf == "A" ) GetMatrix(istrm, A);
      else if( buf == "B" ) GetMatrix(istrm, B);
      else if( buf == "C" ) GetMatrix(istrm, C);
      else if( buf == "K" ) GetMatrix(istrm, K);
      else if( buf == "F" ) GetMatrix(istrm, F);
      else {
	std::cout << "Don't know :" << buf << endl;
	SkipToEOL(istrm);
	break;
      }
    }
    
    if( debug_level > 0 ){
      std::cout << "A=" << A << endl;
      std::cout << "B=" << B << endl;
      std::cout << "C=" << C << endl;
      std::cout << "K=" << K << endl;
      std::cout << "F=" << F << endl;
    }
    
    Nprev = (int)F.size2();
    
    x.resize(B.size1());
    x = dzerovector(B.size1());
    
    ref.resize(Nprev);
    ref = dzerovector(Nprev);
  }
  istrm.close();
  
  return true;
}

void PreviewControl_base::setReference(double dat, int n_now)
{
  ref((n_now-n_start) % Nprev) = dat;
}


//
// Normal Preview Control
//

PreviewControl::PreviewControl(void)
{
 
}

PreviewControl::~PreviewControl()
{
}

void PreviewControl::reset(void)
{
  x = dzerovector(x.size());
}

void PreviewControl::update(double& p, double& v, double& a, double& y, int n_now)
{
  p = x(0);
  v = x(1);
  a = x(2);
  y = ublas::prod(C, x)(0);
  
  u = -ublas::prod(K, x)(0);
  for(int i = 0 ; i < Nprev ; i++){
    u += F(0,i) * ref((i+n_now-n_start+1) % Nprev);
  }
  
  x = ublas::prod(A, x) + ublas::matrix_column<dmatrix>(B, 0) * u;
}

void PreviewControl::update(double& p, double& v, double& a, int n_now)
{
  double y = 0.0;
  PreviewControl::update(p, v, a, y, n_now);
}


//
// Type1 Preview Control
//

PreviewControlType1::PreviewControlType1(void)
{
  PreviewControlType1::reset();
}

PreviewControlType1::~PreviewControlType1()
{
}

void PreviewControlType1::reset(double sx_)
{
  x = dzerovector(x.size());
  sx = sx_;
}

void PreviewControlType1::update(double& p, double& v, double& a, double& y, int n_now)
{
  p = x(0);
  v = x(1);
  a = x(2);
  y = ublas::prod(C, x)(0);
  
  u = -K(0,0)*sx -K(0,1)*x(0) -K(0,2)*x(1) -K(0,3)*x(2);
  for(int i = 0 ; i < Nprev ; i++)
    u += F(0, i) * ref((i+n_now-n_start+1) % Nprev);
  
  x = ublas::prod(A, x) + ublas::matrix_column<dmatrix>(B, 0) * u;
  sx += y - ref((n_now-n_start+1) % Nprev);
}

void PreviewControlType1::update(double& p, double& v, double& a, int n_now)
{
  double y = 0.0;
  PreviewControlType1::update(p, v, a, y, n_now);
}
