/*
 * Copyright (c) 2009,
 * @author Mitsuharu Morisawa
 *
 * AIST
 *
 * All rights reserved.
 *
 * This program is made available under the terms of the Eclipse Public License
 * v1.0 which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 */
#include <fstream>
#include "StateLogger.h"


StateLogger::StateLogger(void)
{
  clearLogData();
}

StateLogger::~StateLogger()
{
}

void StateLogger::clearLogData(void)
{
  waist_queue.clear();
  pos_queue.clear();
  vel_queue.clear();
  zmp_queue.clear();
  hip_queue.clear();
  gsens_queue.clear();
  zmpref_queue.clear();
  for( int leg = RLEG ; leg <= LLEG ; leg++ ){
    pivot_queue[leg].clear();
    ankle_queue[leg].clear();
    foot_queue[leg].clear();
  }
}

void StateLogger::pushWaist(vector3& Pwaist, matrix33& Rwaist, vector3& Vwaist, vector3& Wwaist)
{
  waist_state wstate;
  wstate.p = Pwaist;
  wstate.R = Rwaist;
  wstate.v = Vwaist;
  wstate.w = Wwaist;
  waist_queue.push_back(wstate);
}

void StateLogger::pushAnkle(vector3& RPankle, matrix33& RRankle, vector3& RVankle, vector3& RWankle, vector3& LPankle, matrix33& LRankle, vector3& LVankle, vector3& LWankle)
{
  ankle_state astate;
  astate.p = RPankle;
  astate.R = RRankle;
  astate.v = RVankle;
  astate.w = RWankle;
  ankle_queue[RLEG].push_back(astate);
  
  astate.p = LPankle;
  astate.R = LRankle;
  astate.v = LVankle;
  astate.w = LWankle;
  ankle_queue[LLEG].push_back(astate);
}

void StateLogger::pushFoot(vector3& RPfoot, vector3& RVfoot, vector3& RFoot_rpy, vector3& LPfoot, vector3& LVfoot, vector3& LFoot_rpy)
{
  foot_state fstate;
  fstate.p = RPfoot;
  fstate.v = RVfoot;
  fstate.rpy = RFoot_rpy;
  if( foot_queue[RLEG].empty() )
    fstate.drpy = 0.0;
  else
    fstate.drpy = (RFoot_rpy - (foot_queue[RLEG].end() - 1)->rpy) / DT;
  foot_queue[RLEG].push_back(fstate);
  
  fstate.p = LPfoot;
  fstate.v = LVfoot;
  fstate.rpy = LFoot_rpy;
  if( foot_queue[LLEG].empty() )
    fstate.drpy = 0.0;
  else
    fstate.drpy = (LFoot_rpy - (foot_queue[LLEG].end() - 1)->rpy) / DT;
  foot_queue[LLEG].push_back(fstate);
}

void StateLogger::pushPivot(vector3 *pivot)
{
  pivot_queue[RLEG].push_back(pivot[RLEG]);
  pivot_queue[LLEG].push_back(pivot[LLEG]);
}

void StateLogger::saveLogData(string& filename)
{
  if( !waist_queue.empty() ){
    string filename_waist = filename + ".waist";
    ofstream fWaist(filename_waist.c_str());
    
    double tm = 0.0;
    deque<waist_state>::iterator iter;
    for( iter = waist_queue.begin() ; iter != waist_queue.end() ; ++iter ){
      fWaist << tm << " "
	     << iter->p(X)   << " " << iter->p(Y)   << " " << iter->p(Z) << " "
	     << iter->R(0,0) << " " << iter->R(0,1) << " " << iter->R(0,2) << " "
	     << iter->R(1,0) << " " << iter->R(1,1) << " " << iter->R(1,2) << " "
	     << iter->R(2,0) << " " << iter->R(2,1) << " " << iter->R(2,2) << " "
	     << iter->v(X)   << " " << iter->v(Y)   << " " << iter->v(Z) << " "
	     << iter->w(X)   << " " << iter->w(Y)   << " " << iter->w(Z) << endl;
      
      tm += DT;
    }
    fWaist.close();
  }
  
  if( !pos_queue.empty() ){
    string filename_pos = filename + ".pos";
    ofstream fPos(filename_pos.c_str());
    
    double tm = 0.0;
    deque<vector<double> >::iterator iter;
    for( iter = pos_queue.begin() ; iter != pos_queue.end() ; ++iter ){
      fPos << tm << " ";
      for( int n = 0 ; n < DOF ; n++ )
	fPos << (*iter)[n] << " ";
      fPos << endl;
      
      tm += DT;
    }
    fPos.close();
  }
  
  if( !vel_queue.empty() ){
    string filename_vel = filename + ".vel";
    ofstream fVel(filename_vel.c_str());
    
    double tm = 0.0;
    deque<vector<double> >::iterator iter;
    for( iter = vel_queue.begin() ; iter != vel_queue.end() ; ++iter ){
      fVel << tm << " ";
      for( int n = 0 ; n < DOF ; n++ )
	fVel << (*iter)[n] << " ";
      fVel << endl;
      
      tm += DT;
    }
    fVel.close();
  }
  
  if( !zmp_queue.empty() ){
    string filename_zmp = filename + ".zmp";
    ofstream fZmp(filename_zmp.c_str());
    
    double tm = 0.0;
    deque<vector3>::iterator iter;
    for( iter = zmp_queue.begin() ; iter != zmp_queue.end() ; ++iter ){
      fZmp << tm << " " << (*iter)(X) << " " << (*iter)(Y) << " " << (*iter)(Z) << endl;
      tm += DT;
    }
    fZmp.close();
  }

  if( !hip_queue.empty() ){
    string filename_hip = filename + ".hip";
    ofstream fHip(filename_hip.c_str());
    
    double tm = 0.0;
    deque<vector3>::iterator iter;
    for( iter = hip_queue.begin() ; iter != hip_queue.end() ; ++iter ){
      fHip << tm << " " << (*iter)(X) << " " << (*iter)(Y) << " " << (*iter)(Z) << endl;
      tm += DT;
    }
    fHip.close();
  }
  
  if( !gsens_queue.empty() ){
    string filename_gsens = filename + ".gsens";
    ofstream fGsens(filename_gsens.c_str());
    
    double tm = 0.0;
    deque<vector3>::iterator iter;
    for( iter = gsens_queue.begin() ; iter != gsens_queue.end() ; ++iter ){
      fGsens << tm << " " << (*iter)(X) << " " << (*iter)(Y) << " " << (*iter)(Z) << endl;
      tm += DT;
    }
    fGsens.close();
  }
  
  if( !zmpref_queue.empty() ){
    string filename_zmpref = filename + ".zmpref";
    ofstream fZmpref(filename_zmpref.c_str());
    
    double tm = 0.0;
    deque<vector3>::iterator iter;
    for( iter = zmpref_queue.begin() ; iter != zmpref_queue.end() ; ++iter ){
      fZmpref << tm << " " << (*iter)(X) << " " << (*iter)(Y) << " " << (*iter)(Z) << endl;
      tm += DT;
    }
    fZmpref.close();
  }
  
  if( !pivot_queue[RLEG].empty() ){
    string filename_pivot = filename + ".pivot";
    ofstream fPivot(filename_pivot.c_str());
    
    double tm = 0.0;
    unsigned int size = pivot_queue[RLEG].size();
    for( unsigned int i = 0 ; i < size ; i++ ){
      fPivot << tm << " "
	     << pivot_queue[RLEG][i](X) << " " << pivot_queue[RLEG][i](Y) << " " <<pivot_queue[RLEG][i](Z) << " "
	     << pivot_queue[LLEG][i](X) << " " << pivot_queue[LLEG][i](Y) << " " <<pivot_queue[LLEG][i](Z) << endl;
      tm += DT;
    }
    fPivot.close();
  }
  
  if( !ankle_queue[RLEG].empty() ){
    string filename_ankle = filename + ".ankle";
    ofstream fAnkle(filename_ankle.c_str());
    
    double tm = 0.0;
    unsigned int size = ankle_queue[RLEG].size();
    for( unsigned int i = 0 ; i < size ; i++ ){
      fAnkle << tm << " "
	     << ankle_queue[RLEG][i].p(X)   << " " << ankle_queue[RLEG][i].p(Y)   << " " << ankle_queue[RLEG][i].p(Z) << " "
	     << ankle_queue[RLEG][i].v(X)   << " " << ankle_queue[RLEG][i].v(Y)   << " " << ankle_queue[RLEG][i].v(Z) << " "
	     << ankle_queue[RLEG][i].R(0,0) << " " << ankle_queue[RLEG][i].R(0,1) << " " << ankle_queue[RLEG][i].R(0,2) << " "
	     << ankle_queue[RLEG][i].R(1,0) << " " << ankle_queue[RLEG][i].R(1,1) << " " << ankle_queue[RLEG][i].R(1,2) << " "
	     << ankle_queue[RLEG][i].R(2,0) << " " << ankle_queue[RLEG][i].R(2,1) << " " << ankle_queue[RLEG][i].R(2,2) << " "
	     << ankle_queue[RLEG][i].w(X)   << " " << ankle_queue[RLEG][i].w(Y)   << " " << ankle_queue[RLEG][i].w(Z) << " "
	     << ankle_queue[LLEG][i].p(X)   << " " << ankle_queue[LLEG][i].p(Y)   << " " << ankle_queue[LLEG][i].p(Z) << " "
	     << ankle_queue[LLEG][i].v(X)   << " " << ankle_queue[LLEG][i].v(Y)   << " " << ankle_queue[LLEG][i].v(Z) << " "
	     << ankle_queue[LLEG][i].R(0,0) << " " << ankle_queue[LLEG][i].R(0,1) << " " << ankle_queue[LLEG][i].R(0,2) << " "
	     << ankle_queue[LLEG][i].R(1,0) << " " << ankle_queue[LLEG][i].R(1,1) << " " << ankle_queue[LLEG][i].R(1,2) << " "
	     << ankle_queue[LLEG][i].R(2,0) << " " << ankle_queue[LLEG][i].R(2,1) << " " << ankle_queue[LLEG][i].R(2,2) << " "
	     << ankle_queue[LLEG][i].w(X)   << " " << ankle_queue[LLEG][i].w(Y)   << " " << ankle_queue[LLEG][i].w(Z) << endl;
      tm += DT;
    }
    fAnkle.close();
  }
  
  if( !foot_queue[RLEG].empty() ){
    string filename_foot = filename + ".foot";
    ofstream fFoot(filename_foot.c_str());
    
    double tm = 0.0;
    unsigned int size = ankle_queue[RLEG].size();
    for( unsigned int i = 0 ; i < size ; i++ ){
      fFoot << tm << " "
	    << foot_queue[RLEG][i].p(X)    << " " << foot_queue[RLEG][i].p(Y)    << " " << foot_queue[RLEG][i].p(Z)    << " "
	    << foot_queue[RLEG][i].v(X)    << " " << foot_queue[RLEG][i].v(Y)    << " " << foot_queue[RLEG][i].v(Z)    << " "
	    << foot_queue[RLEG][i].rpy(X)  << " " << foot_queue[RLEG][i].rpy(Y)  << " " << foot_queue[RLEG][i].rpy(Z)  << " "
	    << foot_queue[RLEG][i].drpy(X) << " " << foot_queue[RLEG][i].drpy(Y) << " " << foot_queue[RLEG][i].drpy(Z) << " "
	    << foot_queue[LLEG][i].p(X)    << " " << foot_queue[LLEG][i].p(Y)    << " " << foot_queue[LLEG][i].p(Z)    << " "
	    << foot_queue[LLEG][i].v(X)    << " " << foot_queue[LLEG][i].v(Y)    << " " << foot_queue[LLEG][i].v(Z)    << " "
	    << foot_queue[LLEG][i].rpy(X)  << " " << foot_queue[LLEG][i].rpy(Y)  << " " << foot_queue[LLEG][i].rpy(Z)  << " "
	    << foot_queue[LLEG][i].drpy(X) << " " << foot_queue[LLEG][i].drpy(Y) << " " << foot_queue[LLEG][i].drpy(Z) << endl;
      tm += DT;
    }
    fFoot.close();
  }
}
