/*
 * Copyright (c) 2009,
 * @author Shuuji Kajita, Shizuko Hattori, Mitsuharu Morisawa
 *
 * AIST
 *
 * All rights reserved.
 *
 * Redistribution in source file and binary forms, with or without modification is
 * strictly prohibited. Use in source and binary forms, with or without
 * modification is granted to the members the Joint AIST/CNRS Robotics Lab(JRL)
 * and the LAAS/CNRS lab.
 */
#include "WholeBodyMotionGenerator.h"

using namespace motion_generator;


WholeBodyMotionGenerator::WholeBodyMotionGenerator(BodyPtr body_) : ModelManager(body_)
{
  rootIndex = body->rootLink()->index;
  initCOGInertia();
  initWaistToAnkleJacobian();
  initInertiaMatrix();
}

WholeBodyMotionGenerator::~WholeBodyMotionGenerator()
{
}

void WholeBodyMotionGenerator::initCOGInertia(void)
{
  body->calcForwardKinematics(true);
  
  const LinkTraverse& traverse = body->linkTraverse();
  int n = traverse.numLinks();
  for(int i = n - 1 ; i >= 0 ; --i){
    Link* link = traverse[i];
    linkInr[i].Mass = link->m;
    for(Link* child = link->child; child; child = child->sibling)
      linkInr[i].Mass += linkInr[child->index].Mass;
  }
  Total_mass = linkInr[rootIndex].Mass;
}

void WholeBodyMotionGenerator::initWaistToAnkleJacobian(void)
{
  for( int leg = RLEG ; leg <= LLEG ; leg++ ){
    Jankle[leg] = dzeromatrix(6,6);
    setMatrix33(matrix33(tvmet::identity<matrix33>()), Jankle[leg], 0, 0);
    setMatrix33(matrix33(tvmet::identity<matrix33>()), Jankle[leg], 3, 3);
  }
}

void WholeBodyMotionGenerator::initInertiaMatrix(void)
{
  Mconst = dzeromatrix(6, Nwbm);
  
  for( int leg = RLEG ; leg <= LLEG ; leg++ )
    Mconst_leg[leg] = dzeromatrix(6, 6);

#ifdef WITH_CHEST_JOINT
#ifndef USE_CHEST_RMC
  Mconst_chest.resize(6, waistToChestPath->numJoints());  
#endif
#endif
  
#ifdef WITH_ARM_JOINT
  for( int leg = RLEG ; leg <= LLEG ; leg++ )
    Mconst_arm[leg] = dzeromatrix(6, 6);
#endif
  
#ifdef WITH_NECK_JOINT
  Mconst_neck.resize(6, waistToNeckPath->numJoints());    
#endif
}

void WholeBodyMotionGenerator::calcCOGInertia(vector3 &CoM, matrix33 &Inr)
{
  const LinkTraverse& traverse = body->linkTraverse();
  int n = traverse.numLinks();
  for(int i = n - 1; i >= 0; --i){
    Link* link = traverse[i];
    linkInr[i].Inr = (link->R * link->I) * tvmet::trans(link->R);
    link->wc = link->p + link->R * link->c;
    linkInr[i].Com = link->wc * link->m;
    for(Link* child = link->child; child; child = child->sibling){
      linkInr[i].Com += linkInr[child->index].Mass * linkInr[child->index].Com;
      linkInr[i].Inr += linkInr[child->index].Inr;
    }
    linkInr[i].Com /= linkInr[i].Mass;
    vector3  wc_com(link->wc - linkInr[i].Com);
    matrix33 wc_hat;
    makeCrossMatrix(wc_hat, wc_com);
    linkInr[i].Inr += link->m * wc_hat * tvmet::trans(wc_hat);
    for(Link* child = link->child; child; child = child->sibling){
      vector3  r_com(linkInr[child->index].Com - linkInr[i].Com);
      matrix33 r_hat;
      makeCrossMatrix(r_hat, r_com);
      linkInr[i].Inr += linkInr[child->index].Mass * r_hat * tvmet::trans(r_hat);
    }
    if( link->jointType == Link::ROTATIONAL_JOINT ){
      linkInr[i].Hlin = linkInr[i].Mass * tvmet::cross( link->R*link->a, (linkInr[i].Com-link->p) );
      linkInr[i].Hang = tvmet::cross( linkInr[i].Com, linkInr[i].Hlin ) + linkInr[i].Inr * link->R*link->a;
    }
  }
  CoM = linkInr[rootIndex].Com;
  Inr = linkInr[rootIndex].Inr;
}

//  Jankle = | E   -wPb->wPa | 
//           | 0      E      |
void WholeBodyMotionGenerator::setWaistToAnkleJacobian(State& Waist, State *Ankle)
{
  for( int leg = RLEG ; leg <= LLEG ; leg++ ){
    vector3 wPa(Ankle[leg].P - Waist.P);
    Jankle[leg](0, 4) =  wPa(2);
    Jankle[leg](0, 5) = -wPa(1);
    Jankle[leg](1, 3) = -wPa(2);
    Jankle[leg](1, 5) =  wPa(0);
    Jankle[leg](2, 3) =  wPa(1);
    Jankle[leg](2, 4) = -wPa(0);
  }
}

/**************** MakeInertiaTensor **********************/
//	| MmatB | 	| ME   -MrB->G |   | Mmatleg |		    | Mmatarm |   
//	|       | =	| 0      I     |   |         |[Right/Left]  |         |[Right/Left]   
//
void WholeBodyMotionGenerator::setInertiaMatrix(State& Waist)
{
  const matrix33 Midentity(tvmet::identity<matrix33>()*Total_mass);
  const matrix33 Zero(0.0);
  setMatrix33(Midentity, Mconst, 0, 0);
  setMatrix33(Zero,      Mconst, 3, 0);
  vector3 mcPw(total_com - Waist.P);
  mcPw *= -Total_mass;
  matrix33 mc_hat;
  makeCrossMatrix(mc_hat, mcPw);
  setMatrix33(mc_hat,    Mconst, 0, 3);
  setMatrix33(total_inr, Mconst, 3, 3);
  
  //-------------- Leg momentum tensor -------------
  for(int leg = RLEG ; leg <= LLEG ; leg++){
    for(int j = 0 ; j < 6 ; j++){
      int joint_index = waistToAnklePath[leg]->joint(j)->index;
      vector3 h(linkInr[joint_index].Hang - tvmet::cross(total_com, linkInr[joint_index].Hlin));
      setVector3(linkInr[joint_index].Hlin, Mconst_leg[leg], 0, j);
      setVector3(h, Mconst_leg[leg], 3, j);
    }
  }
  
#ifdef WITH_CHEST_JOINT
  //--------------Chest momentum tensor--------------- 
  for(int j = 0 ; j < DOF_CHEST ; j++){
    int joint_index = waistToChestPath->joint(j)->index;
    vector3 h(linkInr[joint_index].Hang - tvmet::cross(total_com, linkInr[joint_index].Hlin));
#ifndef USE_CHEST_RMC
    setVector3(linkInr[joint_index].Hlin, Mconst_chest, 0, j);
    setVector3(h, Mconst_chest, 3, j);
#else
    setVector3(linkInr[joint_index].Hlin, Mconst, 0, 6+j);
    setVector3(h, Mconst, 3, 6+j);
#endif
  }
#endif
  
#ifdef WITH_ARM_JOINT
  //-------------- Arm momentum tensor -------------
  for(int arm = RLEG ; arm <= LLEG ; arm++){
    for(int j = 0 ; j < 6 ; j++){
      int joint_index = waistToWristPath[arm]->joint(j)->index;
      vector3 h(linkInr[joint_index].Hang - tvmet::cross(total_com, linkInr[joint_index].Hlin));
      setVector3(linkInr[joint_index].Hlin, Mconst_arm[arm], 0, j);
      setVector3(h, Mconst_arm[arm], 3, j);
    }
  }
#endif
  
#ifdef WITH_NECK_JOINT
  //--------------Head momentum tensor---------------
  for(int j = 0 ; j < DOF_NECK ; j++){
    int joint_index = waistToNeckPath->joint(j)->index;
    vector3 h(linkInr[joint_index].Hang - tvmet::cross( total_com, linkInr[joint_index].Hlin ));
    setVector3(linkInr[joint_index].Hlin, Mconst_neck, 0, j);
    setVector3(h, Mconst_neck, 3, j);
  }
#endif
}

//	| Mmat | 	| ME   -MrB->G  |   | Mleg |      | E  -rB->F |  
//	| Hmat | =	|  0        I   | - | Hleg |Jleg-1| 0    E    | 
//        PLfeet = | Mleg |      |   VF    |          PLarm = | Marm |            | Mchest |            | Mhead |
//                 | Hleg |Jleg-1| dankle  |	              | Harm |theta_dot + | Hchest |theta_dot + | Hhead |theta_dot
void WholeBodyMotionGenerator::calcConstraintMomentum(State& Waist, State *Ankle)
{
  dmatrix Jacobian_leg(6,6), Jacobian_leg_star(6,6);
  dvector dankle(6), djoint_leg(6);
  
  setWaistToAnkleJacobian(Waist, Ankle);
#if 1
  PLconst = dzerovector(6);
  for(int leg = RLEG ; leg <= LLEG ; leg++){
    waistToAnklePath[leg]->calcJacobian(Jacobian_leg);
    solveLinearEquationLU(Jacobian_leg, Jankle[leg], Jacobian_leg_star);
#ifndef USE_CHEST_RMC
    Mconst -= ublas::prod(Mconst_leg[leg], Jacobian_leg_star);
#else
    dmatrix66 Mleg(ublas::prod(Mconst_leg[leg], Jacobian_leg_star));
    for(int i = 0 ; i < 6 ; i++ )
      for(int j = 0 ; j < 6 ; j++ )
	Mconst(i,j) -= Mleg(i,j);
#endif
    
    setVector3(Ankle[leg].V, dankle, 0 );
    setVector3(Ankle[leg].W, dankle, 3 );
    solveLinearEquationLU(Jacobian_leg, dankle, djoint_leg);
    PLconst += ublas::prod(Mconst_leg[leg], djoint_leg);
  }
#else
  dmatrix InvJleg[2];
  for(int leg=RLEG; leg<=LLEG; leg++){
    InvJleg[leg].resize(6,6);
    waistToAnklePath[leg]->calcJacobian(Jacobian_leg);
    calcPseudoInverse(Jacobian_leg, InvJleg[leg]);
  }
  dmatrix Mmatleg_star[2];
  PLconst = dzerovector(6);
  for(int leg=RLEG; leg<=LLEG; leg++){
    Mmatleg_star[leg].resize(6,6);
    Mmatleg_star[leg] = prod(Mconst_leg[leg], InvJleg[leg]);
#ifndef USE_CHEST_RMC
    Mconst -= ublas::prod( Mmatleg_star[leg], Jankle[leg]);
#else
    dmatrix66 Mleg(ublas::prod( Mmatleg_star[leg], Jankle[leg]));
    for(int i = 0 ; i < 6 ; i++ )
      for(int j = 0 ; j < 6 ; j++ )
	Mconst(i,j) -= Mleg(i,j);
#endif
    setVector3(Ankle[leg].V, dankle, 0 );
    setVector3(Ankle[leg].W, dankle, 3 );
    PLconst += ublas::prod(Mmatleg_star[leg], dankle);
  }
#endif
  
  PLfree = dzerovector(6);
#ifdef WITH_ARM_JOINT  
  dvector6 djoint_arm;
  for(int arm = RLEG ; arm <= LLEG ; arm++){
    for(int i = 0 ; i < 6 ; i++)
      djoint_arm(i) = waistToWristPath[arm]->joint(i)->dq;
    PLfree += ublas::prod(Mconst_arm[arm], djoint_arm);
  }
#endif
  
#ifdef WITH_CHEST_JOINT  
#ifndef USE_CHEST_RMC
  dvector djoint_chest(DOF_CHEST);
  for(int i = 0; i < DOF_CHEST ; i++)
    djoint_chest(i) = waistToChestPath->joint(i)->dq;
  PLfree += ublas::prod(Mconst_chest, djoint_chest);
#endif
#endif
  
#ifdef WITH_NECK_JOINT
  dvector djoint_neck(DOF_NECK);
  for(int i = 0; i < DOF_NECK ; i++)
    djoint_neck(i) = waistToNeckPath->joint(i)->dq;
  PLfree += ublas::prod(Mconst_neck, djoint_neck);
#endif
}

void WholeBodyMotionGenerator::calcMomentum(dvector6& PL_cmd, State& Waist_cmd, dvector& q_cmd, dvector& qdot_cmd)
{
  waistLink->p = Waist_cmd.P;
  waistLink->R = Waist_cmd.R;
  waistLink->v = Waist_cmd.V;
  waistLink->w = Waist_cmd.W;
  for(int i = 0 ; i < DOF ; i++ ){
    body->joint(i)->q = q_cmd[i];
    body->joint(i)->dq = qdot_cmd[i];
  }
  body->calcForwardKinematics(true);
  
  calcCOGInertia(total_com, total_inr);  
  setInertiaMatrix(Waist_cmd);

  dvector Vctrl(Nwbm);
  setVector3(Waist_cmd.V, Vctrl, 0);
  setVector3(Waist_cmd.W, Vctrl, 3);
  
#ifdef WITH_CHEST_JOINT
#ifdef USE_CHEST_RMC
  for( int i = 0 ; i < DOF_CHEST ; i++ )
    Vctrl(6 + i) = qdot_cmd[DOF_LEG + i];
#endif
#endif
  
  PL_cmd = ublas::prod(Mconst, Vctrl);
  
  dvector6 djoint_leg[2];
  for(int leg = RLEG ; leg <= LLEG ; leg++){
    for(int i = 0 ; i < 6 ; i++){
      djoint_leg[leg](i) = waistToAnklePath[leg]->joint(i)->dq;
      PL_cmd += ublas::prod(Mconst_leg[leg], djoint_leg[leg]);
    }
  }
  
#ifdef WITH_ARM_JOINT
  dvector6 djoint_arm[2];
  for(int arm = RLEG ; arm <= LLEG ; arm++){
    for(int i = 0 ; i < 6 ; i++){
      djoint_arm[arm] = waistToWristPath[arm]->joint(i)->dq;
      PL_cmd += ublas::prod(Mconst_arm[arm], djoint_arm[arm]);
    }
  }
#endif
  
#ifdef WITH_CHEST_JOINT
#ifndef USE_CHEST_RMC
  dvector djoint_chest(DOF_CHEST);
  for(int i = 0 ; i < DOF_CHEST ; i++){
    djoint_chest(i) = waistToChestPath->joint(i)->dq;
    PL_cmd += ublas::prod(Mconst_chest, djoint_chest);
  }
#endif
#endif
  
#ifdef WITH_NECK_JOINT
  dvector djoint_neck(DOF_NECK);
  for(int i = 0; i < DOF_NECK ; i++)
    djoint_neck(i) = waistToNeckPath->joint(i)->dq;
  PL_cmd += ublas::prod(Mconst_neck, djoint_neck);
#endif
  
}
 
void WholeBodyMotionGenerator::generateReferenceMomentum(vector3& COG_cmd, vector3& VCOG_cmd, State& Waist_cmd, vector3& Zmp_cmd, dvector& PLref)
{
  const double CgGain = 10.0;
  
#if 0
  PLref(0) = Total_mass * (CgGain*(COG_cmd(X) - total_com(X)) + VCOG_cmd(X));
  PLref(1) = Total_mass * (CgGain*(COG_cmd(Y) - total_com(Y)) + VCOG_cmd(Y));
  PLref(2) = Total_mass * (CgGain*(COG_cmd(Z) - total_com(Z)) + VCOG_cmd(Z));
  //PLref(2) = Total_mass * CgGain*(0.64 - Waist_cmd.P(Z));

  PLref(3) = 0.0;
  PLref(4) = 0.0;
  PLref(5) = 0.0;
#else
  vector3 Pref(Total_mass * (CgGain*(COG_cmd - total_com) + VCOG_cmd));
  vector3 cz(total_com - Zmp_cmd);
  vector3 Lref( tvmet::cross(cz, Pref));
  Lref(Z) = 0.0;
  setVector3( Pref, PLref, 0);
  setVector3( Lref, PLref, 3);
#endif
}

void WholeBodyMotionGenerator::updateWholeBodyMotion(State& Waist_cmd, dvector& Vctrl_cmd)
{
  Waist_cmd.P += Waist_cmd.V * DT;
  using ::std::numeric_limits;
  double norm_w = norm2(Waist_cmd.W);
  if(norm_w >= numeric_limits<double>::epsilon() ) {
    double th = norm_w * DT;
    vector3 w_n(Waist_cmd.W / norm_w);
    alias(Waist_cmd.R) = rodrigues(w_n, th) * Waist_cmd.R;
  }
  
  getVector3(Waist_cmd.V, Vctrl_cmd, 0);
  getVector3(Waist_cmd.W, Vctrl_cmd, 3);
}

void WholeBodyMotionGenerator::updateWholeBodyMotion(State& Waist_cmd, Jvector& q_cmd, Jvector& qdot_cmd, dvector& Vctrl_cmd)
{
  Waist_cmd.P += Waist_cmd.V * DT;
  using ::std::numeric_limits;
  double norm_w = norm2(Waist_cmd.W);
  if(norm_w >= numeric_limits<double>::epsilon() ) {
    double th = norm_w * DT;
    vector3 w_n(Waist_cmd.W / norm_w);
    alias(Waist_cmd.R) = rodrigues(w_n, th) * Waist_cmd.R;
  }
  
  getVector3(Waist_cmd.V, Vctrl_cmd, 0);
  getVector3(Waist_cmd.W, Vctrl_cmd, 3);
  
  for( int i = 0 ; i < DOF_CHEST ; i++ ){
    q_cmd[DOF_LEG + i] += qdot_cmd[DOF_LEG + i] * DT;
    qdot_cmd[DOF_LEG + i] = Vctrl_cmd[6 + i];
  }
}

void WholeBodyMotionGenerator::calcInverseMomentum(dvector& PL, dvector& Vctrl, vector3& Zmp_cmd, dvector& Vctrl_ref)
{
#if 0
#if 1
  const matrix33 E(tvmet::identity<matrix33>());
  const matrix33 Zero(0.0);
  dmatrix Mzmp(6,6);
  setMatrix33(E,    Mzmp, 0, 0);
  setMatrix33(Zero, Mzmp, 0, 3);
  vector3  mcz(-(total_com - Zmp_cmd));
  matrix33 mcz_hat;
  makeCrossMatrix(mcz_hat, mcz);
  setMatrix33(mcz_hat, Mzmp, 3, 0);
  setMatrix33(E,       Mzmp, 3, 3);
  Mconst -= ublas::prod(Mconst, Mzmp);
#endif
  
  solveLinearEquationLU(Mconst, PL, Vctrl);
#else
  const int LinFlag[3] = {1, 1, 1};
  const int AngFlag[3] = {0, 0, 1};
  const int Nmom = LinFlag[X] + LinFlag[Y] + LinFlag[Z] + AngFlag[X] + AngFlag[Y] + AngFlag[Z];
  dmatrix MH(Nmom, Nwbm), MH_pseu(Nwbm, Nmom), Jnull(Nwbm, Nwbm);
  dvector err(Nmom);
  
  //------------ Pickup considering momentum --------------
  int idx = 0;
  // Linear momentum term (Inertia tensor J, Momentum Pref)
  for(int i = X; i <= Z; i++){
    if( LinFlag[i] == 1 ){
      for(int j = 0 ; j < Nwbm ; j++)
	MH(idx,j) = Mconst(i, j);
      err(idx) = PL(i);
      idx++;
    }
  }
  
  // Angular momentum term (Inertia tensor H, Momentum Lref)
  for(int i = X ; i <= Z ; i++){
    if( AngFlag[i] == 1 ){
      for(int j = 0 ; j < Nwbm ; j++)
	MH(idx,j) = Mconst(i+3, j);
      err(idx) = PL(i+3);
      idx++;
    }
  }

#if 0  
  //------------------------- pseudo inverse --------------------------- 
  // weight for balancing. Bigger weight leads reluctant joint
  double Wpos = 1.0, Watt = 1.0;
  for(int j = 0 ; j < 3 ; j++){
    for(int i = 0 ; i < (int)MH.size1() ; i++)
      MH(i,j) = MH(i,j)/Wpos;
  }
  for(int j = 3 ; j < 6 ; j++){
    for(int i = 0 ; i < (int)MH.size1() ; i++)
      MH(i,j) = MH(i,j)/Watt;
  }
#endif
  
  solveLinearEquationSVD(MH, err, Vctrl);
  calcPseudoInverse(MH, MH_pseu);
  
  //cout << "err=" << err << endl;
  //cout << "dq=" << dq << endl;
  
  //-------- control in the null space ------------
  // I - J_pseu * J
  Jnull = -prod(MH_pseu, MH);
  for(int i = 0 ; i < Nwbm ; i++)
    Jnull(i,i) += 1.0;
  //cout << "Jnull=" << Jnull << endl;
  
  Vctrl += prod(Jnull , Vctrl_ref);        // add modify in null space
  
  // for weight 
  //for(int j = 0 ; j < 3 ; j++) Vctrl(j) /= Wpos;
  //for(int j = 3 ; j < 6 ; j++) Vctrl(j) /= Watt;
#endif
}

void WholeBodyMotionGenerator::makeWholeBodyMotion(State& Waist_cmd, vector3& COG_cmd, vector3& VCOG_cmd, State *Ankle_cmd, vector3& Zmp_cmd, dvector& Vctrl_ref)
{
  waistLink->p = Waist_cmd.P;
  waistLink->R = Waist_cmd.R;
  waistLink->v = Waist_cmd.V;
  waistLink->w = Waist_cmd.W;
  //for(int i = 0 ; i < DOF ; i++ ){
  //body->joint(i)->q = q_cmd(i);
  //body->joint(i)->dq = qdot_cmd(i);
  //}
  
  calcCOGInertia(total_com, total_inr);  
  setInertiaMatrix(Waist_cmd);
  calcConstraintMomentum(Waist_cmd, Ankle_cmd);
  dvector PLref(6);
  generateReferenceMomentum(COG_cmd, VCOG_cmd, Waist_cmd, Zmp_cmd, PLref);
  dvector PL(PLref - PLconst - PLfree);
  dvector Vctrl(Nwbm);
  calcInverseMomentum(PL, Vctrl, Zmp_cmd, Vctrl_ref);
  updateWholeBodyMotion(Waist_cmd, Vctrl);
}

void WholeBodyMotionGenerator::makeWholeBodyMotion(State& Waist_cmd, Jvector& q_cmd, Jvector& qdot_cmd, vector3& COG_cmd, vector3& VCOG_cmd, State *Ankle_cmd, vector3& Zmp_cmd, dvector& Vctrl_ref)
{
  waistLink->p = Waist_cmd.P;
  waistLink->R = Waist_cmd.R;
  waistLink->v = Waist_cmd.V;
  waistLink->w = Waist_cmd.W;
  for(int i = 0 ; i < DOF_CHEST ; i++ ){
    body->joint(DOF_LEG+i)->q = q_cmd[DOF_LEG+i];
    body->joint(DOF_LEG+i)->dq = qdot_cmd[DOF_LEG+i];
  }
  body->calcForwardKinematics(true);
  
  calcCOGInertia(total_com, total_inr);  
  setInertiaMatrix(Waist_cmd);
  calcConstraintMomentum(Waist_cmd, Ankle_cmd);
  dvector PLref(6);
  generateReferenceMomentum(COG_cmd, VCOG_cmd, Waist_cmd, Zmp_cmd, PLref);
  dvector PL(PLref - PLconst - PLfree);
  dvector Vctrl(Nwbm);
  calcInverseMomentum(PL, Vctrl, Zmp_cmd, Vctrl_ref);
  updateWholeBodyMotion(Waist_cmd, q_cmd, qdot_cmd, Vctrl);
}

void WholeBodyMotionGenerator::makeWholeBodyMotion(dvector& Vctrl, dvector& PL_ref, State& Waist_cmd, vector3& COG_cmd, vector3& VCOG_cmd, State *Ankle_cmd, vector3& Zmp_cmd, dvector& Vctrl_ref)
{
  waistLink->p = Waist_cmd.P;
  waistLink->R = Waist_cmd.R;
  waistLink->v = Waist_cmd.V;
  waistLink->w = Waist_cmd.W;
  //for(int i = 0 ; i < DOF ; i++ ){
  //body->joint(i)->q = q_cmd(i);
  //body->joint(i)->dq = qdot_cmd(i);
  //}
  
  calcCOGInertia(total_com, total_inr);  
  setInertiaMatrix(Waist_cmd);
  calcConstraintMomentum(Waist_cmd, Ankle_cmd);
  generateReferenceMomentum(COG_cmd, VCOG_cmd, Waist_cmd, Zmp_cmd, PL_ref);
  dvector PL(PL_ref - PLconst - PLfree);
  calcInverseMomentum(PL, Vctrl, Zmp_cmd, Vctrl_ref);
}

void WholeBodyMotionGenerator::makeWholeBodyMotion(dvector& Vctrl, dvector& PL_ref, State& Waist_cmd, Jvector& q_cmd, Jvector& qdot_cmd, vector3& COG_cmd, vector3& VCOG_cmd, State *Ankle_cmd, vector3& Zmp_cmd, dvector& Vctrl_ref)
{
  waistLink->p = Waist_cmd.P;
  waistLink->R = Waist_cmd.R;
  waistLink->v = Waist_cmd.V;
  waistLink->w = Waist_cmd.W;
  for(int i = 0 ; i < DOF_CHEST ; i++ ){
    body->joint(DOF_LEG+i)->q = q_cmd[DOF_LEG+i];
    body->joint(DOF_LEG+i)->dq = qdot_cmd[DOF_LEG+i];
  }
  body->calcForwardKinematics(true);
  
  calcCOGInertia(total_com, total_inr);  
  setInertiaMatrix(Waist_cmd);
  calcConstraintMomentum(Waist_cmd, Ankle_cmd);
  generateReferenceMomentum(COG_cmd, VCOG_cmd, Waist_cmd, Zmp_cmd, PL_ref);
  dvector PL(PL_ref - PLconst - PLfree);
  calcInverseMomentum(PL, Vctrl, Zmp_cmd, Vctrl_ref);
}
